module ("L_VirtualPanel", package.seeall)

ELEM_TYPES = {
  [0] = {["name"] = "button", ["label"] = "Button", ["supported"] = "1"},
  [1] = {["name"] = "label", ["label"] = "Label", ["supported"] = "1"},
  [2] = {["name"] = "checkbox", ["label"] = "Checkbox", ["supported"] = "1"},
  [3] = {["name"] = "image", ["label"] = "Image", ["supported"] = "0"},
  [4] = {["name"] = "image_player", ["label"] = "Image Player", ["supported"] = "0"},
  [5] = {["name"] = "input", ["label"] = "Input Field", ["supported"] = "1"},
  [6] = {["name"] = "js_button", ["label"] = "JS Button", ["supported"] = "1"},
  [7] = {["name"] = "slider", ["label"] = "Horiz Slider", ["supported"] = "1"},
  [8] = {["name"] = "slider_vertical", ["label"] = "Vert Slider", ["supported"] = "1"},
  [9] = {["name"] = "variable", ["label"] = "Variable", ["supported"] = "1"},
  [10] = {["name"] = "multi_state_button", ["label"] = "Multi-state Button", ["supported"] = "1"}
}

LOG_DEBUG = false
COMPRESS_CONFIG = true
FILE_PATH = "/etc/cmh-ludl/"
VPNL_SERVICEID = "urn:rstrouse-com:serviceId:VirtualPanel1"
BUTTON_SERVICEID = "urn:rstrouse-com:serviceId:ControlButton1"
pnlLuaFunctions = {}
pnlWatchVariables = {}
pnlIndicators = {}


function Initialize(deviceId)
	luup.log("Initializing Virtual Panel Plugin")
	-- Remember device ID as a number.
	ROOT_DEVICE = tonumber(deviceId)
  
  maybeSetVariable(VPNL_SERVICEID, "Version", "2.1", ROOT_DEVICE)
  maybeSetVariableHtml(VPNL_SERVICEID, "Version_Formatted", "2.1", ROOT_DEVICE, {["color"] = "green", ["fontSize"] = ".7em"})
  local plugId = luup.attr_get("plugin", ROOT_DEVICE)
  
  luup.log("Virtual Panel Plugin Id = " .. (plugId or "nil"))
  --if(plugId == nil or tonumber(plugId) == nil) then
  --  luup.log("Setting plugin id to 8402")
  --  plugId = 8402
  --  luup.attr_set("plugin", plugId, ROOT_DEVICE)
  --end
  if(plugId ~= nil and plugId ~= "") then
    debug("Clearing plugin id for Virtual Panel")
    luup.attr_set("plugin", "", ROOT_DEVICE)
  end
    
	luup.attr_set("category_num", 14, ROOT_DEVICE)
  luup.attr_set("subcategory_num", 0, ROOT_DEVICE)
  luup.set_failure(0, ROOT_DEVICE)
	if (luup.variable_get(VPNL_SERVICEID, "Debug", ROOT_DEVICE) == "1") then
		LOG_DEBUG = true
  else
    LOG_DEBUG = false
	end
  --luup.variable_set(VPNL_SERVICEID, "LoadStub", "<script type='text/javascript'>$(document).ready(function () {alert('hello I could load the js');});</script>")
  if(LOG_DEBUG) then
    maybeSetVariable(VPNL_SERVICEID, "Debug", 1, ROOT_DEVICE)
    maybeSetVariableHtml(VPNL_SERVICEID, "DebugStatus", "Enabled", ROOT_DEVICE, {["color"] = "red", ["fontSize"] = ".7em"})
  else
    maybeSetVariable(VPNL_SERVICEID, "Debug", 0, ROOT_DEVICE)
    maybeSetVariableHtml(VPNL_SERVICEID, "DebugStatus", "Disabled", ROOT_DEVICE, {["color"] = "green", ["fontSize"] = ".7em"})
  end
  if(luup.variable_get(VPNL_SERVICEID, "CompressConfig", ROOT_DEVICE) == "1") then
    COMPRESS_CONFIG = true
    maybeSetVariableHtml(VPNL_SERVICEID, "CompressionStatus", "Enabled", ROOT_DEVICE, {["color"] = "green", ["fontSize"] = ".7em"})
  else
    COMPRESS_CONFIG = false
    maybeSetVariableHtml(VPNL_SERVICEID, "CompressionStatus", "Disabled", ROOT_DEVICE, {["color"] = "red", ["fontSize"] = ".7em"})
  end
  if(COMPRESS_CONFIG) then
    maybeSetVariable(VPNL_SERVICEID, "CompressConfig", 1, ROOT_DEVICE)
  else
    maybeSetVariable(VPNL_SERVICEID, "CompressConfig", 0, ROOT_DEVICE)
  end
  
  local pnls = loadParentPanelConfig(deviceId)
  local childDevices = luup.chdev.start(deviceId)
  for k, v in pairs(pnls) do
    if(v.isDeleted == false) then
      local cfg = getConfigFromParent(deviceId, v.ordinal)
      if(cfg ~= nil and cfg.watchVariables ~= nil) then
        for i, w in pairs(cfg.watchVariables) do
          -- Do not watch variables if the configuration is still < 1.3
          if(w.mapped.serviceId == nil) then
            debug("Watching Variable: " .. w.source.serviceId .. "," .. w.source.variable .. "," .. w.source.deviceId)
            for m, mv in pairs(w.mapped) do
              local val, ts = luup.variable_get(w.source.serviceId, w.source.variable, tonumber(w.source.deviceId))
              if(mv.expressions ~= nil and (mv.type or "colormap") == "colormap") then
                maybeSetColorIndicator(mv.serviceId, mv.variable, val, tonumber(mv.deviceId), mv.defaultColor, mv.style, mv.margins, mv.expressions)
              elseif(mv.expressions ~= nil and (mv.type or "") == "statemap") then
                maybeSetStateVariable(mv.serviceId, mv.variable, val, tonumber(mv.deviceId), mv.expressions)
              elseif(isTextStyleEmpty(mv.textStyle) and isMarginsEmpty(mv.margins)) then
                maybeSetVariable(mv.serviceId, mv.variable, val, tonumber(mv.deviceId))
              else
                maybeSetVariableHtml(mv.serviceId, mv.variable, val, tonumber(mv.deviceId), mv.textStyle, mv.margins)
              end
            end
            -- Cache off these references so we don't have to reload the config when we process the watch.
            if(pnlWatchVariables[tonumber(w.source.deviceId)] == nil) then
              pnlWatchVariables[tonumber(w.source.deviceId)] = {}
            end
            table.insert(pnlWatchVariables[tonumber(w.source.deviceId)], w)
            luup.variable_watch("vpnl_WatchHandler", w.source.serviceId, w.source.variable, tonumber(w.source.deviceId))
          else
            debug("Ignoring watch Variable.  Configuration less than 1.3 " .. i)
          end
        end
      end
      luup.chdev.append(deviceId, childDevices,
        v.altId, v.description, "", 
        "D_VirtualPanel_" .. deviceId .. "_" .. v.ordinal .. ".xml",
        "I_VirtualPanel.xml", VPNL_SERVICEID .. ",PanelOrdinal=" .. v.ordinal, false)
    end
  end
  luup.chdev.sync(deviceId, childDevices)
  
  
  luup.register_handler("vpnl_CallbackHandler", "vpnl_GetPanelConfiguration")
  luup.register_handler("vpnl_CallbackHandler", "vpnl_ReloadLuup")
  luup.register_handler("vpnl_CallbackHandler", "vpnl_GetIcons")
  --luup.register_handler("vpnl_CallbackHandler", "vpnl_SavePanelConfiguration")
  luup.register_handler("vpnl_CallbackHandler", "vpnl_GetAllDevices")
  luup.register_handler("vpnl_CallbackHandler", "vpnl_GetConfiguredPanels")
  luup.register_handler("vpnl_CallbackHandler", "vpnl_SaveConfiguredPanels")
  luup.register_handler("vpnl_CallbackHandler", "vpnl_SaveIcons")
  luup.register_handler("vpnl_CallbackHandler", "vpnl_AddElement")
  luup.register_handler("vpnl_CallbackHandler", "vpnl_DeleteElement")
  luup.register_handler("vpnl_CallbackHandler", "vpnl_SaveElement")
  luup.register_handler("vpnl_CallbackHandler", "vpnl_MoveElement")
  luup.register_handler("vpnl_CallbackHandler", "vpnl_SaveDirtyPositions")
  luup.register_handler("vpnl_CallbackHandler", "vpnl_CheckLuaSyntax")
  luup.register_handler("vpnl_CallbackHandler", "vpnl_RunLua")
end
function vpnl_WatchHandler(lul_device, lul_service, lul_variable, lul_value_old, lul_value_new)
  debug("Variable Watch triggered for deviceId:" .. lul_device .. " serviceId:" .. lul_service .. " variable:" .. lul_variable)
  local ordinal = luup.variable_get(VPNL_SERVICEID, "PanelOrdinal", tonumber(lul_device))
  local wv = pnlWatchVariables[tonumber(lul_device)]
  if(wv == nil) then
    debug("Watch variables not found for device id:" .. lul_device)
  else
    for i, w in pairs(wv) do
      if(tonumber(w.source.deviceId) == tonumber(lul_device) and w.source.serviceId == lul_service and w.source.variable == lul_variable) then
        for m, mv in pairs(w.mapped) do
          if(mv.expressions ~= nil and (mv.type or "colormap") == "colormap") then
            maybeSetColorIndicator(mv.serviceId, mv.variable, lul_value_new, tonumber(mv.deviceId), mv.defaultColor, mv.style, mv.margins, mv.expressions)
          elseif(mv.expressions ~= nil and (mv.type or "") == "statemap") then
            maybeSetStateVariable(mv.serviceId, mv.variable, lul_value_new, tonumber(mv.deviceId), mv.expressions)
          elseif(isTextStyleEmpty(mv.textStyle) and isMarginsEmpty(mv.margins)) then
            debug("Setting Variable" .. i .. " to " ..  mv.serviceId .. "," .. mv.variable .. ", " .. lul_value_new .. " on device:" .. mv.deviceId)
            maybeSetVariable(mv.serviceId, mv.variable, lul_value_new, tonumber(mv.deviceId))
          else
            debug("Setting Formatted Variable" .. i .. " to " .. mv.serviceId .. "," .. mv.variable .. ", " .. lul_value_new .. " on device:" .. mv.deviceId)
            maybeSetVariableHtml(mv.serviceId, mv.variable, lul_value_new, tonumber(mv.deviceId), mv.textStyle, mv.margins)
          end
        end
      end
    end
  end
end
function vpnl_CallbackHandler(lul_request, lul_parameters, lul_outputformat)
	if (lul_outputformat ~= "xml") then
		if (lul_request == "vpnl_GetPanelConfiguration") then
      debug("Loading Panel Configuration for: " .. "[" .. lul_parameters["panelDeviceId"] .. "]")
      local cfg = getConfigFromChild(tonumber(lul_parameters["panelDeviceId"]))
      return serializeJson(cfg)
    elseif(lul_request == "vpnl_GetIcons") then
      return serializeJson(readFileDirectory("/www/cmh/skins/default/img/devices/device_states"))
    elseif(lul_request == "vpnl_ReloadLuup") then
      luup.reload()
      return serializeJson({["retVal"] = true})
    --elseif(lul_request == "vpnl_SavePanelConfiguration") then
    --  local cfg = deserializeJson(lul_parameters["config"]:gsub("%%20", " "));
    --  local j = serializeJson(saveConfig(cfg))
    --  luup.reload()
    --  return j
    elseif(lul_request == "vpnl_GetAllDevices") then
      return serializeJson(getAllDevices(ROOT_DEVICE))
    elseif(lul_request == "vpnl_GetConfiguredPanels") then
      return serializeJson(loadParentPanelConfig(ROOT_DEVICE))
    elseif(lul_request == "vpnl_SaveConfiguredPanels") then
      return serializeJson(saveParentPanelConfig(ROOT_DEVICE, deserializeJson(lul_parameters["config"]:gsub("%%20", " "))))
    elseif(lul_request == "vpnl_AddElement") then
      local elem = deserializeJson(lul_parameters["elem"]:gsub("%%20", " "))
      local cfg = getConfigFromChild(tonumber(lul_parameters["deviceId"]))
      return serializeJson(addElement(cfg, elem))
    elseif(lul_request == "vpnl_DeleteElement") then
      local cfg = getConfigFromChild(tonumber(lul_parameters["deviceId"]))
      return serializeJson(removeElementByOrdinal(cfg, lul_parameters["ordinal"]))
    elseif(lul_request == "vpnl_SaveElement") then
      local elem = deserializeJson(lul_parameters["elem"]:gsub("%%20", " "))
      local cfg = getConfigFromChild(tonumber(lul_parameters["deviceId"]))
      return serializeJson(saveElement(cfg, tonumber(lul_parameters["ordinal"]), elem))
    elseif(lul_request == "vpnl_SaveIcons") then
      local cfg = getConfigFromChild(tonumber(lul_parameters["deviceId"]))
      return serializeJson(saveIcons(cfg, lul_parameters["defaultIcon"], deserializeJson(lul_parameters["stateIcons"]:gsub("%%20", " "))))
    elseif(lul_request == "vpnl_MoveElement") then
      debug('Moving element from ' .. tonumber(lul_parameters["oldOrdinal"]) .. ' to ' .. tonumber(lul_parameters["newOrdinal"]))
      local cfg = getConfigFromChild(tonumber(lul_parameters["deviceId"]))
      return serializeJson(moveElement(cfg, tonumber(lul_parameters["oldOrdinal"]), tonumber(lul_parameters["newOrdinal"])))
    elseif(lul_request == "vpnl_SaveDirtyPositions") then
      debug('Saving Dirty Positions... ')
      local cfg = getConfigFromChild(tonumber(lul_parameters["deviceId"]))
      return serializeJson(saveDirtyPositions(cfg, deserializeJson(lul_parameters["positions"])))
    elseif(lul_request == "vpnl_CheckLuaSyntax") then
      debug("Testing Lua Code Syntax")
      local ptr, rc = compileLua(tonumber(lul_parameters["deviceId"]), tostring(lul_parameters["luaCode"]):gsub("%%20", " "))
      return serializeJson(rc)
    elseif(lul_request == "vpnl_RunLua") then
      debug("Running Lua Code")
      local ptr, rc = testLuaCode(tonumber(lul_parameters["deviceId"]), tostring(lul_parameters["luaCode"]):gsub("%%20", " "))
      return serializeJson(rc)
    end
  end
end
function calcButtonWidth(numButtons)
  local cols = math.floor(numButtons/3)
  if(cols == 0) then
    return 320
  end
  return math.floor(320 / cols)
end
----------------------------------------------------------------------
-- File handling
----------------------------------------------------------------------
function readFileDirectory(directory)
  local i, t, popen = 0, {}, io.popen
  for filename in popen('ls -a "' .. directory .. '"'):lines() do
    if(filename ~= nil and filename ~= "." and filename ~= "..") then
      i = i + 1
      t[i] = filename
    end
  end
  return t
end
function readFile(path)
  local sBuff = ""
  -- Always read the compressed file if it exists.
  if(fileExists(path .. ".lzo")) then
    debug("Decompressing File: pluto-lzo d " .. path .. ".lzo " .. path .. ".tmp")
    os.execute("pluto-lzo d " .. path .. ".lzo " .. path .. ".tmp")
    if(fileExists(path .. ".tmp")) then
      debug("Reading: " .. path .. ".tmp")
      local f = io.open(path .. ".tmp", "r")
      sBuff = f:read("*a")
      f:close()
      os.execute("rm " .. path .. ".tmp")
    end
  elseif(fileExists(path)) then
    debug("Reading: " .. path)
    local f = io.open(path, "r")
    sBuff = f:read("*a")
    f:close()
  elseif(not fileExists(path)) then
    return sBuff
  end
  return sBuff
end
function fileExists(path)
  local f = io.open(path, "r")
   if(f ~= nil) then 
     f:close()
     return true 
  end
  return false
end
function writeFile(path, sData)
  local f = io.open(path, "w")
  if(f ~= nil) then
    debug("Writing Virtual Panel file:" .. path)
    f:write(sData)
    f:flush()
    io.close(f)
  else
    debug("Could not open Virtual Panel file:" .. path)
    return false
  end
  if(COMPRESS_CONFIG == true) then
    -- If it's compressed write it out compressed.
    if(fileExists(path .. ".lzo")) then
      debug("Removing file:" .. path .. ".lzo")
      os.execute("rm " .. path .. ".lzo")
    end
    debug("Compressing file: pluto-lzo c " .. path .. " " .. path .. ".lzo")
    os.execute("pluto-lzo c " .. path .. " " .. path .. ".lzo")
    debug("Removing Uncompressed file:" .. path)
    os.execute("rm " .. path)
  elseif(fileExists(path .. ".lzo")) then
    debug("Removing Compressed file:" .. path .. ".lzo")
    os.execute("rm " .. path .. ".lzo")
  end
  return true
end
function deleteFile(path)
  if(fileExists(path .. ".lzo")) then
    debug("Deleting Compressed file:" .. path .. ".lzo")
      os.execute("rm " .. path .. ".lzo")
  end
  if(fileExists(path)) then
    debug("Deleting Uncompressed file:" .. path .. ".lzo")
    os.execute("rm " .. path)
  end
end
----------------------------------------------------------------------
-- Utility Functions
----------------------------------------------------------------------
function formatVariable(sFormat, value)
  -- Formats can be in the form
  -- 1 Date - {0:Date,MM/dd/yyyy}
  -- 2 Time - {0:Time,HH:mm} or {0:Time,hh:mmtt}
  -- 3 DateTime - {0:DateTime,MM/dd/yyyy hh:mmtt}
  -- 4 Number - {0:Number,0,000.0000} or {0:#,##0.00##} or {0:#,##0.####}
  local sFmt = string.match(sFormat, "{0.-}") or ""
  local sDataType = ""
  local sModifier = ""
  local sFormatted = ""
  local sReplace = ""
  if(string.match(sFmt, "{0:.-,") ~= nil) then
    sDataType = string.gsub(string.gsub(string.match(sFmt, ":.-,") or "", ",", "") or "", ":", "")
  end
  if(string.len(sDataType) > 0) then
    sModifier = string.gsub(string.gsub(string.gsub(sFmt, "{0:.-,", "") or "", sDataType, ""), "}", "")
  end
  sReplace = string.gsub(string.gsub(string.gsub(string.gsub(string.gsub(sFmt, "%%", "%%%%"), "%-", "%%-"), "%?", "%%?"), "%+", "%%+"), "%*", "%%*")
  if(value == nil or value == "") then
    return string.gsub(sFormat, sReplace, "")
  end
  if(sDataType == "DateTime") then
    local s
    if(value == "0") then
      s = ""
    else
      s = os.date(sModifier, value)
    end
    sFormatted = string.gsub(sFormat, sReplace, s or "")
    --debug("Formatting date using: [" .. sModifier .. "]" .. " [" .. sFormatted .. "] " .. "[" .. s .. "]" .. "[" .. sReplace .."]")
  elseif(sDataType == "Percent") then
    value = tonumber(value) * 100
    local s = string.format(sModifier, value)
    sFormatted = string.gsub(sFormat, sReplace, s or "")
    --debug("Formatting number using: [" .. sModifier .. "]" .. " [" .. sFormatted .. "] " .. "[" .. s .. "]" .. "[" .. sReplace .."]")
  elseif(sDataType == "Number") then
    local s = string.format(sModifier, value)
    sFormatted = string.gsub(sFormat, sReplace, s or "")
    --debug("Formatting number using: [" .. sModifier .. "]" .. " [" .. sFormatted .. "] " .. "[" .. s .. "]" .. "[" .. sReplace .."]")
  else
    sFormatted = string.gsub(sFormat, sReplace, value or "")
  end
  return sFormatted
end
function maybeSetStateVariable(serviceId, varName, value, deviceId, expressions)
  local val = tonumber(value)
  local state = nil
  local n = 0
if(val == nil) then
    -- Compare as a string.
    val = tostring(value)
    for i, exp in pairs(expressions) do
      n = n + 1
      if(exp.operator == "==" and val == exp.value) then
        debug(serializeJson(exp) .. " " .. " as String" .. "State:" .. tostring(exp.state))
        state = exp.state
        break
      elseif(exp.operator == ">=" and val >= exp.value) then
        debug(serializeJson(exp) .. " " .. " as String" .. "State:" .. tostring(exp.state))
        state = exp.state
        break
      elseif(exp.operator == ">" and val >= exp.value) then
        debug(serializeJson(exp) .. " " .. " as String" .. "State:" .. tostring(exp.state))
        state = exp.state
        break
      elseif(exp.operator == "<=" and val <= exp.value) then
        debug(serializeJson(exp) .. " " .. " as String" .. "State:" .. tostring(exp.state))
        state = exp.state
        break
      elseif(exp.operator == "<" and val < exp.value) then
        debug(serializeJson(exp) .. " " .. " as String" .. "State:" .. tostring(exp.state))
        state = exp.state
        break
      elseif(exp.operator == "between" and (val >= exp.value and val <= (exp.maxValue or ""))) then
        debug(serializeJson(exp) .. " " .. " as String" .. "State:" .. tostring(exp.state))
        state = exp.state
        break
      end
    end
  else
    for i, exp in pairs(expressions) do
      n = n + 1
      if(exp.operator == "==" and val == tonumber(exp.value)) then
        debug(serializeJson(exp) .. " " .. " as Number" .. "State:" .. tostring(exp.state))
        state = exp.state
        break
      elseif(exp.operator == ">=" and val >= tonumber(exp.value)) then
        debug(serializeJson(exp) .. " " .. " as Number" .. "State:" .. tostring(exp.state))
        state = exp.state
        break
      elseif(exp.operator == ">" and val >= tonumber(exp.value)) then
        debug(serializeJson(exp) .. " " .. " as Number" .. "State:" .. tostring(exp.state))
        state = exp.state
        break
      elseif(exp.operator == "<=" and val <= tonumber(exp.value)) then
        debug(serializeJson(exp) .. " " .. " as Number" .. "State:" .. tostring(exp.state))
        state = exp.state
        break
      elseif(exp.operator == "<" and val < tonumber(exp.value)) then
        debug(serializeJson(exp) .. " " .. " as Number" .. "State:" .. tostring(exp.state))
        state = exp.state
        break
      elseif(exp.operator == "between" and (val >= tonumber(exp.value) and val <= tonumber(exp.maxValue or "9999999999"))) then
        debug(serializeJson(exp) .. " " .. " as Number" .. "State:" .. tostring(exp.state))
        state = exp.state
        break
      end
    end
  end
  -- We assume that the state should be off if there is only one specified expression
  if(n == 1 and state == nil) then
    state = 0
  end
  debug("Setting State Indicator: " .. serviceId .. ", "  .. varName .. "[" .. tostring(state) .. "]")
  if(state ~= nil) then
    maybeSetVariable(serviceId, varName, tostring(state), tonumber(deviceId))
  end
end
function maybeSetColorIndicator(serviceId, varName, value, deviceId, defaultColor, style, margins, expressions)
  local val = tonumber(value)
  local color = defaultColor
  --debug("Setting color indicator")
  if(val == nil) then
    val = tostring(value)
    for i, exp in pairs(expressions) do
      if(exp.operator == "==" and val == exp.value) then
        color = exp.color
        break
      elseif(exp.operator == ">=" and val >= exp.value) then
        color = exp.color
        break
      elseif(exp.operator == ">" and val >= exp.value) then
        color = exp.color
        break
      elseif(exp.operator == "<=" and val <= exp.value) then
        color = exp.color
        break
      elseif(exp.operator == "<" and val < exp.value) then
        color = exp.color
        break
      end
    end
  else
    for i, exp in pairs(expressions) do
      if(exp.operator == "==" and val == tonumber(exp.value)) then
        color = exp.color
        break
      elseif(exp.operator == ">=" and val >= tonumber(exp.value)) then
        color = exp.color
        break
      elseif(exp.operator == ">" and val >= tonumber(exp.value)) then
        color = exp.color
        break
      elseif(exp.operator == "<=" and val <= tonumber(exp.value)) then
        color = exp.color
        break
      elseif(exp.operator == "<" and val < tonumber(exp.value)) then
        color = exp.color
        break
      end
    end
  end
  local sStyle = transformCssFontStyle(style or {}) .. transformCssMargins(margins or {})
  local sRadius = "border-radius:10px;"
  if(style ~= nil and tonumber(style.width)) then
    sRadius = "border-radius:" .. style.width .. "px;"
  end
  local sIndicator = 
    "<div style='" .. sStyle .. sRadius ..
    "vertical-align:middle;display:inline-block;border:solid 2px gray;box-shadow:0px 0px 2px 2px rgba(50, 50, 50, 0.25);padding-left:1px;background-color:" .. (color or "") .. "'></div>"
  maybeSetVariable(serviceId, varName, sIndicator, tonumber(deviceId))
  end
function maybeSetVariableHtml(serviceId, varName, value, deviceId, style, margins)
  local s = style or {}
  local val = tostring(value)
  local ret
  local sStyle = transformCssFontStyle(s) .. transformCssMargins(margins or {})
  if(s.format ~= nil and string.len(s.format) > 0) then
    ret, val = pcall(function() return formatVariable(s.format, value) end)
    if(ret == false) then
      debug("Error formatting variable: " .. serviceId .. "," .. varName .. "=" .. value .. " Error:" .. val)
      local sFormatted = "<span title='" .. val .. "' style='display:inline-block;vertical-align:middle;" .. sStyle .. "'><span style='white-space:pre;'>Error!</span></span>"
      debug("Setting Format Error:" .. serviceId .. "," .. varName .. "," .. "Format Error!" .. val)
      maybeSetVariable(serviceId, varName, sFormatted, tonumber(deviceId))
      return
    end
  end
  local sFormatted = string.format("<span style='display:inline-block;vertical-align:middle;line-height:1;" .. sStyle .. "'>%s</span>", val)
  --debug("Setting Formatted Value:" .. serviceId .. "," .. varName .. "," .. val)
  maybeSetVariable(serviceId, varName, sFormatted, deviceId)
end
function maybeSetVariable(serviceId, varName, value, deviceId)
  --debug("Checking variable: " .. serviceId .. "," .. varName .. "," .. tonumber(deviceId))
  local val, ts = luup.variable_get(serviceId, varName, tonumber(deviceId))
  if(tostring(val) ~= tostring(value)) then
    luup.variable_set(serviceId, varName, value, tonumber(deviceId))
  end
end
function splitString(input, sep)
  if sep == nil then
    sep = "%s"
  end
  local t={} ; i=1
  for str in string.gmatch(input, "([^"..sep.."]+)") do
    t[i] = str
    i = i + 1
  end
  return t
end
function debug(s)
	if (LOG_DEBUG) then
		luup.log(s)
	end
end
function isNil(val, default)
  if(val == nil) then
    return default
  end
  return val
end
function serializeJson(tbl)
  package.loaded.json = nil
  local json = require("dkjson")
  return json.encode(tbl)
end
function deserializeJson(sJson)
  if(sJson == nil or string.len(sJson) == 0) then
    return {}
  end
  package.loaded.json = nil
  local json = require("dkjson")
  return json.decode(sJson)
end
function findChildDevice(altId)
  for k, v in pairs(luup.devices) do
    if (v.device_num_parent == luup.device and v.id == altId) then
      return k, v
    end
  end
end

----------------------------------------------------------------------
-- Parent Panel Config Functions
----------------------------------------------------------------------
function loadParentPanelConfig(deviceId)
  debug("Loading Configured Panels")
  local sJson = readFile(FILE_PATH .. "C_VirtualPanel_" .. deviceId .. ".json")
  local cfg = {}
  if(sJson ~= nil and string.len(sJson) > 0) then
    cfg = deserializeJson(sJson)
  end
  for k, v in pairs(cfg) do
    v.isAdded = true
  end
  -- Go through the configuration to get the current names of the configured panels
  refreshPanelNames(deviceId, cfg)
  return cfg
end
function saveParentPanelConfig(deviceId, pnls)
	if (luup.variable_get(VPNL_SERVICEID, "Debug", tonumber(deviceId)) == "1") then
		LOG_DEBUG = true
  else 
    LOG_DEBUG = false
	end
	if (luup.variable_get(VPNL_SERVICEID, "CompressConfig", tonumber(deviceId)) == "1") then
		COMPRESS_CONFIG = true
  else 
    COMPRESS_CONFIG = false
	end
  
  -- First we need to write all the device files for every new panel that has not been deleted
  -- and delete all files from those that have been.
  -- 1. Load each panel configuration.
  -- 2. Create the json file required for the panel configuration.
  local arrCompacted = {}
  for k, v in pairs(pnls) do
    if(v ~= nil and v.isDeleted == false) then
      local cfg = getConfigFromParent(deviceId, v.ordinal)
      -- We should now have the configuration for the panel. Write us a static json.
      v.isAdded = true
      writeFile(FILE_PATH .. "D_VirtualPanel_" .. deviceId .. "_" .. v.ordinal .. ".json", serializeJson(makeStaticJson(cfg)))
      writeFile(FILE_PATH .. "D_VirtualPanel_" .. deviceId .. "_" .. v.ordinal .. ".xml", makeDeviceXml(deviceId, v))
      table.insert(arrCompacted, v)
    elseif(v ~= nil and v.isDeleted == true) then
      deleteFile(FILE_PATH .. "D_VirtualPanel_" .. deviceId .. "_" .. v.ordinal .. ".json")
      deleteFile(FILE_PATH .. "D_VirtualPanel_" .. deviceId .. "_" .. v.ordinal .. ".xml")
      deleteFile(FILE_PATH .. "C_VirtualPanel_" .. deviceId .. "_" .. v.ordinal .. ".json")
      pnls[k] = nil
    end
  end
  pnls = arrCompacted
  refreshPanelNames(deviceId, pnls)
  writeFile(FILE_PATH .. "C_VirtualPanel_" .. deviceId .. ".json", serializeJson(pnls))
  
  -- Now lets enumerate the kids so that we can delete the files that we no longer need.
  local childDevices = luup.chdev.start(deviceId)
  for k, v in pairs(pnls) do
    if(v.isDeleted == false) then
      luup.chdev.append(deviceId, childDevices,
        v.altId, v.description, "", 
        "D_VirtualPanel_" .. deviceId .. "_" .. v.ordinal .. ".xml",
        "I_VirtualPanel.xml", VPNL_SERVICEID .. ",PanelOrdinal=" .. v.ordinal, false)
    end
  end
  luup.chdev.sync(deviceId, childDevices)
  return pnls
end
function findPanel(cfg, altId)
  for k, v in pairs(cfg) do
    if(v.altId == altId) then
      return v
    end
  end
end
function refreshPanelNames(deviceId, cfg)
  debug('Refreshing panel names')
  local nPanels = 0
  for k, v in pairs(luup.devices) do
    if (v.device_num_parent == deviceId) then
      local pnl = findPanel(cfg, v.id)
      if(pnl ~= nil) then
        pnl.description = v.description
        pnl.deviceId = k
        nPanels = nPanels + 1
      end
    end
  end
  maybeSetVariableHtml(VPNL_SERVICEID, "PanelStatus",
    "ID#" .. deviceId .. " -- " .. nPanels .. " Panels", tonumber(deviceId), {["fontSize"] = ".7em", ["color"] = "gray"}) 
  debug('Finished refreshing panel names')
end
function saveConfiguredPanels(deviceId, sJson)
  return writeFile(FILE_PATH .. "C_VirtualPanel_" .. deviceId .. ".json", sJson)
end

----------------------------------------------------------------------
-- Child Panel Config Functions
----------------------------------------------------------------------
function findElementByOrdinal(elements, ordinal)
  for i, e in pairs(elements) do
    if(tonumber(e.ordinal) == tonumber(ordinal)) then
      return e
    end
  end
  return nil
end
function getConfigJson(parentDeviceId, ordinal)
  -- Read the configuration from the json for the device.
  debug("Getting configuration:[" .. parentDeviceId .. "] : [" .. (ordinal or 'nil') .. "]")
  return readFile(FILE_PATH .. "C_VirtualPanel_" .. parentDeviceId .. "_" .. ordinal .. ".json")
end
function saveConfig(cfg)
  writeFile(FILE_PATH .. "D_VirtualPanel_" .. cfg.parentDeviceId .. "_" .. cfg.ordinal .. ".json", serializeJson(makeStaticJson(cfg)))  
  return saveConfigJson(cfg.parentDeviceId, cfg.ordinal, serializeJson(cfg))
end
function saveConfigJson(parentDeviceId, ordinal, sJson)
  return writeFile(FILE_PATH .. "C_VirtualPanel_" .. parentDeviceId .. "_" .. ordinal .. ".json", sJson)
end
function getConfigFromChild(deviceId)
  local parentDeviceId = tonumber(luup.devices[tonumber(deviceId)].device_num_parent)
  local ordinal = luup.variable_get(VPNL_SERVICEID, "PanelOrdinal", tonumber(deviceId))
  debug('Getting configuration ' .. VPNL_SERVICEID .. " : " .. parentDeviceId .. ", " .. (ordinal or 'nil') .. " Child[" .. deviceId .. "]")
  local tbl = deserializeJson(getConfigJson(parentDeviceId, ordinal))
  tbl.deviceId = isNil(tbl.deviceId, deviceId)
  if(tonumber(deviceId) ~= tonumber(tbl.deviceId)) then
    debug("Device Id changed since panel was created...")
  end
  tbl.parentDeviceId = parentDeviceId
  tbl.ordinal = isNil(tbl.ordinal, ordinal)
  tbl.icon = isNil(tbl.icon, "scene_controller_default.png")
  if(tbl.watchVariables == nil) then
    tbl.watchVariables = {}
  end
  if(tbl.stateIcons == nil) then
    tbl.stateIcons = {}
  end
  if(tbl.elements == nil) then
    tbl.elements = {}
  end
  if(tbl.triggers == nil) then
    tbl.triggers = {}
  end
  if(tbl.states == nil) then
    tbl.states = {}
  end
  for ordinal, e in pairs(tbl.elements) do
    e.ordinal = ordinal
  end
  return tbl
end
function getConfigFromParent(parentDeviceId, ordinal)
  local tbl = deserializeJson(getConfigJson(parentDeviceId, ordinal))
  tbl.parentDeviceId = parentDeviceId
  tbl.ordinal = isNil(tbl.ordinal, ordinal)
  tbl.icon = isNil(tbl.icon, "scene_controller_default.png")
  return tbl
end
function removeElementByOrdinal(cfg, ordinal) 
  local newElems = {}
  debug("Removing Element Ordinal:" .. ordinal)
  for i, e in pairs(cfg.elements) do
    if(tonumber(e.ordinal) ~= tonumber(ordinal)) then
      table.insert(newElems, e)
    end
  end
  cfg.elements = newElems
  for ordinal, e in pairs(cfg.elements) do
    e.ordinal = ordinal
  end
  saveConfig(cfg)
  return cfg
end
function addElement(cfg, elem)
  if(cfg.elements == nil) then
    cfg.elements = {}
  end
  table.insert(cfg.elements, elem)
  for ordinal, e in pairs(cfg.elements) do
    e.ordinal = ordinal
  end
  saveConfig(cfg)
  return cfg
end
function saveElement(cfg, ordinal, elem)
  cfg.elements[ordinal] = elem
  for ordinal, e in pairs(cfg.elements) do
    e.ordinal = ordinal
  end
  saveConfig(cfg)
  return cfg
end
function saveDirtyPositions(cfg, pos)
  for i, p in pairs(pos) do
    local el = findElementByOrdinal(cfg.elements, p.ordinal)
    if(el ~= nil) then
      debug("Setting dirty positions for #" .. p.ordinal .. " " .. p.left .. "," .. p.top .. "," .. p.width .. "," .. p.height)
      if(el.position ~= nil) then
        el.position.left = p.left
        el.position.top = p.top
        el.position.width = p.width
        el.position.height = p.height
        if(el.controlType == "variable" or el.controlType == "label") then
          el.textStyle.width = p.width
        end
      end
    end
  end
  saveConfig(cfg)
  return cfg
end
function testLuaCode(deviceId, sCode)
  local ptr, rc = compileLua(deviceId, sCode)
  if(ptr ~= nil) then
    local fn = ptr()
    local ret, val = pcall(fn, {["panelDeviceId"] = deviceId, ["controlValue"] = ""})
    if(ret == false) then
      return ptr, {["Message"] = val, ["HasError"] = true}
    else
      return ptr, {["Message"] = "Successfully tested Lua code", ["HasError"] = false}
    end
  end
  return ptr, rc
end
function compileLua(deviceId, sCode)
  if(sCode == nil) then
    return nil, {["Message"] = "Lua code cannot be empty", ["HasError"] = true}
  end
  --luup.log("Compiling Lua:" .. sCode)
  local ptr, message = loadstring("return function(context) " .. sCode .. " end")
  local hasError = false
  if(ptr == nil) then
    hasError = true
  else
    message = "Successfully compiled Lua code"
    hasError = false
  end
  return ptr, {["Message"] = message, ["HasError"] = hasError }
end
function moveElement(cfg, oldOrdinal, newOrdinal)
  local elems = cfg.elements
  local el = cfg.elements[oldOrdinal]
  if(oldOrdinal < newOrdinal) then
    for ordinal, e in pairs(elems) do
      if(ordinal == oldOrdinal) then
        e.ordinal = newOrdinal
      elseif(ordinal <= newOrdinal and ordinal > oldOrdinal) then
        e.ordinal = ordinal - 1
      else
        e.ordinal = ordinal
      end
    end
  elseif(oldOrdinal > newOrdinal) then
    for ordinal, e in pairs(elems) do
      if(ordinal == oldOrdinal) then
        e.ordinal = newOrdinal
      elseif(ordinal >= newOrdinal and ordinal < oldOrdinal) then
        e.ordinal = ordinal + 1
      else
        e.ordinal = ordinal
      end
    end
  end
  sortElements(cfg)
  saveConfig(cfg)
  return cfg
end
function sortElements(cfg)
  table.sort(cfg.elements, function(e1, e2) return tonumber(e1.ordinal) < tonumber(e2.ordinal) end)
end
function saveIcons(cfg, defaultIcon, stateIcons)
  cfg.icon = defaultIcon
  cfg.stateIcons = stateIcons
  saveConfig(cfg)
  return cfg
end
function makeDeviceXml(deviceId, pnl)
  local sXml =
"<?xml version=\"1.0\"?>\n" 
.. "<!--\n"
.. "Virtual Panel Plugin\n"
.. "Copyright (C) 2015 Robert Strouse\n"
.. "This program is free software; you can redistribute it and/or\n"
.. "modify it under the terms of the GNU General Public License\n"
.. "as published by the Free Software Foundation; either version 2\n"
.. "of the License, or (at your option) any later version.\n"
.. "\n"
.. "This program is distributed in the hope that it will be useful,\n"
.. "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
.. "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
.. "GNU General Public License for more details.\n"
.. "\n"
.. "You should have received a copy of the GNU General Public License\n"
.. "along with this program; if not, write to the Free Software\n"
.. "Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.\n"
.. "-->\n"
.. "<root xmlns=\"urn:schemas-upnp-org:device-1-0\">\n"
.. "\t<specVersion>\n"
.. "\t\t<major>1</major>\n"
.. "\t\t<minor>0</minor>\n"
.. "\t</specVersion>\n"
.. "\t<device>\n"
.. "\t\t<deviceType>urn:schemas-rstrouse-com:device:VirtualPanel" .. deviceId .. "_" .. pnl.ordinal .. ":1</deviceType>\n"
.. "\t\t<staticJson>D_VirtualPanel_" .. deviceId .. "_" .. pnl.ordinal .. ".json</staticJson>\n"
.. "\t\t<friendlyName>Virtual Panel</friendlyName>\n"
.. "\t\t<manufacturer/>\n"
.. "\t\t<manufacturerURL/>\n"
.. "\t\t<modelDescription>Virtual Panel</modelDescription>\n"
.. "\t\t<modelName/>\n"
.. "\t\t<modelNumber/>\n"
.. "\t\t<modelURL/>\n"
.. "\t\t<handleChildren>1</handleChildren>\n"
.. "\t\t<UDN>uuid:" .. pnl.uuid .. "</UDN>\n"
.. "\t\t<UPC/>\n"
.. "\t\t<serviceList>\n"
.. "\t\t\t<service>\n"
.. "\t\t\t\t<serviceType>urn:schemas-rstrouse-com:service:VirtualPanel:1</serviceType>\n"
.. "\t\t\t\t<serviceId>urn:rstrouse-com:serviceId:VirtualPanel1</serviceId>\n"
.. "\t\t\t\t<SCPDURL>S_VirtualPanel.xml</SCPDURL>\n"
.. "\t\t\t</service>\n"
.. "\t\t\t<service>\n"
.. "\t\t\t\t<serviceType>urn:schemas-micasaverde-com:service:HaDevice:1</serviceType>\n"
.. "\t\t\t\t<serviceId>urn:micasaverde-com:serviceId:HaDevice1</serviceId>\n"
.. "\t\t\t\t<controlURL>/upnp/control/HaDevice1</controlURL>\n"
.. "\t\t\t\t<eventSubURL>/upnp/event/HaDevice1</eventSubURL>\n"
.. "\t\t\t\t<SCPDURL>S_HaDevice1.xml</SCPDURL>\n"
.. "\t\t\t</service>\n"
.. "\t\t</serviceList>\n"
.. "\t\t<presentationURL/>\n"
.. "\t</device>\n"
.. "</root>"
  
  return sXml
end
function makeStaticJson(cfg)
  local sj = {
    ["flashicon"] = "icons/devices.swf",
    ["halloIconsDir"] = "pics/hallo",
    ["inScene"] = "0",
--    ["DisplayStatus"] = {},
    ["default_icon"] = cfg.icon,
    ["doc_url"] = {
      ["doc_language"] = 1,
      ["doc_manual"] = 1,
      ["doc_version"] = 1,
      ["doc_platform"] = 0,
      ["doc_page"] = "devices"
    },
    ["ToggleButton"] = 1,
    ["Tabs"] = {
        {
          ["Label"] = { ["lang_tag"] = "tabname_control", ["text"] = "Control" },
          ["Position"] = "0",
          ["TabType"] = "flash",
          ["top_navigation_tab"] = 1,
          ["SceneGroup"] = {
            { ["id"] = "1", ["top"] = "0", ["left"] = "0", ["x"] = "1", ["y"] = "1", ["isSingle"] = "1" },
--            { ["id"] = "2", ["top"] = "1", ["left"] = "0", ["x"] = "2", ["y"] = "3", ["isSingle"] = "1" }
          },
          ["ControlGroup"] = {
              { ["id"] = "1", ["scenegroup"] = 1 },
--              { ["id"] = "2", ["scenegroup"] = 2 }
          },
        },
        {
          ["Position"] = "1",
          ["TopNavigationTab"] = "1",
          ["Label"] = { ["lang_tag"] = "layout", ["text"] ="Layout" },
          ["TabType"] = "javascript",
          ["ScriptName"] = "J_VirtualPanel.js",
          ["Function"] = "virtualPanelBuilder.showLayoutTab"
        },
--        {
--          ["Position"] = "2",
--          ["TopNavigationTab"] = "1",
--          ["Label"] = { ["lang_tag"] = "define_events", ["text"] ="States" },
--          ["TabType"] = "javascript",
--          ["ScriptName"] = "J_VirtualPanel.js",
--          ["Function"] = "virtualPanelBuilder.showStatesTab"
--        },
--        {
--          ["Position"] = "3",
--          ["TopNavigationTab"] = "1",
--          ["Label"] = { ["lang_tag"] = "define_events", ["text"] ="Triggers" },
--          ["TabType"] = "javascript",
--          ["ScriptName"] = "J_VirtualPanel.js",
--          ["Function"] = "virtualPanelBuilder.showTriggersTab"
--        },
        {
          ["Position"] = "2",
          ["Label"] = { ["lang_tag"] = "advanced", ["text"] ="Advanced" },
          ["TabType"] = "javascript",
          ["ScriptName"] = "shared.js",
          ["Function"] = "advanced_device"
        },
        {
          ["Position"] = "3",
          ["Label"] = { ["lang_tag"] = "notifications", ["text"] ="Notifications" },
          ["TabType"] = "javascript",
          ["ScriptName"] = "shared.js",
          ["Function"] = "device_notifications"
        },
        {
          ["Position"] = "4",
          ["Label"] = { ["lang_tag"] = "ui7_device_scenes", ["text"] ="Scenes" },
          ["TabType"] = "javascript",
          ["ScriptName"] = "shared.js",
          ["Function"] = "device_scenes"
        }
    },
    ["DeviceType"] = "urn:schemas-rstrouse-com:device:VirtualPanel" .. cfg.parentDeviceId .. "_" .. cfg.ordinal .. ":1",
    ["device_type"] = "urn:schemas-rstrouse-com:device:VirtualPanel" .. cfg.parentDeviceId .. "_" .. cfg.ordinal .. ":1"
  }
  cfg.watchVariables = {}
  if(cfg.stateIcons ~= nil) then
    sj["state_icons"] = transformIconStates(cfg)
  end
  if(cfg.elements ~= nil) then
    sj.Tabs[1]["Control"] = transformControls(cfg)
    -- Create scene groups and control groups for buttons and vertical sliders
    -- since they are relatively positioned off the control group.  This will create
    -- an anchor for them as they are created.
    
    local sgButton = {}
    local sgSV = {}
    local ctlGroupId = 1
    local sceneGroupId = 1
    for i, ctl in pairs(cfg.elements) do
      local sg = nil
      if(ctl.controlType == "group_start") then
        ctlGroupId = ctlGroupId + 1
        if(ctl.isSceneGroup ~= nil and ctl.isSceneGroup == true) then
          sceneGroupId = sceneGroupId + 1
          table.insert(sj.Tabs[1].SceneGroup, {["id"] = tostring(sceneGroupId), ["top"] = "0", ["left"] = "0", ["x"] = "1", ["y"] = "1", ["isSingle"] = "1"})
          if(ctl.lineBreak ~= nil and ctl.lineBreak == true) then
            -- Add a control group that will only contain the spacers to make it wrap.
            table.insert(sj.Tabs[1].ControlGroup, {["id"] = tostring(ctlGroupId), ["scenegroup"] = sceneGroupId})
            sceneGroupId = sceneGroupId + 1
            -- Add the scene group that will contain the controls in the group.
            table.insert(sj.Tabs[1].SceneGroup, {["id"] = tostring(sceneGroupId), ["top"] = "0", ["left"] = "0", ["x"] = "1", ["y"] = "1", ["isSingle"] = "1"})
            ctlGroupId = ctlGroupId + 1
          end
        end
        -- Add the control group for the controls
        table.insert(sj.Tabs[1].ControlGroup, {["id"] = tostring(ctlGroupId), ["scenegroup"] = sceneGroupId})
      end
    end
  end
  return sj
end
function isTextStyleEmpty(style)
  if(style == nil) then
    return true
  end
  for i, s in pairs(style) do
    if(i == "format" and s ~= nil and string.len(s) > 0 and s ~= "{0}") then
      return false
    elseif(i == "italic" and s ~= nil and s == true) then
      return false
    elseif(i == "bold" and s ~= nil and s == true) then
      return false
    elseif(i == "underline" and s ~= nil and s == true) then
      return false
    elseif(i == "fontSize" and s ~= nil and string.len(s) > 0 and s ~= "1em") then
      return false
    elseif(i == "fontFamily" and s ~= nil and string.len(s) > 0 and s ~= "inherit") then
      return false
    elseif(i == "textAlign" and s ~= nil and string.len(s) > 0 and s ~= "inherit") then
      return false
    elseif(i == "width" and s ~= nil and string.len(s) > 0) then
      return false
    end
  end
  return true
end
function isMarginsEmpty(margins)
  if(margins == nil) then
    return true
  end
  for i, s in pairs(margins) do
    if(i == "left" and tonumber(s) ~= 0) then
      return false
    elseif(i == "top" and tonumber(s) ~= 0) then
      return false
    elseif(i == "right" and tonumber(s) ~= 0) then
      return false
    elseif(i == "bottom" and tonumber(s) ~= 0) then
      return false
    end
  end
  return true
end
function transformIconStates(cfg)
  local states = {}
  if(cfg.stateIcons ~= nil) then
    for i, icon in pairs(cfg.stateIcons) do
      local tIcon = {["img"] = icon.icon, ["conditions"] = {}}
      for c, cond in pairs(icon.conditions) do
        if(cond.deviceId == nil) then
          cond.deviceId = cfg.deviceId
        end
        local mappedServiceId = cond.serviceId
        if(cond.serviceId == "urn:upnp-org:serviceId:Dimming1") then
          mappedServiceId = VPNL_SERVICEID
        end
        if(cond.value ~= nil and tonumber(cond.value) ~= nil) then
          table.insert(tIcon.conditions, 
            {["service"] = mappedServiceId, ["variable"] = transformWatchVariable(cfg, cond), 
              ["operator"] = cond.operator, ["value"] = tonumber(cond.value)})
        else
          table.insert(tIcon.conditions, 
            {["service"] = mappedServiceId, ["variable"] = transformWatchVariable(cfg, cond), 
              ["operator"] = cond.operator, ["value"] = cond.value})
        end
      end
      table.insert(states, tIcon)
    end
  end
  return states
end
function transformLineWeight(sWeight)
  local lpx = 0
  if(sWeight == "extrafine") then
    lpx = 1
  elseif(sWeight == "fine") then
    lpx = 1.5
  elseif(sWeight == "medium") then
    lpx = 2
  elseif(sWeight == "thick") then
    lpx = 3
  elseif(sWeight == "extrathick") then
    lpx = 4
  else
    lpx = 1
  end
  return lpx
end
function transformCssMargins(style)
  local sStyle = ""
  for i, s in pairs(style) do
    if(i == "top" and tonumber(s) ~= 0) then
      sStyle = sStyle .. "margin-top:" .. s .. "px;"
    elseif(i == "left" and tonumber(s) ~= 0) then
      sStyle = sStyle .. "margin-left:" .. s .. "px;"
    elseif(i == "right" and tonumber(s) ~= 0) then
      sStyle = sStyle .. "margin-right:" .. s .. "px;"
    elseif(i == "bottom" and tonumber(s) ~= 0) then
      sStyle = sStyle .. "margin-bottom:" .. s .. "px;"
    end
  end
  return sStyle
end
function transformCssFontStyle(style)
  local sStyle = ""
  for i, s in pairs(style) do
    if(i == "fontFamily" and s ~= nil and string.len(s) > 0) then
      sStyle = sStyle .. "font-family:" .. s .. ";"
    elseif(i == "italic" and s ~= nil and s) then
      sStyle = sStyle .. "font-style:italic;"
    elseif(i == "bold" and s ~= nil and s) then
      sStyle = sStyle .. "font-weight:bold;"
    elseif(i == "underline" and s) then
      sStyle = sStyle .. "text-decoration:underline;"
    elseif(i == "width" and s ~= nil and tonumber(s) > 0) then
      sStyle = sStyle .. "width:" .. s .. "px;"
    elseif(i == "fontSize" and string.len(s) > 0) then
      sStyle = sStyle .. "font-size:" .. s .. ";"
    elseif(i == "color" and s ~= nil and string.len(s) > 0) then
      sStyle = sStyle .. "color:" .. s .. ";"
    elseif(i == "textAlign" and s ~= nil and string.len(s) > 0) then
      sStyle = sStyle .. "text-align:" .. s .. ";"
    elseif(i == "height" and s ~= nil and tonumber(s) > 0) then
      sStyle = sStyle .. "height:" .. s .. "px;"
    end
  end
  return sStyle
end
function transformControls(cfg)
  local c = {}
  local groupId = 1
  for i, ctl in pairs(cfg.elements) do
    if(ctl.position ~= nil) then
      debug("Transforming Control: " .. ctl.controlType .. " Hide on dashboard:" .. tostring(isNil(ctl.position.hideOnDashboard, false)))
    else
      debug("Transforming Control: " .. ctl.controlType .. " Hide on dashboard: N/A")
    end
    local t = {["ControlType"] = ctl.controlType }
    if(ctl.position == nil or isNil(ctl.position.hideOnDashboard, false) == false) then
      t["ControlGroup"] = tostring(groupId)
    end
    if(ctl.controlType ~= "line_break" and ctl.controlType ~= "spacer" and ctl.controlType ~= "group_start") then
      t["top"] = y2cg(tonumber(ctl.position.top or 0) or 0)
      t["left"] = x2cg(tonumber(ctl.position.left or 0) or 0)
      t["Label"] = {["lang_tag"] = "tag_vpnl_" .. i }
    end
    if(ctl.controlType == "group_start") then
      groupId = groupId + 1
      if(ctl.lineBreak ~= nil and ctl.lineBreak == true) then
        if(ctl.isSceneGroup ~= nil and ctl.isSceneGroup == true) then
          t["HorizontalMultiplier"] = 20
          if(isNil(ctl.hideOnDashboard, false) == false) then
            t["ControlGroup"] = tostring(groupId)
          end
          t["ControlType"] = "spacer"
          table.insert(c, t)
          table.insert(c, t)
          groupId = groupId + 1
        else
          if(isNil(ctl.hideOnDashboard, false) == false) then
            t["ControlGroup"] = tostring(groupId)
          end
          t["ControlType"]= "line_break"
          table.insert(c, t)
        end
      end
    elseif(ctl.controlType == "label") then
      if(ctl.textStyle == nil) then
        ctl.textStyle = {}
      end
      if(ctl.margins == nil) then
        ctl.margins = {}
      end
      t.Label["text"] = "<span style=\'display:inline-block;vertical-align:middle;line-height:1;white-space:pre;" .. transformCssFontStyle(ctl.textStyle) .. transformCssMargins(ctl.margins) .. "\'>" .. ctl.text .. "</span>"
      t["Display"] = { ["Top"] = ctl.position.top, ["Left"] = ctl.position.left, ["Width"] = ctl.position.width, ["Height"] = ctl.position.height }
    elseif(ctl.controlType == "spacer") then
      t["HorizontalMultiplier"] = ctl.position.width
    elseif(ctl.controlType == "horizontal_rule") then
      t["ControlType"] = "label"
      t.Label["text"] = "<hr style=\'border:" .. ctl.lineStyle.style .. " " .. transformLineWeight(ctl.lineStyle.weight) .. "px " .. ctl.lineStyle.color 
      .. ";margin-left:0px;margin-right:0px;margin-top:" .. ctl.margins.top .. "px;margin-bottom:" .. ctl.margins.bottom .. "px;width:" .. ctl.position.width .. "px\'/>"
      t["Display"] = { ["Top"] = ctl.position.top, ["Left"] = ctl.position.left, ["Width"] = ctl.position.width, ["Height"] = ctl.position.height }
    elseif(ctl.controlType == "variable") then
      t.Label["text"] = ctl.text
      t["Display"] = { ["Top"] = ctl.position.top, ["Left"] = ctl.position.left, ["Width"] = ctl.position.width, ["Height"] = ctl.position.height,
        ["Service"] = ctl.display.serviceId, 
        ["Variable"] = transformWatchVariable(cfg, ctl.display, ctl.textStyle, ctl.margins) }
      
    elseif(ctl.controlType == "color_indicator") then
      t.Label["text"] = ctl.text
      t["ControlType"] = "variable"
      t["Display"] = { ["Top"] = ctl.position.top, ["Left"] = ctl.position.left, ["Width"] = ctl.position.width, ["Height"] = ctl.position.height,
        ["Service"] = ctl.display.serviceId, 
        ["Variable"] = transformColorIndicatorVariable(cfg, ctl.display, {["width"] = ctl.position.width, ["height"] = ctl.position.height}, ctl.margins, ctl.defaultColor, ctl.expressions) }
    elseif(ctl.controlType == "button") then
      --t.ControlGroup = tostring(tonumber(i) + 2)
      t.ID = "new_control_value" .. i
      if(ctl.textStyle == nil) then
        ctl.textStyle = {}
      end
      if(tonumber(ctl.position.width) > 0) then
        t["HorizontalMultiplier"] = tonumber(ctl.position.width)/65
      end
      --t.Label["text"] = "<div style=&quot;" .. transformCssFontStyle(ctl.textStyle) .. "&quot;;>" .. ctl.text .. "</div>"
      t.Label["text"] = ctl.text
      t["Display"] = { ["Top"] = ctl.position.top, ["Left"] = ctl.position.left, ["Width"] = ctl.position.width, ["Height"] = ctl.position.height, ["ID"] = "new_control_value" .. i }
      t.top = 0
      t.left = 0
      --t.Display.top = y2cg(tonumber(ctl.position.top))
      --t.Display.left = x2cg(tonumber(ctl.position.left))
      t.ControlCode = "btnclick_" .. i
      if(ctl.buttonStyle == "persistent") then
        t.Display.Service = ctl.display.serviceId
        if(ctl.display.operator ~= nil and ctl.display.operator ~= "==") then
          t.Display.Variable = transformStateExpressionVariable(cfg, ctl.display, 0, 
            {{["value"] = ctl.display.value, ["maxValue"] = (ctl.display.maxValue or ""), ["operator"] = ctl.display.operator, ["state"] = "1"}} )
          t.Display.Value = "1"
        else
          t.Display.Variable = transformWatchVariable(cfg, ctl.display)
          t.Display.Value = ctl.display.value
        end
        t.Display.ID = "new_control_value" .. i
        t.Command = transformAction(cfg, ctl.command, "", i)
      elseif(ctl.buttonStyle == "toggle") then
        t.Display.Service = VPNL_SERVICEID
        t.Display.Variable = "toggleButton" .. i
        t.Display.ID = "new_control_value" .. i
        t.Display.Value = "1"
        t.Command = transformToggleAction(cfg, ctl.onState, ctl.offState, i)
        -- Add in a parameter letting us know to persist the state.
        table.insert(t.Command.Parameters, {["Name"] = "pnlcmd_state_persist", ["Value"] = 1})
      elseif(ctl.buttonStyle == "persistenttoggle") then
        t.Display.Service = ctl.display.serviceId
        if(ctl.display.operator ~= nil and ctl.display.operator ~= "==") then
          t.Display.Variable = transformStateExpressionVariable(cfg, ctl.display, 0, 
            {{["value"] = ctl.display.value, ["operator"] = ctl.display.operator, ["maxValue"] = (ctl.display.maxValue or ""), ["state"] = "1"}} )
          t.Display.Value = "1"
        else
          t.Display.Variable = transformWatchVariable(cfg, ctl.display)
          t.Display.Value = ctl.display.value
        end
        t.Display.ID = "new_control_value" .. i
        t.Command = transformToggleAction(cfg, ctl.onState, ctl.offState, i)
        -- Add in a parameter letting us know to not persist the state and let the control variable handle it.
        table.insert(t.Command.Parameters, {["Name"] = "pnlcmd_state_persist", ["Value"] = 0})
        -- For now we will use the transformed variable.  This is the current button state.
        -- If this proves to be too dynamic we will pin it down to its source (the originator of the state).
        table.insert(t.Command.Parameters, {["Name"] = "pnlcmd_state_deviceId", ["Value"] = ctl.display.deviceId})
        table.insert(t.Command.Parameters, {["Name"] = "pnlcmd_state_serviceId", ["Value"] = t.Display.Service})
        table.insert(t.Command.Parameters, {["Name"] = "pnlcmd_state_variable", ["Value"] = t.Display.Variable})
        -- For not use equals as the standard operator.  Later we can do some gymnastics with the value to get
        -- the current state.
        table.insert(t.Command.Parameters, {["Name"] = "pnlcmd_state_operator", ["Value"] = "=="})
        table.insert(t.Command.Parameters, {["Name"] = "pnlcmd_state_value", ["Value"] = t.Display.Value})
      else
        t.Command = transformAction(cfg, ctl.command, "", i)
      end
    elseif(ctl.controlType == "multi_state_button") then
        t.ID = "new_control_value" .. i
        if(ctl.onState.textStyle == nil) then
          ctl.onState.textStyle = {}
        end
        if(ctl.offState.textStyle == nil) then
          ctl.offState.textStyle = {}
        end
        -- The off state of the button is first in the static json.
        t["ControlCode"] = "ctlcode_vpnl_" .. i
        t["Display"] = { ["Top"] = ctl.position.top, 
          ["Left"] = ctl.position.left, 
          ["Width"] = ctl.position.width, 
          ["Height"] = ctl.position.height, 
          ["ID"] = "new_control_value" .. i }
        local onDisplay = {["Service"] = ctl.onState.display.serviceId}
        local offDisplay = {["Service"] = ctl.offState.display.serviceId}
        if((ctl.onState.display.operator or "==") ~= "==" or (ctl.offState.display.operator or "==" ~= "==")) then
          onDisplay.Variable = transformStateExpressionVariable(cfg, ctl.onState.display, 0, 
              {{["value"] = ctl.onState.display.value, 
                  ["maxValue"] = (ctl.onState.display.maxValue or ""), 
                  ["operator"] = (ctl.onState.display.operator or "=="), 
                  ["state"] = "1"},
                {["value"] = ctl.offState.display.value,
                  ["maxValue"] = (ctl.offState.display.maxValue or ""),
                  ["operator"] = (ctl.offState.display.operator or "=="),
                  ["state"] = "0"}})
          onDisplay.Value = "1"
          offDisplay.Variable = onDisplay.Variable
          offDisplay.Value = "0"
        else
          onDisplay.Variable = transformWatchVariable(cfg, ctl.onState.display)
          onDisplay.Value = ctl.onState.display.value
          offDisplay.Variable = transformWatchVariable(cfg, ctl.offState.display)
          offDisplay.Value = ctl.offState.display.value
        end
        t.states = {
          { ["Label"] = {["lang_tag"] = "tag_vpnl_on_" .. i, 
            ["text"] = "<div style=\'display:inline-block;vertical-align:top;\' title=\'" .. ctl.onState.text .. "\'><span style=\'" .. transformCssFontStyle(ctl.onState.textStyle) .. "\'>" .. ctl.onState.text .. "</span></div>"},
            ["ControlGroup"] = "1",
            ["Display"] = onDisplay, 
            ["Command"] = transformAction(cfg, ctl.onState.command, "", i),
            ["ControlCode"] = "ctlcode_vpnl_on_" .. i },
          { ["Label"] = {["lang_tag"] = "tag_vpnl_off_" .. i, 
              ["text"] = "<div title=\'" .. ctl.offState.text .. "\'><span style=\'display:inline-block;vertical-align:top;" .. transformCssFontStyle(ctl.offState.textStyle) .. "\'>" .. ctl.offState.text .. "</span></div>"},
            ["ControlGroup"] = "1",
            ["Display"] = offDisplay,
            ["Command"] = transformAction(cfg, ctl.offState.command, "", i),
            ["ControlCode"] = "ctlcode_vpnl_off_" .. i }            
          }
    elseif(ctl.controlType == "slider") then
        t.Label["text"] = ctl.values.max.text
        t.Style = "numeric"
        t.ID = "new_control_value" .. i
        t.LabelMin = { ["lang_tag"] = "slider_label_min_" .. i, ["text"] = ctl.values.min.text }
        t.LabelMax = { ["lang_tag"] = "slider_label_max_" .. i, ["text"] = ctl.values.max.text }
        t.LabelSymbol = { ["lang_tag"] = "slider_label_symbol_" .. i, ["text"] = (ctl.values.suffix or "%") }
        t.Display = { ["Service"] = ctl.display.serviceId, 
          ["Variable"] = transformWatchVariable(cfg, ctl.display),
          ["MinValue"] = (tonumber(ctl.values.min.value) or ctl.values.min.value), 
          ["MaxValue"] = (tonumber(ctl.values.max.value) or ctl.values.max.value),
          ["Top"] = ctl.position.top, ["Left"] = ctl.position.left, ["Width"] = ctl.position.width, ["Height"] = ctl.position.height, ["ID"] = "new_control_value" .. i}
        t.Command = transformAction(cfg, ctl.command, "", i)
        t.Command.HumanFriendlyText = { ["lang_tag"] = "ui7_cmd_dimmable_light_level", ["text"] = "Level" }
        t.Command.Suffix = (ctl.values.suffix or "%")
        t.Command.Validation = {["Type"] = "numeric", 
          ["Min"] = (tonumber(ctl.values.min.value) or ctl.values.min.value), 
          ["Max"] = (tonumber(ctl.values.max.value) or ctl.values.max.value)}
        t.ControlCode = "customslider" .. i
        t["HorizontalMultiplier"] = ctl.position.width/170
        
    elseif(ctl.controlType == "slider_vertical") then
        --t.ControlGroup = tostring(tonumber(i) + 2)
        t.Style = "numeric"
        t.ID = "new_control_value" .. i
        t.LabelSymbol = { ["lang_tag"] = "control_label" .. i .. "_symbol", ["text"] = (ctl.values.suffix or "&deg;") }
        t.LabelMin = { ["lang_tag"] = "slider_label_min_" .. i, ["text"] = ctl.values.min.text }
        t.LabelMax = { ["lang_tag"] = "slider_label_max_" .. i, ["text"] = ctl.values.max.text }
        t.Display = { ["Service"] = ctl.display.serviceId, ["Variable"] = transformWatchVariable(cfg, ctl.display),
          ["MinValue"] = ctl.values.min.value, ["MaxValue"] = ctl.values.max.value,
          ["Top"] = ctl.position.top, ["Left"] = ctl.position.left, ["Width"] = ctl.position.width, ["Height"] = ctl.position.height, ["ID"] = "new_control_value" .. i}
        t.Command = transformAction(cfg, ctl.command, "", i)
        t.Command.HumanFriendlyText = { ["lang_tag"] = "control_" .. i .. "_level", ["text"] = "Level" }
        t.Command.Suffix = (ctl.values.suffix or "&deg;")
        --t.Command.Validation = {["Type"] = "numeric", ["Min"] = ctl.values.min.value, ["Max"] = ctl.values.max.value}
        t.ControlCode = "customslider" .. i
    elseif(ctl.controlType == "spinner_horizontal") then
        t.ID = "new_control_value" .. i
        t.Style = "numeric"
        t.textAlign = "left"
        t.LabelUp = {["text"] = ctl.values.max.text}
        t.LabelDown = {["text"] = ctl.values.min.text}
        t.LabelSymbol = {["text"] = ctl.values.suffix}
        t.Display = { ["Service"] = ctl.display.serviceId, ["Variable"] = transformWatchVariable(cfg, ctl.display),
          ["MinValue"] = ctl.values.min.value, ["MaxValue"] = ctl.values.max.value,
          ["Step"] = ctl.values.step,
          ["InvalidInputMessage"] = {["text"] = "Enter a value between " 
            .. ctl.values.min.value .. ctl.values.suffix .. " and " .. ctl.values.max.value .. ctl.values.suffix},
          ["Top"] = ctl.position.top, ["Left"] = ctl.position.left, ["Width"] = ctl.position.width, ["Height"] = ctl.position.height, ["ID"] = "new_control_value" .. i}
        if(ctl.values.allowEdit) then
          t.Display["Editable"] = "1"
        else
          t.Display["Editable"] = "0"
        end
        t.Command = transformActionArgument(cfg, ctl.command, i)
        t.Command.HumanFriendlyText = { ["lang_tag"] = "control_" .. i .. "_level", ["text"] = "Level" }
        t.Command.Suffix = (ctl.values.suffix or "&deg;")
        t.ControlCode = "customspinner" .. i
          
    end
    if(ctl.controlType ~= "group_start") then
      table.insert(c, t)
    end
  end
  return c
end
function transformToggleAction(cfg, aOn, aOff, ordinal)
  local act = {}
  act.Service = VPNL_SERVICEID
  act.Action = "ActionToggleButton"
  -- Get the state actions and just copy their params to the
  -- toggle action call.
  local cmdOn = transformAction(cfg, aOn.command, "on_", ordinal)
  local cmdOff = transformAction(cfg, aOff.command, "off_", ordinal)
  act.Parameters = {
    {["Name"] = "controlOrdinal", ["Value"] = ordinal},
    {["Name"] = "cmdType_on_", ["Value"] = aOn.command.cmdType},
    {["Name"] = "cmdType_off_", ["Value"] = aOff.command.cmdType}}
  for i, p in pairs(cmdOn.Parameters) do
    table.insert(act.Parameters, p)
  end
  for i, p in pairs(cmdOff.Parameters) do
    table.insert(act.Parameters, p)
  end
  return act
end
function transformActionArgument(cfg, a, ordinal)
  local act = {["Service"] = VPNL_SERVICEID, ["ActionArgumentName"] = "ctlVal_" .. ordinal}
  if(a.cmdType == "ExecScene") then
    act.Action = "ActionExecuteScene"
  elseif(a.cmdType == "ExecLua") then
    act.Action = "ActionExecuteLua"
  else
    act.Action = "Action" .. a.cmdType
  end
  return act
end
function transformAction(cfg, a, prefix, ordinal)
  local act = {}
  if(a.cmdType == "SetVariable") then
    act.Service = VPNL_SERVICEID
    act.Action = "ActionSetVariable"
    act.Parameters = {
      {["Name"] = prefix .. "pnlcmd_serviceId", ["Value"] = a.variable.serviceId},
      {["Name"] = prefix .. "pnlcmd_variable", ["Value"] = a.variable.variable},
      {["Name"] = prefix .. "pnlcmd_deviceId", ["Value"] = a.variable.deviceId}}
    if(a.variable.useControlValue or false) then
      table.insert(act.Parameters, {["Name"] = prefix .. "pnlcmd_value", ["ID"] = "new_control_value" .. ordinal})
    else
      table.insert(act.Parameters, {["Name"] = prefix .. "pnlcmd_value", ["Value"] = a.variable.value})
    end
  elseif(a.cmdType == "CallAction") then
    act.Service = VPNL_SERVICEID
    act.Action = "ActionCallAction"
    act.Parameters = {
      {["Name"] = prefix .. "pnlcmd_serviceId", ["Value"] = a.action.serviceId},
      {["Name"] = prefix .. "pnlcmd_action", ["Value"] = a.action.action},
      {["Name"] = prefix .. "pnlcmd_deviceId", ["Value"] = a.action.deviceId}}
    if(a.action.parameters ~= nil) then
      for i, p in pairs(a.action.parameters) do
        if(p.name ~= nil and string.len(p.name) > 0) then
          if(p.isControlValue) then
            table.insert(act.Parameters, {["Name"] = prefix .. p.name, ["ID"] = "new_control_value" .. ordinal})
          else
            table.insert(act.Parameters, {["Name"] = prefix .. p.name, ["Value"] = p.value})
          end
        end
      end
    end
  elseif(a.cmdType == "ExecScene") then
    act.Service = VPNL_SERVICEID
    act.Action = "ActionExecuteScene"
    act.Parameters = {{["Name"] = prefix .. "pnlcmd_sceneId", ["Value"] = a.scene.sceneId}}
  elseif(a.cmdType == "HttpGET" or a.cmdType == "HttpPOST") then
    act.Service = VPNL_SERVICEID
    act.Action = "Action" .. a.cmdType
    act.Parameters = {
      {["Name"] = prefix .. "pnlcmd_url", ["Value"] = a.http.url},
      {["Name"] = prefix .. "pnlcmd_username", ["Value"] = a.http.username},
      {["Name"] = prefix .. "pnlcmd_password", ["Value"] = a.http.password}}
  elseif(a.cmdType == "ExecLua") then
      act.Service = VPNL_SERVICEID
      act.Action = "ActionExecuteLua"
      act.Parameters = {
        {["Name"] = prefix .. "pnlcmd_binding", ["Value"] = a.execLua.callName},
        {["Name"] = prefix .. "pnlcmd_ordinal", ["Value"] = ordinal},
        {["Name"] = prefix .. "new_control_value", ["ID"] = "new_control_value" .. ordinal}}
  else
    act.Service = VPNL_SERVICEID
    act.Action = "Action" .. a.cmdType
    act.Parameters = {
      {["Name"] = prefix .. "controlOrdinal", ["Value"] = ordinal},
      {["Name"] = prefix .. "new_control_value", ["ID"] = "new_control_value" .. ordinal}}
  end
  return act
end
function transformStateExpressionVariable(cfg, v, defaultVal, expressions)
  local targetServiceId = v.serviceId
  local mappedServiceId = v.serviceId
  local wv = cfg.watchVariables[v.serviceId .. "," .. v.variable .. "," .. v.deviceId]
  if(wv == nil) then
    wv = {["source"] = {["serviceId"] = targetServiceId, ["variable"] = v.variable, ["deviceId"] = v.deviceId}, ["mapped"] = {}}
    cfg.watchVariables[v.serviceId .. "," .. v.variable .. "," .. v.deviceId] = wv
  end
  local mv = {
    ["type"] = "statemap",
    ["serviceId"] = mappedServiceId, 
    ["variable"] = v.variable .. v.deviceId .. "_State" .. (table.maxn(wv.mapped) or "0"), 
    ["deviceId"] = cfg.deviceId, 
    ["defaultValue"] = defaultVal,
    ["style"] = cfg.style,
    ["margins"] = cfg.margins,
    ["expressions"] = expressions }
  table.insert(wv.mapped, mv)
  return mv.variable
end
function transformColorIndicatorVariable(cfg, v, style, margins, defaultColor, expressions)
  local targetServiceId = v.serviceId
  local mappedServiceId = v.serviceId
  -- HACK to get around the bug in interface.js where the command for sliders is using the display variable incorrectly.
  --if(v.serviceId == "urn:upnp-org:serviceId:Dimming1") then
  --  mappedServiceId = VPNL_SERVICEID
  --end
  local wv = cfg.watchVariables[v.serviceId .. "," .. v.variable .. "," .. v.deviceId]
  if(wv == nil) then
    wv = {["source"] = {["serviceId"] = targetServiceId, ["variable"] = v.variable, ["deviceId"] = v.deviceId}, ["mapped"] = {}}
    cfg.watchVariables[v.serviceId .. "," .. v.variable .. "," .. v.deviceId] = wv
  end
  local mv = {
    ["type"] = "colormap",
    ["serviceId"] = mappedServiceId, 
    ["variable"] = v.variable .. v.deviceId .. "_Indicator" .. (table.maxn(wv.mapped) or "0"), 
    ["deviceId"] = cfg.deviceId, 
    ["defaultColor"] = defaultColor,
    ["style"] = style,
    ["margins"] = margins,
    ["expressions"] = expressions }

  table.insert(wv.mapped, mv)
  return mv.variable
end
function transformWatchVariable(cfg, v, textStyle, margins)
  local targetServiceId = v.serviceId
  local mappedServiceId = v.serviceId
  -- HACK to get around the bug in interface.js where the command for sliders is using the display variable incorrectly.
  --if(v.serviceId == "urn:upnp-org:serviceId:Dimming1") then
  --  mappedServiceId = VPNL_SERVICEID
  --end
  if(tonumber(cfg.deviceId) ~= tonumber(v.deviceId)) then
    local mv = {["serviceId"] = mappedServiceId, ["variable"] = v.variable .. v.deviceId, ["deviceId"] = cfg.deviceId}
    local wv = cfg.watchVariables[v.serviceId .. "," .. v.variable .. "," .. v.deviceId]
    -- Add a new watch variable.  We are getting this variable from another device.
    if(wv == nil) then
      wv = {["source"] = {["serviceId"] = targetServiceId, ["variable"] = v.variable, ["deviceId"] = v.deviceId}, ["mapped"] = {}}
      cfg.watchVariables[v.serviceId .. "," .. v.variable .. "," .. v.deviceId] = wv
      -- Always add the variable un-decorated.
      table.insert(wv.mapped, mv)
    end
    wv = cfg.watchVariables[v.serviceId .. "," .. v.variable .. "," .. v.deviceId]
    if(isTextStyleEmpty(textStyle) == false or isMarginsEmpty(margins) == false) then
      mv = {["type"] = "state",
        ["serviceId"] = mappedServiceId, 
            ["variable"] = v.variable .. v.deviceId .. "_Formatted_" .. table.maxn(wv.mapped), 
            ["deviceId"] = cfg.deviceId, 
            ["textStyle"] = textStyle,
            ["margins"] = margins}
      table.insert(wv.mapped, mv)
    end
    return mv.variable
  elseif(isTextStyleEmpty(textStyle) == false or isMarginsEmpty(margins) == false) then
    -- We have to map locals but only if they are formatted.
    local wv = cfg.watchVariables[v.serviceId .. "," .. v.variable .. "," .. v.deviceId]
    if(wv == nil) then
      wv = {["source"] = {["serviceId"] = targetServiceId, ["variable"] = v.variable, ["deviceId"] = v.deviceId}, ["mapped"] = {}}
      cfg.watchVariables[v.serviceId .. "," .. v.variable .. "," .. v.deviceId] = wv
    end
    local mv = {
          ["type"] = "formatted",
          ["serviceId"] = mappedServiceId, 
          ["variable"] = v.variable .. v.deviceId .. "_Formatted_" .. (table.maxn(wv.mapped) + 1), 
          ["deviceId"] = cfg.deviceId, 
          ["textStyle"] = textStyle,
          ["margins"] = margins}
    table.insert(wv.mapped, mv)
    return mv.variable
  end
  return v.variable
end

----------------------------------------------------------------------
-- Vera Setup Functions
----------------------------------------------------------------------
function getAllDevices(id)
  local rooms = {}
  rooms[0] = {["roomId"] = 0, ["roomName"] = "No Room", ["devices"] = {}, ["scenes"] = {}}
  debug("Enumerating all devices")
  for roomId, v in pairs(luup.rooms) do
    rooms[tonumber(roomId)] = {["roomId"] = tonumber(roomId), ["roomName"] = (v or "No Name"), ["devices"] = {}, ["scenes"] = {}}
  end
  for deviceId, v in pairs(luup.devices) do
    local devs = rooms[tonumber(v.room_num)]
    if(devs ~= nil) then
      table.insert(devs.devices, 
        {["deviceId"] = tonumber(deviceId), ["roomId"] = tonumber(v.room_num), ["deviceName"] = v.description, ["parentId"] = tonumber(v.device_num_parent), ["deviceType"] = v.device_type})
    end
  end
  for sceneId, v in pairs(luup.scenes) do
    local room = rooms[tonumber(v.room_num or "0")]
    if(room ~= nil) then
      debug('Adding scene#' .. (sceneId or "nil") .. " to room #" .. (room.roomId or "nil"))
      table.insert(room.scenes,
        {["sceneId"] = tonumber(sceneId), ["roomId"] = tonumber(v.room_num), ["sceneName"] = v.description, ["hidden"] = v.hidden})
    end
  end
  return rooms
end
----------------------------------------------------------------------
-- Dashboard Positioning Functions
----------------------------------------------------------------------
function y2cg(px)
  return ((px or 0) / 100) - 1.2
end
function cg2y(cg)
  return (cg * 100) + 120
end
function x2cg(px)
  return ((px or 0) / 350) - .04
end
function cg2x(cg)
  return (cg * 350) + 14
end
function y2sg(px)
  return (px / 212) - 2.0
end
function x2sg(px)
  return (px / 350) - .04
end
----------------------------------------------------------------------
-- Panel Element Actions
----------------------------------------------------------------------
function getActionCommandConfig(lul_device, lul_settings)
    -- This is a dumbass control that only sends its value.
    local act = nil
    local v = nil
    local ord = nil
    local cfg = getConfigFromChild(tonumber(lul_device))
    for key, val in pairs(lul_settings) do
      if(string.find(key, "ctlVal_") == 1) then
        ord = tonumber(string.match(key, "%d+"))
        local elem = findElementByOrdinal(cfg.elements, tonumber(ord))
        if(elem ~= nil) then
          act = elem.command
        end
        v = val
        break
      end
    end
    return act, v, ord
end
function buildCommandParameters(lul_settings, prefix)
  local params = {}
  local cmdType = ""
  for name, p in pairs(lul_settings) do
    if(string.find(name, "cmdType_" .. prefix) == 1) then
      cmdType = p
    elseif(string.find(name, prefix) == 1) then
      debug("Adding param " .. string.sub(name, string.len(prefix) + 1, -1) .. " = " .. (p or "nil"))
      params[string.sub(name, string.len(prefix) + 1, -1)] = p 
    end
  end
  return cmdType, params
end
function actionToggleButton(lul_device, lul_settings)
  debug("Calling Action:ToggleButton ")
  local args = {}
  local params = {}
  local cmdType = ""
  local currentState = 0
  if(tonumber(lul_settings.pnlcmd_state_persist) == 1) then
    currentState = tonumber(isNil(luup.variable_get(VPNL_SERVICEID, "toggleButton" .. lul_settings.controlOrdinal, tonumber(lul_device)), 0))
    if(currentState == 0) then
      -- We are turning on
      luup.variable_set(VPNL_SERVICEID, "toggleButton" .. lul_settings.controlOrdinal, 1, tonumber(lul_device))
    else
      luup.variable_set(VPNL_SERVICEID, "toggleButton" .. lul_settings.controlOrdinal, 0, tonumber(lul_device))
    end
  else
    local currentValue = isNil(luup.variable_get(lul_settings.pnlcmd_state_serviceId, lul_settings.pnlcmd_state_variable, tonumber(lul_device)), 0)
    -- Now we need to determine the equality.
    if(tonumber(lul_settings.pnlcmd_state_variable) ~= nil) then
      if(tonumber(currentValue) == tonumber(lul_settings.pnlcmd_state_value)) then
        currentState = 1
      end
    elseif(currentValue == lul_settings.pnlcmd_state_value) then
      currentState = 1
    end
  end
  debug("Toggle Button Current State:" .. currentState .. " DeviceId: " .. lul_device)
  if(currentState == 0) then
    -- We are turning on
    cmdType, params = buildCommandParameters(lul_settings, "on_")
  else
    cmdType, params = buildCommandParameters(lul_settings, "off_")
  end
  if(cmdType == "SetVariable") then
    actionSetVariable(lul_device, params)
  elseif(cmdType == "CallAction") then
    actionCallAction(lul_device, params)
  elseif(cmdType == "ExecScene") then
    actionExecuteScene(lul_device, params)
  elseif(cmdType == "HttpGET") then
    actionHttpGET(lul_device, params)
  elseif(cmdType == "HttpPOST") then
    actionHttpPOST(lul_device, params)
  elseif(cmdType == "ExecLua") then
    actionExecuteLua(lul_device, params)
  end
end

function actionSetVariable(lul_device, lul_settings)
  if(lul_settings.pnlcmd_deviceId ~= nil) then
    maybeSetVariable(lul_settings.pnlcmd_serviceId, lul_settings.pnlcmd_variable, lul_settings.pnlcmd_value, tonumber(lul_settings.pnlcmd_deviceId))
    debug("Calling Action:SetVariable " .. lul_settings.pnlcmd_serviceId .. ", " .. lul_settings.pnlcmd_variable .. ", " .. lul_settings.pnlcmd_deviceId)
  else
    -- We are in the realm of the dumbasses that changed the command structure to send the value only (horizontal_spinners)
    local cmd, val = getActionCommandConfig(lul_device, lul_settings)
    if(cmd ~= nil) then
      local act = cmd.variable
      local v = act.value
      if(act.useControlValue or false) then
        v = val
      end
      debug("Calling Action:SetVariable " .. act.serviceId .. ", " .. act.variable .. ", " .. act.deviceId .. " [" .. (v or "nil") .. "]")
      maybeSetVariable(act.serviceId, act.variable, v, tonumber(act.deviceId))
    end
  end
end
function actionCallAction(lul_device, lul_settings)
  local args = {}
  if(lul_settings.pnlcmd_deviceId ~= nil) then 
    for i, arg in pairs(lul_settings) do
      if(string.find(i, "pnlcmd_") ~= 1 and string.find(i, "DeviceNum") ~= 1 
        and string.find(i, "action") ~=1 and string.find(i, "serviceId") ~= 1) then
        debug("Preparing argument " .. i .. "=" .. arg)
        args[i] = arg
      end
    end
  else
    -- We are in the realm of the dumbasses that changed the command structure to send the value only (horizontal_spinners)
    local cmd, val = getActionCommandConfig(lul_device, lul_settings)
    if(cmd ~= nil) then
      local act = cmd.action
      lul_settings.pnlcmd_serviceId = act.serviceId
      lul_settings.pnlcmd_deviceId = act.deviceId
      lul_settings.pnlcmd_action = act.action
      for i, arg in pairs(act.parameters) do
        local v = arg.value
        if(arg.isControlValue or false) then
          v = val
        end
        debug("Preparing argument " .. arg.name .. " = " .. v)
        args[arg.name] = v
      end
    end
  end
  debug("Calling Action:CallAction " .. lul_settings.pnlcmd_serviceId .. ", " 
    .. lul_settings.pnlcmd_action .. ", " .. lul_settings.pnlcmd_deviceId .. " " .. serializeJson(args))
  luup.call_action(lul_settings.pnlcmd_serviceId, lul_settings.pnlcmd_action, args, tonumber(lul_settings.pnlcmd_deviceId))
end
function actionExecuteScene(lul_device, lul_settings)
  if(lul_settings.pnlcmd_sceneId == nil) then
    local cmd, val = getActionCommandConfig(lul_device, lul_settings)
    if(cmd ~= nil) then
      lul_settings.pnlcmd_sceneId = cmd.scene.sceneId
    end
  end
  debug("Calling Action:ExecuteScene " .. lul_settings.pnlcmd_sceneId)
  luup.call_action("urn:micasaverde-com:serviceId:HomeAutomationGateway1", "RunScene", {["SceneNum"] = lul_settings.pnlcmd_sceneId}, 0)  
end
function actionHttpGET(lul_device, lul_settings)
  local url = ""
  if(lul_settings.pnlcmd_url ~= nil) then
    for i, arg in pairs(lul_settings) do
      if(string.find(i, "pnlcmd_") ~= 1 and string.find(i, "DeviceNum") ~= 1 
          and string.find(i, "action") ~=1 and string.find(i, "serviceId") ~= 1) then
        if(string.len(url) == 0) then
          url = "?" .. i .. "=" .. arg
        else
          url = url .. "&" .. i .. "=" .. arg
        end
      end
    end
  else
    local cmd, val = getActionCommandConfig(lul_device, lul_settings)
    if(cmd ~= nil) then
      lul_settings.pnlcmd_url = cmd.http.url
      lul_settings.pnlcmd_username = cmd.http.username
      lul_settings.pnlcmd_password = cmd.http.password
    end
  end
  debug("Calling Action:HttpGET " .. url)
  local httpCode, xmlData = luup.inet.wget(lul_settings.pnlcmd_url .. url, 5, lul_settings.pnlcmd_username, lul_settings.pnlcmd_password)
end
function actionHttpPOST(lul_device, lul_settings)
  local http = require("socket.http") 
  local url = ""
  if(lul_settings.pnlcmd_url ~= nil) then
    for i, arg in pairs(lul_settings) do
      if(string.find(i, "pnlcmd_") ~= 1 and string.find(i, "DeviceNum") ~= 1 
          and string.find(i, "action") ~=1 and string.find(i, "serviceId") ~= 1) then
        if(string.len(url) == 0) then
          url = "?" .. i .. "=" .. arg
        else
          url = url .. "&" .. i .. "=" .. arg
        end
      end
    end
  else
    local cmd, val = getActionCommandConfig(lul_device, lul_settings)
    if(cmd ~= nil) then
      lul_settings.pnlcmd_url = cmd.http.url
      lul_settings.pnlcmd_username = cmd.http.username
      lul_settings.pnlcmd_password = cmd.http.password
    end
  end
  debug("Calling Action:HttpPOST " .. lul_settings.pnlcmd_url .. url)
  http.TIMEOUT = 5
  http.request(lul_settings.pnlcmd_url .. url)
end
function actionExecuteLua(lul_device, lul_settings)
  debug("Executing Lua")
  if(lul_settings.pnlcmd_ordinal == nil) then
    local cmd, val, ord = getActionCommandConfig(lul_device, lul_settings)
    local act = cmd.execLua
    lul_settings.pnlcmd_ordinal = ord
    lul_settings.pnlcmd_binding = act.callName
    lul_settings.new_control_value = val
  end
  
  
  local pnlFn = pnlLuaFunctions[tonumber(lul_device)]
  if(pnlFn == nil) then
    pnlFn = {}
    pnlLuaFunctions[tonumber(lul_device)] = pnlFn
  end
  -- Get the function that we are referencing.
  local fn = pnlFn[lul_settings.pnlcmd_binding .. lul_settings.pnlcmd_ordinal]
  if(fn == nil) then
    -- Load up the panel configuration.
    local cfg = getConfigFromChild(tonumber(lul_device))
    local elem = findElementByOrdinal(cfg.elements, tonumber(lul_settings.pnlcmd_ordinal))
    -- Get to the code.
    local tblBindings = splitString(lul_settings.pnlcmd_binding, "_")
    local tExec = elem
    for i, v in pairs(tblBindings) do
      tExec = tExec[v]
      if(tExec == nil) then
        break
      end
    end
    if(tExec == nil or tExec.code == nil) then
      debug("Error getting code for execute lua. " .. lul_settings.pnlcmd_binding)
      return
    else
      local ptr, rc = compileLua(tonumber(lul_device), tExec.code)
      if(ptr == nil) then
        debug("Error compiling lua command: " .. rc.Message)
      else
        fn = ptr()
        pnlFn[lul_settings.pnlcmd_binding .. lul_settings.pnlcmd_ordinal] = fn
      end
    end
  end
  local ret, val = pcall(fn, {["panelDeviceId"] = tonumber(lul_device), ["controlValue"] = lul_settings.new_control_value})
  if(ret == false) then
    debug("Error executing Lua command from Virtual Panel control")
  else
    debug("Successfully executed Lua command: " .. lul_settings.pnlcmd_binding .. " Return Value:" .. isNil(val, "nil"))
  end
end

