//@ sourceURL=J_VirtualPanel_UI7.js
//# sourceURL=J_VirtualPanel_UI7.js
/*
 * Virtual Panel Control Plugin
 * Copyright (C) 2015 Robert Strouse

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */
function textFormatter() {
    function pad(d, n, p) {
        var s = '' + d;
        p = p || '0';
        while (s.length < n)
            s = p + s;
        return s;
    }
    var locales = {
        en: {
            A: ['Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday'],
            a: ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'],
            B: ['January', 'February', 'March', 'April', 'May', 'June', 'July', 'August', 'September', 'October', 'November', 'December'],
            b: ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]
        }
    }
    var formats = {
        A: function (d, locale) { return locales[locale].A[d.getDay()]; },
        a: function (d, locale) { return locales[locale].a[d.getDay()]; },
        B: function (d, locale) { return locales[locale].B[d.getMonth()]; },
        b: function (d, locale) { return locales[locale].b[d.getMonth()]; },
        C: function (d) { return Math.floor(d.getFullYear() / 100); },
        c: function (d) { return d.toString(); },
        D: function (d) { return formats.m(d) + '/' + formats.d(d) + '/' + formats.y(d); },
        d: function (d) { return pad(d.getDate(), 2, '0'); },
        e: function (d) { return pad(d.getDate(), 2, ' '); },
        F: function (d) { return formats.Y(d) + '-' + formats.m(d) + '-' + formats.d(d); },
        H: function (d) { return pad(d.getHours(), 2, '0'); },
        I: function (d) { return pad((d.getHours() % 12 || 12), 2); },
        k: function (d) { return pad(d.getHours(), 2, ' '); },
        l: function (d) { return pad((d.getHours() % 12 || 12), 2, ' '); },
        M: function (d) { return pad(d.getMinutes(), 2, '0'); },
        m: function (d) { return pad((d.getMonth() + 1), 2, '0'); },
        n: function (d) { return "\n"; },
        p: function (d) { return (d.getHours() > 11) ? 'PM' : 'AM'; },
        P: function (d) { return formats.p(d).toLowerCase(); },
        R: function (d) { return formats.H(d) + ':' + formats.M(d); },
        r: function (d) { return formats.I(d) + ':' + formats.M(d) + ':' + formats.S(d) + ' ' + formats.p(d); },
        S: function (d) { return pad(d.getSeconds(), 2, '0'); },
        s: function (d) { return Math.floor(d.getTime() / 1000); }, T: function (d) { return formats.H(d) + ':' + formats.M(d) + ':' + formats.S(d); },
        t: function (d) { return "\t"; },
        u: function (d) { return (d.getDay() || 7); },
        v: function (d) { return formats.e(d) + '-' + formats.b(d) + '-' + formats.Y(d); },
        w: function (d) { return d.getDay(); },
        X: function (d) { return d.toTimeString(); }, // wrong? 
        x: function (d) { return d.toDateString(); }, // wrong? 
        Y: function (d) { return d.getFullYear(); },
        y: function (d) { return pad((d.getYear() % 100), 2); },
        '%': function (d) { return '%'; }
    };
    var re = { not_string: /[^s]/, number: /[dief]/, text: /^[^\x25]+/, modulo: /^\x25{2}/, placeholder: /^\x25(?:([1-9]\d*)\$|\(([^\)]+)\))?(\+)?(0|'[^$])?(-)?(\d+)?(?:\.(\d+))?([b-fiosuxX])/, key: /^([a-z_][a-z_\d]*)/i, key_access: /^\.([a-z_][a-z_\d]*)/i, index_access: /^\[(\d+)\]/, sign: /^[\+\-]/ };
    var cache = {};
    
    formats['+'] = formats.c;
    formats.h = formats.b;
    var defaultLocale = 'en';

    this.strftime = function(date, fmt, locale) {
        var r = '';
        var n = 0;
        if (!locale) { locale = defaultLocale; }
        while (n < fmt.length) {
            var c = fmt.substring(n, n + 1);
            if (c == '%') {
                c = fmt.substring(++n, n + 1);
                r += (formats[c]) ? formats[c](date, locale) : c;
            } else r += c;
            ++n;
        }
        return r;
    }
    function parseSprintf(fmt) {
        var _fmt = fmt, match = [], parse_tree = [], arg_names = 0
        while (_fmt) {
            if ((match = re.text.exec(_fmt)) !== null)
                parse_tree[parse_tree.length] = match[0];
            else if ((match = re.modulo.exec(_fmt)) !== null)
                parse_tree[parse_tree.length] = "%"
            else if ((match = re.placeholder.exec(_fmt)) !== null) {
                if (match[2]) {
                    arg_names |= 1;
                    var field_list = [], replacement_field = match[2], field_match = [];
                    if ((field_match = re.key.exec(replacement_field)) !== null) {
                        field_list[field_list.length] = field_match[1];
                        while ((replacement_field = replacement_field.substring(field_match[0].length)) !== "") {
                            if ((field_match = re.key_access.exec(replacement_field)) !== null)
                                field_list[field_list.length] = field_match[1];
                            else if ((field_match = re.index_access.exec(replacement_field)) !== null)
                                field_list[field_list.length] = field_match[1];
                            else
                                throw new SyntaxError("[sprintf] failed to parse named argument key");
                        }
                    }
                    else
                        throw new SyntaxError("[sprintf] failed to parse named argument key");
                    match[2] = field_list;
                }
                else
                    arg_names |= 2;
                if (arg_names === 3)
                    throw new Error("[sprintf] mixing positional and named placeholders is not (yet) supported");
                parse_tree[parse_tree.length] = match;
            }
            else
                throw new SyntaxError("[sprintf] unexpected placeholder");
            _fmt = _fmt.substring(match[0].length);
        }
        return parse_tree;
    }
    function get_type(variable) { 
        return Object.prototype.toString.call(variable).slice(8, -1).toLowerCase() 
    } 
    function str_repeat(input, multiplier) { 
        return Array(multiplier + 1).join(input) 
    }
    this.sprintf = function () {
        var key = arguments[0], c = cache;
        if (!(c[key] && c.hasOwnProperty(key))) {
            c[key] = parseSprintf(key)
        }
        return format.call(null, c[key], arguments)
    }
    function format(parse_tree, argv) {
        var cursor = 1, tree_length = parse_tree.length, node_type = "", arg, output = [], i, k, match, pad, pad_character, pad_length, is_positive = true, sign = "";
        for (i = 0; i < tree_length; i++) {
            node_type = get_type(parse_tree[i]);
            if (node_type === "string") {
                output[output.length] = parse_tree[i];
            }
            else if (node_type === "array") {
                match = parse_tree[i]; // convenience purposes only 
                if (match[2]) { // keyword argument 
                    arg = argv[cursor]
                    for (k = 0; k < match[2].length; k++) {
                        if (!arg.hasOwnProperty(match[2][k]))
                            throw new Error(sprintf("[sprintf] property '%s' does not exist", match[2][k]));
                        arg = arg[match[2][k]];
                    }
                }
                else if (match[1])
                    arg = argv[match[1]];
                else
                    arg = argv[cursor++];
                if (get_type(arg) == "function") {
                    arg = arg();
                }
                if (re.not_string.test(match[8]) && (get_type(arg) != "number" && isNaN(arg))) {
                    throw new TypeError(sprintf("[sprintf] expecting number but found %s", get_type(arg)));
                }
                if (re.number.test(match[8])) {
                    is_positive = arg >= 0;
                }
                switch (match[8]) {
                    case "b":
                        arg = arg.toString(2);
                        break;
                    case "c":
                        arg = String.fromCharCode(arg);
                        break;
                    case "d":
                    case "i":
                        arg = parseInt(arg, 10);
                        break;
                    case "e":
                        arg = match[7] ? arg.toExponential(match[7]) : arg.toExponential();
                        break;
                    case "f":
                        arg = match[7] ? parseFloat(arg).toFixed(match[7]) : parseFloat(arg);
                        break;
                    case "o":
                        arg = arg.toString(8);
                        break;
                    case "s":
                        arg = ((arg = String(arg)) && match[7] ? arg.substring(0, match[7]) : arg);
                        break;
                    case "u":
                        arg = arg >>> 0;
                        break;
                    case "x":
                        arg = arg.toString(16);
                        break;
                    case "X":
                        arg = arg.toString(16).toUpperCase();
                        break;
                }
                if (re.number.test(match[8]) && (!is_positive || match[3])) {
                    sign = is_positive ? "+" : "-";
                    arg = arg.toString().replace(re.sign, "");
                }
                else
                    sign = "";
                pad_character = match[4] ? match[4] === "0" ? "0" : match[4].charAt(1) : " ";
                pad_length = match[6] - (sign + arg).length;
                pad = match[6] ? (pad_length > 0 ? str_repeat(pad_character, pad_length) : "") : "";
                output[output.length] = match[5] ? sign + arg + pad : (pad_character === "0" ? sign + pad + arg : pad + sign + arg);
            }
        }
        return output.join("");
    }

    //Date.prototype.strftime = function (fmt, locale) {
    //    return strftime(this, fmt, locale);
    //};
    //Date.prototype.strftime.formats = formats;
    //Date.prototype.strftime.setDefaultLocale = function (locale) {
    //    defaultLocale = locale;
    //};
    //Date.prototype.strftime.locales = locales;
}
textFormatter.formatVariable = function (ctl, value) {
    var sFormatted = '';
    try
    {
        var sFormat = ctl.textStyle.format;
        var arrFmt = sFormat.match(/\{0:.*\}/g);
        var sFmt = '{0}';
        var val = '';
        if(arrFmt && typeof(arrFmt.length) != 'undefined' && arrFmt.length > 0) {
            sFmt = arrFmt[0];
            var sModifier = '';
            var sDataType = '';
            var f = null;
            sDataType = sFmt.substring(sFmt.indexOf(':') + 1, sFmt.indexOf(','));
            sModifier = sFmt.substring(sFmt.indexOf(',') + 1, sFmt.lastIndexOf('}'));
            switch(sDataType) {
                case 'Percent':
                    val = parseFloat(value) * 100;
                    if (!isNaN(val)) {
                        f = new textFormatter();
                        value = f.sprintf(sModifier, val);
                    }
                    else
                        value = '';
                    break;
                case 'DateTime':
                    if(!isNaN(parseInt(value, 10))) {
                        val = new Date(parseInt(value, 10) * 1000);
                        f = new textFormatter();
                        value = f.strftime(val, sModifier);
                    }
                    else
                        value = '';
                    break;
                case 'Number':
                    val = parseFloat(value);
                    if (!isNaN(val)) {
                        f = new textFormatter();
                        value = f.sprintf(sModifier, val);
                    }
                    else
                        value = '';
                    break;
            }
        }
        sFormatted = sFormat.replace(/\{0:.*\}/g, value);
        sFormatted = sFormatted.replace(/\{0.*\}/g, value);
    }
    catch(error) {
        sFormatted = 'Format Error';
    }
    return sFormatted;
}


function dialogManager(elParent) {
    if ($(elParent).find('#vpnl_modalOverlay').length = 0) {
        if (elParent && typeof elParent != 'undefined')
            $(elParent).html(this.createHtml());
    }
    this.isIntialized = false;
}
dialogManager.prototype.createHtml = function () {
    var html = '<div id="vpnl_modalOverlay" style="opacity:.5;display:none;background-color:#EEEEEE;height:100%;left:0;top:0;min-height:100%;min-width:100%;position:fixed;text-align:center;z-index:10002;"/>';
    html += '<div id="vpnl_modalDialog" class="vpnl_dialogShadow" onclick="textStyler.hidePopups();" style="font-size:8pt;border-top-left-radius:4px;border-top-right-radius:4px;display:none;position:absolute;margin:0px auto;width:400px;border:solid 1px silver;background-color:white;z-index:10003">';
    html += '<div id="vpnl_modalTitle" class="vpnl_modalTitle vpnl_titleGradient" style=""><span id="vpnl_modalTitleText" style="color:white;font-weight:bold;">Modal Title</span><div class="pull-right" style="border:solid 1px silver;border-radius:2px;margin-top:-2px;display:inline-block;background-color:gainsboro;cursor:pointer;margin-bottom:2px;padding-bottom:4px;vertical-align:middle;height:18px;" onmouseover="this.style.backgroundColor=\'orange\';" onmouseout="this.style.backgroundColor=\'gainsboro\';"><img style="height:14px;margin:0px auto;vertical-align:middle;" src="skins/default/img/other/button_close.png" onclick="$(\'#vpnl_modalDialog\').hide(); $(\'#vpnl_modalOverlay\').hide();"/></div></div>';
    html += '<div id="vpnl_modalContent" style="font-size:8pt;width:100%;display:block;position:relative;font-size:9pt;padding-bottom:4px;"></div>';
    html += '<div id="vpnl_modalButtons" style="width:100%;padding-top:7px;border-top:solid 3px orange;height:50px;display:block;"></div>'
    html += '</div>';
    return html;
}
dialogManager.createGradient = function (gradFrom, gradTo) {
    return 'background-image: -webkit-gradient(linear, 0% 0%, 0% 100%, from(#' + gradFrom + '), to(#' + gradTo + ')); '
        + 'background-image: -webkit-linear-gradient(top, #' + gradFrom + ', #' + gradTo + '); '
        + 'background-image: -moz-linear-gradient(top, #' + gradFrom + ', #' + gradTo + '); '
        + 'background-image: -ms-linear-gradient(top, #' + gradFrom + ', #' + gradTo + '); '
        + 'background-image: -o-linear-gradient(top, #' + gradFrom + ', #' + gradTo + '); '
        + 'background-image: linear-gradient(top, #' + gradFrom + ', #' + gradTo + '); '
        + 'filter:progid:DXImageTransform.Microsoft.Gradient(GradientType=0, StartColorStr=#FF' + gradFrom + ', EndColorStr=#FF' + gradTo + ');'
}
dialogManager.bindValue = function (o, el, val, arrayRef) {
    var binding = $(el).data('bind');
    if ($(el).is(':hidden')) {
        var bindIfHidden = $(el).data('bindifhidden');
        if (typeof (bindIfHidden) != 'undefined' && bindIfHidden != null && !bindIfHidden)
            return;
    }
    var dataType = $(el).data('datatype');
    if (binding && binding.length > 0) {
        var sRef = '';
        var arr = binding.split('.');
        var t = o;
        for (var i = 0; i < arr.length - 1; i++) {
            s = arr[i];
            if (s == undefined || s.length == 0) continue;
            sRef += ('.' + s);
            var ndx = s.lastIndexOf('[')
            if (ndx != -1) {
                var v = s.substring(0, ndx);
                if (arrayRef[sRef] == undefined) {
                    if (t[v] == undefined) {
                        t[v] = new Array();
                        t[v].push(new Object());
                        t = t[v][0];
                        arrayRef[sRef] = 0;
                    }
                    else {
                        k = arrayRef[sRef];
                        if (k == undefined) {
                            a = t[v];
                            k = a.length;
                            arrayRef[sRef] = k;
                            a.push(new Object());
                            t = a[k];
                        }
                        else
                            t = t[v][k];
                    }
                }
                else {
                    k = arrayRef[sRef];
                    if (typeof(k) == 'undefined') {
                        a = t[v];
                        k = a.length;
                        arrayRef[sRef] = k;
                        a.push(new Object());
                        t = a[k];
                    }
                    else
                        t = t[v][k];
                }
            }
            else if (typeof(t[s]) == 'undefined') {
                t[s] = new Object();
                t = t[s];
            }
            else
                t = t[s];
        }
        // After all that we should have the reference to the object in t.
        switch (dataType) {
            case 'int':
                var v = parseInt(val, 10);
                t[arr[arr.length - 1]] = isNaN(v) ? 0 : v;
                break;
            case 'float':
                var v = parseFloat(val);
                t[arr[arr.length - 1]] = isNaN(v) ? 0 : v;
                break;
            default:
                t[arr[arr.length - 1]] = val;
                break;
        }
    }
}
dialogManager.getBindings = function (el) {
    var binding = $(el).data('bind');
    if (typeof binding != 'undefined' && binding && binding.length > 0) {
        var arr = binding.split('.');
        var b = new Object();
        if (n != -1)
            return { bindObj: binding.substring(0, n), bindProp: binding.substring(n + 1) };
    }
}
dialogManager.fromDialog = function (el) {
    if (typeof (el) == 'undefined')
        el = $('#vpnl_modalContent');
    var arrayRef = new Array();
    var o = new Object;
    var that = this;
    el.find('input')
        .each(function () {
            if (this.type == 'checkbox')
                dialogManager.bindValue(o, this, this.checked, arrayRef);
            else
                dialogManager.bindValue(o, this, this.value, arrayRef);
        });
    el.find('textarea')
        .each(function () {
            console.log('Binding text area value: ' + $(this).val());
            dialogManager.bindValue(o, this, $(this).val(), arrayRef);
        });
    el.find('select')
        .each(function () {
            dialogManager.bindValue(o, this, this.value, arrayRef);
        });
    el.find('checkbox')
        .each(function () {
            dialogManager.bindValue(o, this, this.checked, arrayRef);
        });

    el.find('div.vpnl_btn')
        .each(function () {
            dialogManager.bindValue(o, this, $(this).hasClass('vpnl_btnSelected'), arrayRef);
        });
    el.find('div.vpnl_radiobtncontainer')
        .each(function () {
            dialogManager.bindValue(o, this, $(this).data('radiovalue'), arrayRef);
            $(this).find('div.vpnl_radiobtn').filter('.vpnl_btnSelected:first').each(function () {
                dialogManager.bindValue(o, this, $(this).data('radiovalue'), arrayRef);
            });
        });
    return o;
}
dialogManager.beautifyJson = function (sJson) {
    var s = sJson.replace(/{/g, '<div style="padding-left:7px;display:block;">{');
    s = s.replace(/}/g, '}</div>');
    s = s.replace(/: <div style="padding-left:7px;display:block;"/g, ': <div style="padding-left:7px;display:inline-block;"');
    s = s.replace(/}<\/div>,/g, '},</div>');
    //s = s.replace(/},<\/div><br\/>\s+<div/g, '},</div><div');
    return s;
}
dialogManager.generateGuid = function () {
    var result, i, j;
    result = '';
    for (j = 0; j < 32; j++) {
        if (j == 8 || j == 12 || j == 16 || j == 20)
            result = result + '-';
        i = Math.floor(Math.random() * 16).toString(16).toUpperCase();
        result = result + i;
    }
    return result
}
dialogManager.createStyles = function () {
    // Blue Hue
    //var gradFrom = '42729b';
    //var gradTo = '295982';
    // Dark Green Hue
    //var gradFrom = '1b9345';
    //var gradTo = '295982';
    // Vera Hue
    var gradFrom = '119d2d';
    var gradTo = '295982';
    $('<style>')
        .prop('type', 'text/css')
        .html('.vpnl_modalTitle { cursor:pointer;width:100%;border-top-left-radius:4px;border-top-right-radius:4px;display:block;border:solid 1px silver;border-bottom:solid 3px orange;padding:4px;background-color:steelblue;}'
            + '.vpnl_titleGradient {' + dialogManager.createGradient(gradFrom, gradTo) + 'color:white;font-weight:bold;}'
            + '.vpnl_dialogShadow { box-shadow: 3px 3px 5px 6px rgba(0,1,1,0.4);}'
            + '.vpnl_popupContainer { border-radius:7px; padding:4px; border:solid 1px silver; background-color:lightyellow; box-shadow: 2px 2px 3px 3px rgba(0,1,1,0.4);}'
            + '.vpnl_dialogSubHeader {' + dialogManager.createGradient(gradFrom, gradTo) + 'color:white;font-weight:bold;}'
            + 'input.vpnl_editField { border-radius:4px; font-size:inherit; font-family:OpenSansLight; border:solid 1px silver; color:gray; padding-left:2px; padding-right:2px;margin-top:2px;}'
            + 'select.vpnl_editField { border-radius:4px; font-size:inherit; font-family:OpenSansLight; border:solid 1px silver; color:gray; padding-left:2px; padding-right:2px; margin-top:2px;}' )
    .appendTo("head");

}
dialogManager.prototype.closeDialog = function () { $('#vpnl_modalDialog').hide(); $('#vpnl_modalOverlay').hide(); }
dialogManager.prototype.setListSelected = function (tbl, rowIndex) {
    $(tbl).find('tbody:first').children('tr')
        .each(function () {
            if (rowIndex == this.rowIndex) {
                this.style.backgroundColor = 'steelblue';
                this.style.color = 'white';
                $(this).data('selected', true);
            }
            else {
                this.style.backgroundColor = '';
                this.style.color = '';
                $(this).data('selected', false);
            }
        });
}
dialogManager.prototype.showModalDialog = function (attrs) {
    var dlg = $('#vpnl_modalDialog');
    if (!this.isIntialized) {
        dlg.draggable({ handle: '#vpnl_modalTitle', containment: '#vpnl_modalOverlay', scroll: false });
    }
    var ovly = $('#vpnl_modalOverlay');

    $('#vpnl_modalTitleText').html(attrs.title);
    dlg[0].style.width = attrs.width;
    $('#vpnl_modalContent').html(attrs.content);
    ovly.show();
    dlg[0].style.left = ((ovly[0].parentElement.offsetWidth / 2) - (parseInt(attrs.width, 10) / 2)) + 'px';
    dlg[0].style.top = $(window).scrollTop() + "px";
    if (attrs.init != undefined) {
        attrs.init();
    }

    if (attrs.buttons && attrs.buttons.length > 0) {
        var btnPane = $('#vpnl_modalButtons');
        vpnl_modalButtons.style.display = '';
        btnPane.html('');
        $('<div style="display:inline;padding-left:4px;padding-right:4px;color:red;" id="vpnl_modalStatusBar"/>').appendTo(btnPane[0]);
        for (var i = 0; i < attrs.buttons.length; i++) {
            var btn = attrs.buttons[i];
            if (typeof (btn.align) != 'undefined' && btn.align == 'left')
                $('<input id="vpnl_modalButton' + i + '" class="vBtn pull-left" type="button" style="margin:2px;" value="' + btn.text + '"></input>').appendTo(btnPane[0]);
            else
                $('<input id="vpnl_modalButton' + i + '" class="vBtn pull-right" type="button" style="margin:2px;" value="' + btn.text + '"></input>').appendTo(btnPane[0]);
            var b = $('#vpnl_modalButton' + i);
            if (btn.event != undefined)
                b.click(btn.event);
        }
        btnPane.show();
    }
    else
        vpnl_modalButtons.style.display = "none";
    for (var p in attrs.params) {
        $('#vpnl_modalContent').data(p, attrs.params[p]);
    }
    dlg.show();
    dlg.css("position", "absolute");
}
dialogManager.prototype.resizeTo = function (width, bCenter) {
    var dlg = $('#vpnl_modalDialog');
    var ovly = $('#vpnl_modalOverlay');
    dlg[0].style.width = width;
    if (bCenter) {
        dlg[0].style.left = ((ovly[0].parentElement.offsetWidth / 2) - (parseInt(width, 10) / 2)) + 'px';
    }
}
dialogManager.prototype.showStatusDialog = function (attrs) {
    attrs.content = '<div style="padding:7px;text-align:center;">' + attrs.message + '</div>';
    this.showModalDialog(attrs);
}
dialogManager.prototype.showMessageDialog = function (attrs) {
    attrs.content = '<div style="padding:7px;text-align:center;">' + attrs.message + '</div>';
    if (!attrs.buttons || attrs.buttons.length == 0) {
        attrs.buttons = [{ text: "Close", event: function () { this.closeDialog(); } }];
    }
    this.showModalDialog(attrs);
}
dialogManager.prototype.getSelectedRow = function (tbl) {
    var rowSelected = null;
    $(tbl).find('tbody:first').children('tr')
        .each(function () {
            if ($(this).data('selected') == true) {
                rowSelected = this;
                return false;
            }
        });
    return rowSelected;
}
dialogManager.stopPropagation = function(e) {
    if(typeof(window.event) != 'undefined')
        e = e || window.event;
    if(typeof(e) != 'undefined' && typeof(e.stopPropagation) != 'undefined')  {
        e.stopPropagation();
    }
}
dialogManager.createTextPopupButton = function (id, text) {
    return '<div style="display:inline-block;" id="' + id + '"><div id="' + id + '_Button" onclick="dialogManager.toggleTextPopup(arguments[0], \'' + id + '\'); return false;" onmouseover="this.style.backgroundColor=\'orange\'" onmouseout="this.style.backgroundColor=\'silver\'" style="cursor:pointer;border-radius:20px;border:outset 1px gray;background-color:silver;display:inline-block;width:20px;height:20px;text-align:center;margin-left:4px;" title="' + text + '"><span style="color:white;font-weight:bold;">?</span></div></div>';
}
dialogManager.createTextPopup = function (parent, id, html) {
    var div = $('<div id="' + id + '_Container" class="vpnl_popupContainer" style="display:none;position:absolute;">' + html + '</div>');
    div.appendTo(parent);
}
dialogManager.toggleTextPopup = function (e, id) {
    $('#vpnl_modalDialog').find('#' + id + '_Container')
        .each(function () {
            if (this.style.display == 'none') {
                // We are showing a new one so hide the others.
                $('div.vpnl_textStylerContainer').hide();
                $('div.vpnl_popupContainer').hide();
                $(this).show();
            }
            else
                $(this).hide();
        });
    dialogManager.stopPropagation(e);
    return false;
}
dialogManager.toggleRadioButton = function (e, el) {
    var bSelected = $(el).hasClass('vpnl_btnSelected');
    var elCont = $(el).parents('div.vpnl_radiobtncontainer:first');
    elCont.find('div.vpnl_radiobtn').removeClass('vpnl_btnSelected').addClass('vpnl_btnUnselected');
    if (!bSelected) {
        $(el).removeClass('vpnl_btnUnselected').addClass('vpnl_btnSelected');
        elCont.data('radiovalue', $(el).data('radiovalue'));
    }
    else
        elCont.data('radiovalue', '');
}
function textStyler(sId, sBinding, flags) {
    this.id = sId;
    this.binding = sBinding;
    this.showAlignment = false;
    if (typeof (flags) != 'undefined') {
        if (typeof (flags.showAlignment) != 'undefined') this.showAlignment = flags.showAlignment;
    }
    this.createHtml = function (opt, sampleText) {
        if (typeof (opt) == 'undefined')
            opt = { underline: false, bold: false, italic: false, fontSize: "", fontFamily: "", color: "" };
        if (typeof (sampleText) == 'undefined' || sampleText.length == 0)
            sampleText = 'Sample Text';
        var html = '<div class="vpnl_textStylerShadow vpnl_textStylerSample" id="' + this.id
            + '_Sample" style="cursor:pointer;display:block;vertical-align:middle;padding-left:4px;padding-right:4px;border:solid 1px silver;border-radius:4px;margin-bottom:0px;"'
            + ' onclick="return textStyler.togglePopup(arguments[0], \'' + this.id + '\');" >'
            + textStyler.createSample(sampleText, opt)

        + '<div id="' + this.id + '_Container" data-bind="' + this.binding + '" class="vpnl_textStyler vpnl_textStylerContainer vpnl_shadow vpnl_textStylerGradient" style="display:none;position:absolute;margin-top:0px;width:347px;" onclick="dialogManager.stopPropagation(arguments[0]);">'
        + '<div style="vertical-align:top;margin:0px;display:inline-block;">'
        + this.createFontStyleButton('italic', (typeof (opt) == 'undefined' || typeof (opt.italic) == 'undefined' || !opt.italic || parseInt(opt.italic, 10) == 0) ? false : true)
        + this.createFontStyleButton('bold', (typeof (opt) == 'undefined' || typeof (opt.bold) == 'undefined' || !opt.bold || parseInt(opt.bold, 10) == 0) ? false : true)
        + this.createFontStyleButton('underline', (typeof (opt) == 'undefined' || typeof (opt.underline) == 'undefined' || !opt.underline || parseInt(opt.underline, 10) == 0) ? false : true)
        + '</div>';
        if (this.showAlignment) {
            html += '<div class="vpnl_radiobtncontainer" data-bind="' + this.binding + '.textAlign"  style="vertical-align:top;margin:0px;margin-left:7px;display:inline-block;">'
            + this.createTextAlignmentButton('left', (typeof (opt) == 'undefined' || typeof (opt.textAlign) == 'undefined' || opt.textAlign != 'left') ? false : true)
            + this.createTextAlignmentButton('center', (typeof (opt) == 'undefined' || typeof (opt.textAlign) == 'undefined' || opt.textAlign != 'center') ? false : true)
            + this.createTextAlignmentButton('right', (typeof (opt) == 'undefined' || typeof (opt.textAlign) == 'undefined' || opt.textAlign != 'right') ? false : true)
            + '</div>';
        }
        html += '<div>'
        + this.createFontFamily(opt.fontFamily)
        + this.createFontSize(opt.fontSize)
        + this.createFontColor(opt.color)
        + '</div></div>';
        return html;
    }
    this.createFontStyleButton = function(style, bSelected) {
        return '<div class="vpnl_btn' + (bSelected ? ' vpnl_btnSelected' : ' vpnl_btnUnselected')
            + '" data-bind="' + this.binding + '.' + style
            + '" style="padding-top:0px;'
            + (style == 'italic' ? 'font-style:italic;' : '')
            + (style == 'bold' ? 'font-weight:bold;' : '')
            + (style == 'underline' ? 'text-decoration:underline;' : '')
            + '" onclick="if($(this).hasClass(\'vpnl_btnSelected\')) {$(this).removeClass(\'vpnl_btnSelected\').addClass(\'vpnl_btnUnselected\');' + this.createUpdateEvent() + ' return dialogManager.stopPropagation(arguments[0]);} else {$(this).removeClass(\'vpnl_btnUnselected\').addClass(\'vpnl_btnSelected\'); ' + this.createUpdateEvent() + ' return dialogManager.stopPropagation(arguments[0]);}"'
            + '><span style="margin:0px auto;">' + style.substring(0, 1).toUpperCase() + '</span></div>';
    }
    this.createTextAlignmentButton = function (style, bSelected) {
        return '<div class="vpnl_radiobtn' + (bSelected ? ' vpnl_btnSelected' : ' vpnl_btnUnselected')
            + '" data-bind="' + this.binding + '.textAlign'
            + '" data-radiovalue="' + style
            + '" title="Align ' + style + '"'
            + '" style="padding-top:0px;'
            + '" onclick="dialogManager.toggleRadioButton(arguments[0], this); ' + this.createUpdateEvent() + '">'
            + '<div style="padding:4px;vertical-align:middle;text-align:' + style + '">'
            + '<div style="height:2px;width:100%;font-size:4pt;">------</div>'
            + '<div style="height:2px;width:100%;font-size:4pt;">--</div>'
            + '<div style="height:2px;width:100%;font-size:4pt;">----</div>'
            + '<div style="height:2px;width:100%;font-size:4pt;">---</div>'
            + '<div style="height:2px;width:100%;font-size:4pt;">----</div>'
            + '</div></div>';
    }
    this.createFontFamily = function (sFamily) {
        var html = '<div style="display:inline-block;"><div class="vpnl_textStylerLabel">Face</div><select data-bind="' + this.binding + '.fontFamily" class="vpnl_fontFamily" style="margin-left:2px;border-radius:4px;border:solid 1px silver;" onchange="' + this.createUpdateEvent() + '">'
        html += '<option value=""' + (!sFamily || sFamily.length == 0 ? 'selected="selected"' : '') + '>[default]</option>'
        for (var i = 0; i < textStyler.fontList.length; i++) {
            html += '<option value="' + textStyler.fontList[i].toLowerCase()
                + '" style="font-family:' + textStyler.fontList[i].toLowerCase() + ';" '
                + (typeof(sFamily) != 'undefined' && sFamily && sFamily.toLowerCase() == textStyler.fontList[i].toLowerCase() ? 'selected="selected"' : '') + '>'
                + textStyler.fontList[i]
                + '</option>';
        }
        html += '</select></div>'
        return html;
    }
    this.createUpdateEvent = function () {
        return 'textStyler.updateSample(\'' + this.id + '\');';
    }
    this.createFontSize = function (sSize) {
        var createOption = function (s) {
            return '<option value="' + s + '"' + (sSize == s ? 'selected="selected"' : '') + '>' + s + '</option>';
        };
        if (!sSize || sSize.length == 0) sSize = "1em";
        var html = '<div style="display:inline-block"><div class="vpnl_textStylerLabel">Size</div><select data-bind="' + this.binding + '.fontSize" class="vpnl_fontSize" style="margin-left:2px;border-radius:4px;border:solid 1px silver;" onchange="' + this.createUpdateEvent() + '">'
        html += createOption(".5em");
        html += createOption(".7em");
        html += createOption(".8em");
        html += createOption(".9em");
        html += createOption("1em");
        html += createOption("1.1em");
        html += createOption("1.2em");
        html += createOption("1.5em");
        html += createOption("1.7em");
        html += createOption("2em");
        html += createOption("2.1em");
        html += createOption("2.2em");
        html += createOption("2.5em");
        html += createOption("2.7em");
        html += createOption("3em");
        html += createOption("3.1em");
        html += createOption("3.2em");
        html += createOption("3.5em");
        html += createOption("3.7em");
        html += createOption("4em");
        html += '</select></div>'
        return html;
    }
    this.createFontColor = function (sColor) {
        var html = '<div style="display:inline-block;"><div class="vpnl_textStylerLabel">Color</div><input type="text" data-bind="' + this.binding + '.color" class="vpnl_fontColor" list="vpnl_colordatalist" style="margin-left:2px;border-radius:4px;border:solid 1px silver;" value="' + sColor + '" onchange="' + this.createUpdateEvent() + '" >'
        html += '<datalist id="vpnl_colordatalist">';
        for (var clr in textStyler.colorList) {
            if(typeof(window.event) != 'undefined')
                html += '<option value="' + clr + '" style="color:' + clr + ';">' + textStyler.colorList[clr] + '</option>';
            else
                html += '<option value="' + clr + '" style="color:' + clr + ';">' + clr + '</option>';

        }
        html += '</datalist></input></div>';
        return html;
    }

    if (typeof (this.binding) == 'undefined') this.binding = '.';
    if (typeof (this.id) == 'undefined') this.id = '';
}
textStyler.createSample = function (sText, opt) {
    var html = '<div class="" style="font-size:14px;">'
        + '<span class="vpnl_textStylerSample"'
        + ' style="margin-top:4px;text-align:center;background-color:white;'
        + textStyler.createCss(opt)
        + '">' + sText + '</span></div>'
    return html;
}
textStyler.createCss = function(opt) {
    if (typeof (opt) == 'undefined')
        opt = { color: '', bold: false, underline: false, italic: false, fontSize: '', fontFamily: '' }
    var css = 'color:' + ((typeof (opt.color) == 'undefined' || opt.color == '') ? 'inherit' : opt.color)
    +  ((typeof(opt.width) == 'undefined' || opt.width == '' || opt.width == '0') ? '' : ';width:' + opt.width + 'px;overflow-x:hidden;')
    + ';font-weight:' + (opt.bold ? 'bold' : 'normal')
    + ';text-decoration:' + (opt.underline ? 'underline' : 'none')
    + ';font-family:' + ((typeof (opt.fontFamily) == 'undefined' || opt.fontFamily == '') ? 'inherit' : opt.fontFamily)
    + ';font-style:' + (opt.italic ? 'italic' : 'normal')
    + ';font-size:' + ((typeof (opt.fontSize) == 'undefined' || opt.fontSize == '') ? '1em' : opt.fontSize) 
    + ';text-align:' + ((typeof(opt.textAlign) == 'undefined' || opt.textAlign == '') ? 'center' : opt.textAlign) + ';'
    return css;
}
textStyler.createCssMargins = function (opt) {
    if (typeof (opt) == 'undefined' || opt == null || !opt) return '';
    return 'margin-left:' + opt.left + 'px;margin-top:' + opt.top + 'px;margin-right:' + opt.right + 'px;margin-bottom:' + opt.bottom + 'px;';
}
textStyler.colorList = { aliceblue: '#F0F8FF', antiquewhite: '#FAEBD7', aqua: '#00FFFF', aquamarine: '#7FFFD4', azure: '#F0FFFF', beige: '#F5F5DC', bisque: '#FFE4C4', black: '#000000', blanchedalmond: '#FFEBCD', blue: '#0000FF', blueviolet: '#8A2BE2', brown: '#A52A2A', burlywood: '#DEB887', cadetblue: '#5F9EA0', chartreuse: '#7FFF00', chocolate: '#D2691E', coral: '#FF7F50', cornflowerblue: '#6495ED', cornsilk: '#FFF8DC', crimson: '#DC143C', cyan: '#00FFFF', darkblue: '#00008B', darkcyan: '#008B8B', darkgoldenrod: '#B8860B', darkgray: '#A9A9A9', darkgrey: '#A9A9A9', darkgreen: '#006400', darkkhaki: '#BDB76B', darkmagenta: '#8B008B', darkolivegreen: '#556B2F', darkorange: '#FF8C00', darkorchid: '#9932CC', darkred: '#8B0000', darksalmon: '#E9967A', darkseagreen: '#8FBC8F', darkslateblue: '#483D8B', darkslategray: '#2F4F4F', darkslategrey: '#2F4F4F', darkturquoise: '#00CED1', darkviolet: '#9400D3', deeppink: '#FF1493', deepskyblue: '#00BFFF', dimgray: '#696969', dimgrey: '#696969', dodgerblue: '#1E90FF', firebrick: '#B22222', floralwhite: '#FFFAF0', forestgreen: '#228B22', fuchsia: '#FF00FF', gainsboro: '#DCDCDC', ghostwhite: '#F8F8FF', gold: '#FFD700', goldenrod: '#DAA520', gray: '#808080', grey: '#808080', green: '#008000', greenyellow: '#ADFF2F', honeydew: '#F0FFF0', hotpink: '#FF69B4', indianred: '#CD5C5C', indigo: '#4B0082', ivory: '#FFFFF0', khaki: '#F0E68C', lavender: '#E6E6FA', lavenderblush: '#FFF0F5', lawngreen: '#7CFC00', lemonchiffon: '#FFFACD', lightblue: '#ADD8E6', lightcoral: '#F08080', lightcyan: '#E0FFFF', lightgoldenrodyellow: '#FAFAD2', lightgray: '#D3D3D3', lightgrey: '#D3D3D3', lightgreen: '#90EE90', lightpink: '#FFB6C1', lightsalmon: '#FFA07A', lightseagreen: '#20B2AA', lightskyblue: '#87CEFA', lightslategray: '#778899', lightslategrey: '#778899', lightsteelblue: '#B0C4DE', lightyellow: '#FFFFE0', lime: '#00FF00', limegreen: '#32CD32', linen: '#FAF0E6', magenta: '#FF00FF', maroon: '#800000', mediumaquamarine: '#66CDAA', mediumblue: '#0000CD', mediumorchid: '#BA55D3', mediumpurple: '#9370D8', mediumseagreen: '#3CB371', mediumslateblue: '#7B68EE', mediumspringgreen: '#00FA9A', mediumturquoise: '#48D1CC', mediumvioletred: '#C71585', midnightblue: '#191970', mintcream: '#F5FFFA', mistyrose: '#FFE4E1', moccasin: '#FFE4B5', navajowhite: '#FFDEAD', navy: '#000080', oldlace: '#FDF5E6', olive: '#808000', olivedrab: '#6B8E23', orange: '#FFA500', orangered: '#FF4500', orchid: '#DA70D6', palegoldenrod: '#EEE8AA', palegreen: '#98FB98', paleturquoise: '#AFEEEE', palevioletred: '#D87093', papayawhip: '#FFEFD5', peachpuff: '#FFDAB9', peru: '#CD853F', pink: '#FFC0CB', plum: '#DDA0DD', powderblue: '#B0E0E6', purple: '#800080', rebeccapurple: '#663399', red: '#FF0000', rosybrown: '#BC8F8F', royalblue: '#4169E1', saddlebrown: '#8B4513', salmon: '#FA8072', sandybrown: '#F4A460', seagreen: '#2E8B57', seashell: '#FFF5EE', sienna: '#A0522D', silver: '#C0C0C0', skyblue: '#87CEEB', slateblue: '#6A5ACD', slategray: '#708090', slategrey: '#708090', snow: '#FFFAFA', springgreen: '#00FF7F', steelblue: '#4682B4', tan: '#D2B48C', teal: '#008080', thistle: '#D8BFD8', tomato: '#FF6347', turquoise: '#40E0D0', violet: '#EE82EE', wheat: '#F5DEB3', white: '#FFFFFF', whitesmoke: '#F5F5F5', yellow: '#FFFF00', yellowgreen: '#9ACD32' };
textStyler.styleList = { italic: 'Italic', bold: 'bold', underline: 'underline' }
textStyler.fontList = ['Arial', 'Arial Black', 'Book Antiqua', 'Comic Sans MS', 'Courier New', 'Impact', 'Georgia', 'Lucida Sans Unicode', 'Lucida Console', 'OpenSansLight', 'Tahoma', 'Trebuchet MS', 'Verdana'];
textStyler.createStyles = function () {
    var gradFrom = 'DAD8D2';
    var gradTo = 'ECECE8';
    $('<style>')
        .prop('type', 'text/css')
        .html('.vpnl_textStyler { display:inline-block;border-radius:4px;border:outset 2px gainsboro;padding:4px;background-color:gainsboro;font-family:arial;font-size:7pt; } '
            + '.vpnl_btn { border: vertical-align:middle;text-align:center;margin:1px;cursor:pointer;display:inline-block;width:24px;height:24px;font-size:12pt;font-family:times new roman; } '
            + '.vpnl_radiobtn { border: vertical-align:middle;text-align:center;margin:1px;cursor:pointer;display:inline-block;width:24px;height:24px;font-size:12pt;font-family:times new roman; } '
            + '.vpnl_btnUnselected { border: outset 1px silver; ' + dialogManager.createGradient(gradFrom, gradTo) + '} '
            + '.vpnl_btnSelected { border: inset 1px silver; ' + dialogManager.createGradient('FF9900', 'FFE0B2') + '} '
            + '.vpnl_textStylerGradient {' + dialogManager.createGradient(gradFrom, gradTo) + 'font-weight:bold;} '
            + '.vpnl_textStylerLabel { padding-left:4px; }'
            + 'div.vpnl_textStylerSample { margin-bottom:4px; margin-top:2px; }'
            + '.vpnl_textStylerShadow { box-shadow: 0px 2px 3px rgba(0,1,1,1.3), inset 0px 0px 2px rgba(255,255,255,0.8);} ')
    .appendTo("head");
}
textStyler.createHtml = function (id, binding, opt, sampleText, flags) {
    var ts = new textStyler(id, binding, flags);
    return ts.createHtml(opt, sampleText);
}
textStyler.updateSample = function (id) {
    var popup = $('#' + id + '_Container');
    var sBinding = popup.data('bind');
    var d = dialogManager.fromDialog($('#' + id + '_Container'));
    var opt = eval('d' + sBinding);

    if (typeof (opt) != 'undefined') {
        $('#' + id + '_Sample').find('span.vpnl_textStylerSample').css(
            {
                fontSize: opt.fontSize,
                color: opt.color,
                fontFamily: typeof (opt.fontFamily) == 'undefined' || opt.fontFamily == '' || opt.fontFamily == '[default]' ? 'inherit' : opt.fontFamily,
                fontWeight: opt.bold ? 'bold' : 'normal',
                textDecoration: opt.underline ? 'underline' : 'none',
                fontStyle: opt.italic ? 'italic' : 'normal'
            });
    }
}
textStyler.updateSampleText = function (id, sampleText) {
    $('#' + id + '_Sample').find('span.vpnl_textStylerSample').html(sampleText);
}
textStyler.hidePopups = function() {
    $('div.vpnl_textStylerContainer').hide();
    $('div.vpnl_popupContainer').hide();
}
textStyler.togglePopup = function (e, id) {
    //alert(e);
    $(this).find('#' + id + '_Container')
        .each(function() {
            if (this.style.display == 'none') {
                // We are showing a new one so hide the others.
                $('div.vpnl_textStylerContainer').hide();
                $('div.vpnl_popupContainer').hide();
                $(this).show();
            }
            else
                $(this).hide();
        });
    dialogManager.stopPropagation(e);
    return false;
}
textStyler.transformVariableFormat = function(textStyle, sText) {
    function format() {
        var s = arguments[0];
        for (var i = 0; i < arguments.length - 1; i++) {
            var reg = new RegExp('\\{' + i + ':', 'g');
            s = s.replace(reg, '{' + arguments[i + 1] + ':');
            reg = new RegExp('\\{' + i + '\\}', 'g');
            s = s.replace(reg, '{' + arguments[i + 1] + '}');
        }
        return s;
    }
    if (typeof (textStyle) == 'undefined' || typeof (textStyle.format) == 'undefined' || textStyle.format.length == 0)
        return '{' + sText + '}';
    return format(textStyle.format, sText);
}

var virtualPanelManager = (function (api) {
    var device = api.getCpanelDeviceId();
    var modalDialog = new dialogManager();
    var uuid = dialogManager.generateGuid();
    var configuredPanels = new Array();
    var myModule = {};
    function showPanelsTab() {
        device = api.getCpanelDeviceId();
        var html = modalDialog.createHtml();
        html += '<div id="vpnl_showconfig"><img src="skins/default/img/other/plusOpenBtn.png" onclick="if($(\'#vpnl_configresponse\')[0].style.display==\'none\') { $(\'#vpnl_configresponse\').show(); this.src=\'skins/default/img/other/minusLightGreen.png\'; } else { $(\'#vpnl_configresponse\').hide(); this.src=\'skins/default/img/other/plusOpenBtn.png\'; }" />';
        html += '<div style="display:none;font-size:7pt;white-space:pre;" id="vpnl_configresponse"></div></div>';
        html += '<div id="vpnl_deviceresponse"></div>';
        html += '<div id="vpnl_panelconfig">Getting Configuration...</div>';
        html += '<input class="vBtn pull-right" type="button" id="vpnl_btnSaveConfiguration" style="margin-top:4px;display:none;" value="Save Panels" onclick="virtualPanelManager.saveConfiguredPanels(\'vpnl_panelconfig\');"/>';
        api.setCpanelContent(html);
        loadAvailablePanels('vpnl_panelconfig');
    }
    function loadAvailablePanels(sPanelId) {
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_GetConfiguredPanels",
                    deviceId: device,
                    rand: Math.random(),
                    output_format: "json"
                },
                onSuccess: function (response) {
                	//console.log(response);
                    configuredPanels = JSON.parse(response.responseText);
                    $('#vpnl_configresponse').html(JSON.stringify(configuredPanels, null, '\t'));
                    if (configuredPanels) {
                        $('#' + sPanelId).html(buildConfiguredPanels(configuredPanels));
                        $('#vpnl_btnSaveConfiguration').show();
                    }
                    else {
                        $('#vpnl_btnSaveConfiguration').hide();
                        $('#' + sPanelId).html('Vera is currently too busy doing something else. Please reload your browser.');
                    }
                },
                onFailure: function (response) {
                    $('#vpnl_deviceresponse').html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Could not get panels from Vera:' + response.statusText + '</p>');
                }
            });
        }
        catch (err) {
            $('#vpnl_deviceresponse').html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Error getting panels from Vera.' + err.message + '</p>');
        }
    }
    function buildConfiguredPanels(configuration) {
        var html = '<table style="width:100%;border:solid 1px black;margin-top:4px;border-radius-top:4px;"><tbody>';
        html += '<tr class="vpnl_titleGradient" style="background-color:steelblue;color:white;font-weight:bold;"><td style="padding-left:2px;padding-right:2px;border-bottom:solid 3px orange;">Settings</td></tr>';

        html += '<tr title="" onmouseover="jQuery(this).find(\'div.vpnl_inst\').show();" onmouseout="jQuery(this).find(\'div.vpnl_inst\').hide();">';
        html += '<td style="padding-left:4px;">'
        html += '<input type="checkbox" style="" '
                + (parseInt(api.getDeviceState(device, 'urn:rstrouse-com:serviceId:VirtualPanel1', 'Debug'), 10) == 1 ? 'checked="checked"' : '')
                + ' onclick="virtualPanelManager.onSetDebug(this.checked);"'
                + '>Debug to Log</input>';
        html += '<div class="vpnl_inst" style="position:absolute;width:350px;margin-left:54px;display:none;font-size:8pt;border:solid 1px teal;background-color:lightyellow;border-radius:4px;padding-left:4px;padding-right:4px;">When this box is detailed information will be written to the luup log.  Uncheck this box if you are not troubleshooting issues with your panels.<div>'
        html += '</td></tr>';
        html += '<tr title="" onmouseover="jQuery(this).find(\'div.vpnl_inst\').show();" onmouseout="jQuery(this).find(\'div.vpnl_inst\').hide();">';
        html += '<td style="padding-left:4px;">'
        html += '<input type="checkbox" style="" '
                + (parseInt(api.getDeviceState(device, 'urn:rstrouse-com:serviceId:VirtualPanel1', 'CompressConfig'), 10) == 1 ? 'checked="checked"' : '')
                + ' onclick="virtualPanelManager.onSetCompression(this.checked);"'
                + '>Compress Configuration Files</input>';
        html += '<div class="vpnl_inst" style="position:absolute;width:350px;margin-left:54px;display:none;font-size:8pt;border:solid 1px teal;background-color:lightyellow;border-radius:4px;padding-left:4px;padding-right:4px;">When this box is checked the plugin will store its configuration files in compressed on your Vera to conserve Vera resources.  If the box is not checked the files will be stored uncompressed.<div>'
        html += '</td></tr>';
        html += '</tbody></table>';

        html += '<table style="width:100%;border:solid 1px black;margin-top:4px;border-radius-top:4px;"><tbody>';
        html += '<tr style="background-color:steelblue;color:white;font-weight:bold;" class="vpnl_titleGradient"><td colspan="4" style="padding-left:2px;padding-right:2px;border-bottom:solid 3px orange;"><span>Configured Panels</span></td></tr>';
        html += '<tr style="background-color:gainsboro;"><td style="width:24px;"/><td style="width:200px;padding-left:4px;padding-right:4px;">Panel Name</td><td style="width:200px;padding-left:4px;padding-right:4px;">Panel Id</td><td style=padding-left:4px;padding-right:4px;"></td>';
        if (configuration && configuration.length > 0) {
            for (var i = 0; i < configuration.length; i++) {
                e = configuration[i];
                if (e == null || e.isDeleted) continue;
                html += '<tr data-ordinal="' + e.ordinal + '"><td><img style="width:24px;" src="skins/default/img/scenes/create_scene_trigger_button_remove.png" onmouseover="this.src = \'skins/default/img/scenes/create_scene_trigger_button_remove_hover.png\'" onmouseout="this.src = \'skins/default/img/scenes/create_scene_trigger_button_remove.png\'" onclick="virtualPanelManager.deletePanel($(this).parents(\'tr:first\').data(\'ordinal\'));" /></td>';
                html += '<td>' + e.description + '</td>';
                html += '<td>' + e.altId + '</td><td/></tr>';
            }
        }
        html += '<tr><td><img style="width:22px;cursor:pointer;" onclick="virtualPanelManager.addPanel(' + device + ');" src="skins/default/img/other/plusOpenBtn.png"/></td><td colspan="3"><span style="white-space:nowrap;font-size:8pt;font-style:italic;">Click + to add another panel.  To configure that panel open the layout tab on the newly created device on your dashboard.</span></td></tr></tbody></table>';
        return html;
    }
    function getNextAltOrdinal() {
        var ord = 0;
        for (var i = 0; i < configuredPanels.length; i++) {
            if (configuredPanels[i].ordinal >= ord)
                ord = configuredPanels[i].ordinal;
        }
        return ord + 1;
    }
    function addPanel() {
        modalDialog.showModalDialog({
            title: "Provide a name for your new panel",
            width: "270px",
            content: '<div style=padding:4px;"><input data-bind="' + '.description" style="width:100%;border-radius:4px;" type="text" value=""/></div>',
            init: function () { },
            buttons: [
                { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                {
                    text: "Add Panel", event: function () {
                        var pnl = dialogManager.fromDialog();
                        pnl.deviceId = 0
                        pnl.isDeleted = false;
                        pnl.ordinal = getNextAltOrdinal();
                        pnl.altId = 'VirtualPanel_' + device + '_' + pnl.ordinal;
                        pnl.uuid = dialogManager.generateGuid();
                        configuredPanels.push(pnl);
                        $('#vpnl_panelconfig').html(buildConfiguredPanels(configuredPanels));
                        $('#vpnl_configresponse').html(JSON.stringify(configuredPanels, null, '\t'));
                        modalDialog.closeDialog();
                    }
                }
            ]
        });
    }
    function deletePanel(ordinal) {
        var pnl = null;
        var ndx = 0;
        for (var i = 0; i < configuredPanels.length; i++) {
            if (parseInt(configuredPanels[i].ordinal, 10) == parseInt(ordinal, 10)) {
                pnl = configuredPanels[i];
                ndx = i;
            }
        }
        modalDialog.showMessageDialog({
            title: 'Confirm Delete ' + pnl.description + ' (' + pnl.altId + ')',
            message: 'Are you sure you want to delete panel ' + pnl.description + ' (' + pnl.altId + ')?  If so press the Delete button otherwise press Cancel. <p style="font-size:8pt">NOTE: Deleted items will be saved when the Save Panels button is pressed.',
            width: '500px',
            buttons: [{ text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                {
                    text: 'Delete ' + pnl.description,
                    event: function () {
                        if (parseInt(pnl.deviceId, 10) == 0 || (typeof(pnl.isAdded) != 'undefined' && pnl.isAdded == false))
                            configuredPanels.splice(ndx, 1);
                        else
                            pnl.isDeleted = true;
                        $('#vpnl_panelconfig').html(buildConfiguredPanels(configuredPanels));
                        $('#vpnl_configresponse').html(JSON.stringify(configuredPanels, null, '\t'));
                        modalDialog.closeDialog();
                    }
                }
            ]
        });
    }
    function saveConfiguredPanels(sPanelId) {
        modalDialog.showStatusDialog({
            title: "Saving Configuration...",
            width: '350px',
            message: 'Saving Configuration... Please wait!'
        });
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_SaveConfiguredPanels",
                    rand: Math.random(),
                    config: JSON.stringify(configuredPanels).replace(/ /g, '%20'),
                    output_format: "json"
                },
                onSuccess: function (response) {
                    modalDialog.closeDialog();
                    configuredPanels = JSON.parse(response.responseText);
                    $('#vpnl_configresponse').html(JSON.stringify(configuredPanels, null, '\t'));
                    if (configuredPanels) {
                        $('#' + sPanelId).html(buildConfiguredPanels(configuredPanels));
                        $('#vpnl_btnSaveConfiguration').show();
                    }
                    else {
                        $('#vpnl_btnSaveConfiguration').hide();
                        $('#' + sPanelId).html('Vera is currently too busy doing something else. Please reload your browser.');
                    }

                },
                onFailure: function () {
                    modalDialog.showMessageDialog({
                        title: 'Error',
                        width: '350px',
                        message: 'Error saving configuration'
                    });
                }
            });
        }
        catch (err) {
        }
    }
    function onSetDebug(bDebug) {
        api.setDeviceStatePersistent(device, 'urn:rstrouse-com:serviceId:VirtualPanel1', 'Debug', bDebug == true ? '1' : '0');
        if (!bDebug)
            api.setDeviceStatePersistent(device, 'urn:rstrouse-com:serviceId:VirtualPanel1', 'DebugStatus', '<div style="color:green;font-size:.7em;">Disabled</div>');
        else
            api.setDeviceStatePersistent(device, 'urn:rstrouse-com:serviceId:VirtualPanel1', 'DebugStatus', '<div style="color:red;font-size:.7em;">Enabled</div>');
    }
    function onSetCompression(bCompress) {
        api.setDeviceStatePersistent(device, 'urn:rstrouse-com:serviceId:VirtualPanel1', 'CompressConfig', bCompress == true ? '1' : '0');
        if (!bDebug)
            api.setDeviceStatePersistent(device, 'urn:rstrouse-com:serviceId:VirtualPanel1', 'CompressionStatus', '<div style="color:red;font-size:.7em;">Disabled</div>');
        else
            api.setDeviceStatePersistent(device, 'urn:rstrouse-com:serviceId:VirtualPanel1', 'CompressionStatus', '<div style="color:green;font-size:.7em;">Enabled</div>');
    }

    function cleanUp() { }
    function init() {
        // register to events...
        api.registerEventHandler('on_ui_cpanel_before_close', myModule, 'onBeforeCpanelClose');
        dialogManager.createStyles();
        textStyler.createStyles();
    }
    function onBeforeCpanelClose() {
        console.log('handler for before cpanel close');
    }
    myModule = {
        uuid: uuid,
        init: init,
        addPanel: addPanel,
        saveConfiguredPanels: saveConfiguredPanels,
        deletePanel: deletePanel,
        showPanelsTab: showPanelsTab,
        onBeforeCpanelClose: onBeforeCpanelClose,
        onSetDebug: onSetDebug,
        onSetCompression : onSetCompression,
        cleanUp: cleanUp
    };
    return myModule;
})(api);
var virtualPanelBuilder = (function (api) {
    var device = api.getCpanelDeviceId();
    var modalDialog = new dialogManager();
    var uuid = dialogManager.generateGuid();
    var panelConfig = {};
    var iconList = {};
    var availableDevices = {};
    var deviceList = {};
    var dirtyPositions = {};
    /********************************************************
     * Entry Points
     ********************************************************/
    function showLayoutTab() {
        device = api.getCpanelDeviceId();
        var html = modalDialog.createHtml();
        html += '<img src="skins/default/img/other/plusOpenBtn.png" onclick="if($(\'#vpnl_configresponse\')[0].style.display==\'none\') { $(\'#vpnl_configresponse\').show(); this.src=\'skins/default/img/other/minusLightGreen.png\'; } else { $(\'#vpnl_configresponse\').hide(); this.src=\'skins/default/img/other/plusOpenBtn.png\'; }" />';
        html += '<div style="display:none;font-size:7pt;white-space:pre;" id="vpnl_configresponse"></div>';
        html += '<div id="vpnl_deviceresponse"></div>';
        html += '<div id="vpnl_panelconfig">Getting Layout Configuration...</div>';
        html += '<table><tbody><tr><td style="vertical-align:top;"><div style="display:inline-block;"><span style="font-size:8pt;font-weight:bold;margin:0px;">Drag and Drop elements within the grid to position them on the dashboard view.  Elements on the dashboard are drawn from top to bottom/left to right and are reflected on the Dashboard View as you build your panel.  To position items on the details tab click on the item to highlight it.  You can then use your mouse to drag it to the desired location or your arrow keys to nudge items into position.</span>';
        html += '</div></td></tr><tr><td><div><span style="font-size:9pt;font-weight:bold;margin:0px;">NOTE: When finished adding or changing controls on your panel you must press the Reload Luup button.  Wait for luup to finish loading, then refresh your browser to see your changes.</span></div></td><td style="vertical-align:top;"></div></td></tr></tbody></table>';
        api.setCpanelContent(html);
        loadPanelConfig('vpnl_panelconfig');
        loadAvailableDevices();
        loadUserData();
    }
    function showTriggersTab() {
        device = api.getCpanelDeviceId();
        var html = modalDialog.createHtml();
        html += '<img src="skins/default/img/other/plusOpenBtn.png" onclick="if($(\'#vpnl_configresponse\')[0].style.display==\'none\') { $(\'#vpnl_configresponse\').show(); this.src=\'skins/default/img/other/minusLightGreen.png\'; } else { $(\'#vpnl_configresponse\').hide(); this.src=\'skins/default/img/other/plusOpenBtn.png\'; }" />';
        html += '<div style="display:none;font-size:7pt;white-space:pre;" id="vpnl_configresponse"></div>';
        html += '<div id="vpnl_deviceresponse"></div>';
        html += '<div id="vpnl_panelconfig">Getting Trigger Configuration...</div>';
        html += '<table style="width:100%;"><tbody><tr><td><div><span style="font-size:9pt;font-weight:bold;margin:0px;">NOTE: When finished adding or changing triggers for your panel you must press the Reload Luup button.  Wait for luup to finish loading, then refresh your browser to see your changes.</span></div></td><td style="vertical-align:top;"><input class="vBtn pull-right" type="button" id="vpnl_btnSaveConfiguration" style="margin-top:4px;display:inline-block;" value="Reload Luup" onclick="virtualPanelBuilder.onReloadLuup();"/></div></td></tr></tbody></table>';
        api.setCpanelContent(html);
        loadTriggerConfig('vpnl_panelconfig');
        loadAvailableDevices();
        loadUserData();
    }
    function showStatesTab() {
        device = api.getCpanelDeviceId();
        var html = modalDialog.createHtml();
        html += '<img src="skins/default/img/other/plusOpenBtn.png" onclick="if($(\'#vpnl_configresponse\')[0].style.display==\'none\') { $(\'#vpnl_configresponse\').show(); this.src=\'skins/default/img/other/minusLightGreen.png\'; } else { $(\'#vpnl_configresponse\').hide(); this.src=\'skins/default/img/other/plusOpenBtn.png\'; }" />';
        html += '<div style="display:none;font-size:7pt;white-space:pre;" id="vpnl_configresponse"></div>';
        html += '<div id="vpnl_deviceresponse"></div>';
        html += '<div id="vpnl_panelconfig">Getting State Configuration...</div>';
        html += '<table style="width:100%;"><tbody><tr><td><div><span style="font-size:9pt;font-weight:bold;margin:0px;">NOTE: When finished adding or changing states for your panel you must press the Reload Luup button.  Wait for luup to finish loading, then refresh your browser to see your changes.</span></div></td><td style="vertical-align:top;"><input class="vBtn pull-right" type="button" id="vpnl_btnSaveConfiguration" style="margin-top:4px;display:inline-block;" value="Reload Luup" onclick="virtualPanelBuilder.onReloadLuup();"/></div></td></tr></tbody></table>';
        api.setCpanelContent(html);
        loadStateConfig('vpnl_panelconfig');
        loadAvailableDevices();
        loadUserData();
    }
    /********************************************************
     * Main Templates
     ********************************************************/
    function buildPanelConfig(configuration) {
        var html = '<table  style="width:100%;border:solid 1px black;margin-top:4px;border-radius-top:4px;"><tbody>';
        html += '<tr style="background-color:steelblue;color:white;font-weight:bold;" class="vpnl_titleGradient"><td colspan="3" style="padding-left:2px;padding-right:2px;border-bottom:solid 3px orange;">Panel Configuration</td></tr>';
        html += '<tr title="" style="vertical-align:middle;">';
        html += '<td style="width:200px;padding-left:4px;padding-right:4px;">Panel Icons</td>';
        html += '<td style="padding-left:4px;padding-right:4px;width:77px;" onmouseover="jQuery(this).find(\'div.spInst\').show();" onmouseout="jQuery(this).find(\'div.spInst\').hide();">';
        // Default Icon
        html += '<div style="display:block;width:70px;"><div style="border:outset 2px;cursor:pointer;padding:1px;width:37px;display:block;border-radius:5px;margin:0px auto;"><img id="vpnl_panelIcon" onclick="virtualPanelBuilder.onClickSelectDefaultIcon()" style="width:32px;" src="skins/default/img/devices/device_states/' + configuration.icon + '"/></div>';
        html += '<div style="font-size:8pt;font-style:italic;text-align:center;display:block;"><span style="white-space:nowrap;">Default Icon</span></div></div>';
        html += '<div class="spInst" style="position:absolute;z-index:30000;width:300px;margin-left:54px;display:none;font-size:8pt;border:solid 1px teal;background-color:lightyellow;border-radius:4px;padding-left:4px;padding-right:4px;">The currently selected default icon for the device in the dashboard.  Click here to select a new icon.</div>'
        html += '<td id="vpnl_stateIconDisplay" style="vertical-align:middle;border-left:inset 2px;">'
        // State Icons
        html += buildStateIconConfig(configuration);
        html += '</tbody></table>';
        html += '<div id="vpnl_panelElements">' + buildPanelElements(configuration) + '</div>';
        return html;
    }
    function buildStateIconConfig(configuration) {
        var html = '<table style=""><tbody><tr>';
        if (configuration.stateIcons) {
            for (var i = 0; i < configuration.stateIcons.length; i++) {
                var icon = configuration.stateIcons[i];
                html += '<td style="padding-left:4px;padding-right:4px;width:50px;" onmouseover="jQuery(this).find(\'div.spInst\').show();" onmouseout="jQuery(this).find(\'div.spInst\').hide();">';
                html += '<div style="display:block;width:50px;"><div style="border:outset 2px;cursor:pointer;padding:1px;width:37px;display:block;border-radius:5px;margin:0px auto;"><img id="vpnl_stateIcon' + i + '" onclick="virtualPanelBuilder.onClickEditStateIcon(' + i + ')" style="width:32px;" src="skins/default/img/devices/device_states/' + icon.icon + '"/></div>';
                html += '<div style="font-size:8pt;font-style:italic;text-align:center;display:block;"><span style="white-space:nowrap;">State Icon</span></div></div>';
                html += '<div class="spInst" style="position:absolute;width:377px;z-index:30000;margin-left:54px;display:none;font-size:8pt;border:solid 1px teal;background-color:lightyellow;border-radius:4px;padding-left:4px;padding-right:4px;">This icon will show when the following conditions are true.';
                html += '<ul>'
                for (var k = 0; k < icon.conditions.length; k++) {
                    html += '<li>' + icon.conditions[k].serviceId + '[' + icon.conditions[k].variable + '] ' + icon.conditions[k].operator + ' [' + icon.conditions[k].value + ']';
                }
                html += '</ul>';
                html += 'Click here to edit this icon and its conditions.</div>'
                html += '</td>'
            }
        }
        html += '<td><div style="display:inline-block;" onmouseover="jQuery(this).find(\'div.spInst\').show();" onmouseout="jQuery(this).find(\'div.spInst\').hide();"><img style="vertical-align:middle;display:inline-block;" src="skins/default/img/other/plusOpenBtn.png" onclick="virtualPanelBuilder.onClickCreateStateIcon();"/><div class="spInst" style="z-index:30000;position:absolute;width:170px;margin-left:54px;display:none;font-size:8pt;border:solid 1px teal;background-color:lightyellow;border-radius:4px;padding-left:4px;padding-right:4px;">Click here to create a new state icon.</div></div></td>';
        html += '</tr></tbody></table>';
        return html;
    }
    function buildPanelElements(configuration) {
        dirtyPositions = [];
        var html = '<table id="vpnl_elementsContainer"><tbody><tr><td style="vertical-align:top;"><div id="vpnl_sortableParent" style="position:relative;">';
        html += '<table id="vpnl_elements" style="border:solid 1px black;margin-top:4px;border-radius-top:4px;"><thead>';
        html += '<tr style="background-color:steelblue;color:white;font-weight:bold;" class="vpnl_titleGradient"><td colspan="5" style="padding-left:2px;padding-right:2px;border-bottom:solid 3px orange;">Panel Elements</td></tr>';
        html += '<tr style="background-color:gainsboro;"><td style="width:24px;"/><td style="width:24px;"/><td style="width:200px;padding-left:4px;padding-right:4px;">Element</td><td style="width:37px;padding-left:4px;padding-right:4px;">Type</td><td style=padding-left:4px;padding-right:4px;"></td></tr></thead><tbody style="position:relative;">';
        if (configuration.elements && configuration.elements.length > 0) {
            for (var i = 0; i < configuration.elements.length; i++) {
                e = configuration.elements[i];
                html += '<tr data-ordinal="' + i + '" ' + 'title=""><td><img style="width:24px;" src="skins/default/img/scenes/create_scene_trigger_button_remove.png" onmouseover="this.src = \'skins/default/img/scenes/create_scene_trigger_button_remove_hover.png\'" onmouseout="this.src = \'skins/default/img/scenes/create_scene_trigger_button_remove.png\'" onclick="virtualPanelBuilder.onDeleteControlElement($(this).parents(\'tr:first\')[0].rowIndex - 2)" /></td>';
                if (e.controlType != 'line_break')
                    html += '<td><img style="width:22px;" src="skins/default/img/scenes/scene_edit_button_static.png" onmouseover="this.src = \'skins/default/img/scenes/scene_edit_button_hover.png\'" onmouseout="this.src = \'skins/default/img/scenes/scene_edit_button_static.png\'" onclick="virtualPanelBuilder.onEditControlElement($(this).parents(\'tr:first\')[0].rowIndex - 2)" /></td>';
                else
                    html += '<td/>';
                html += '<td style="cursor:n-resize;padding-left:2px;padding-right:2px;">' + buildControlLabel(e) + '</td>';
                html += '<td style="font-size:8pt;">' + e.controlType + '</td><td/></tr>';
            }
        }
        html += '</tbody><tfoot><tr><td><img style="width:22px;cursor:pointer;" onclick="virtualPanelBuilder.onAddControlElement();" src="skins/default/img/other/plusOpenBtn.png"/></td><td colspan="4"><span style="white-space:nowrap;font-size:8pt;font-style:italic;">Click + to add another button, slider, image ...etc. to your panel.</span></td></tr></tfoot></table>';
        html += '</div></td><td style="vertical-align:top;"><div id="vpnl_panelSamples" class="" style="position:relative;top:' + document.body.scrollTop + 'px;vertical-align:top;margin-top:10px;margin-left:7px;display:inline-block;">'
            + buildDashboardSample(configuration) + '<input class="vBtn" type="button" id="vpnl_btnSaveConfiguration" style="margin-left:10px;display:inline-block;vertical-align:top;" value="Reload Luup" onclick="virtualPanelBuilder.onReloadLuup();"/><div style="display:block;height:10px;"/>' + buildDetailSample(configuration) + '</div></td></tr></tbody></table>';
        return html;
    }
    function buildDetailSample(configuration) {
        var html = '<div id="vpnl_panelDashboardSample" class="vpnl_dialogShadow" style="display:inline-block;border:solid 1px silver;border-radius:7px;padding:10px;background-color:azure;'
            + dialogManager.createGradient('119d2d', 'EE9A00') + ';"><div class="" style="color:gray;background-color:whitesmoke;padding-left:15px;padding-right:15px;">'
            + '<div style="margin-left:20px;padding:2px;font-size:9pt;font-weight:bold;display:inline-block;">Detail Tab View</div><input id="vpnl_btnSavePositions" type="button" class="vBtn pull-right" style="display:inline-block;padding:4px;font-size:8pt;border-radius:4px;display:none;" onclick="virtualPanelBuilder.onSaveDirtyPositions();" value="Save Positions" /><input id="vpnl_btnRevertPositions" type="button" class="vBtn pull-right" style="display:inline-block;padding:4px;font-size:8pt;border-radius:4px;display:none;margin-left:4px;margin-right:4px;" onclick="virtualPanelBuilder.onRevertDirtyPositions();" value="Revert Positions" />'
            + '<div class="clearfix position_relative"><div id="cpanel_control_container" class="device_cpanel_controls" style="background-color:white;width:600px;text-align:left;height:400px;font-size:12px;border:dashed 1px silver;">';
        if (configuration.elements && configuration.elements.length > 0) {
            for (var i = 0; i < configuration.elements.length; i++) {
                ctl = configuration.elements[i];
                switch (ctl.controlType) {
                    case 'label':
                        html += '<div id="vpnl_ctlElem' + i + '" class="cpanel_device_control_label vpnl_samplecontrol vpnl_ctlElemWidth" data-controlordinal="' + i + '"  data-control_type="label" style="cursor:pointer;background-color:white;position:absolute;top:' + ctl.position.top + 'px;left:' + ctl.position.left + 'px;width:' + ctl.position.width + 'px;">'
                            + '<div class="vpnl_ctlElemWidth" style="' + textStyler.createCss(ctl.textStyle) + textStyler.createCssMargins(ctl.margins) + '"><span style="white-space:pre;">' + ctl.text + '</span></div></div>';
                        break;
                    case 'horizontal_rule':
                        html += '<div id="vpnl_ctlElem' + i + '" class="cpanel_device_control_label vpnl_samplecontrol vpnl_ctlElemWidth" data-controlordinal="' + i + '" data-control_type="label" style="cursor:pointer;background-color:white;position:absolute;top:' + ctl.position.top + 'px;left:' + ctl.position.left + 'px;width:' + ctl.position.width + 'px;">'
                            + '<hr class="vpnl_ctlElemWidth" style="border:' + ctl.lineStyle.style
                            + ' ' + translateLineWeight(ctl.lineStyle.weight)
                            + ' ' + ctl.lineStyle.color + ';' + textStyler.createCssMargins(ctl.margins)
                            + 'width:' + ctl.position.width + 'px;"/></div>'
                        break;
                    case 'button':
                        html += '<div id="vpnl_ctlElem' + i + '" class="cpanel_device_control_button vpnl_samplecontrol vpnl_ctlElemWidth vpnl_ctlElemHeight" data-control_type="button" data-controlordinal="' + i + '" style="cursor:pointer;position:absolute;text-align:center;width:' + ctl.position.width + 'px;top:' + ctl.position.top + 'px;left:' + ctl.position.left + 'px;height:' + ctl.position.height + 'px;">' + ctl.text + '</div>';
                        break;
                    case 'variable':
                        html += '<div id="vpnl_ctlElem' + i + '" class="cpanel_device_control_label vpnl_samplecontrol vpnl_ctlElemWidth" data-control_type="label" data-controlordinal="' + i + '" style="cursor:pointer;background-color:white;position:absolute;top:' + ctl.position.top + 'px;left:' + ctl.position.left + 'px;width:' + ctl.position.width + 'px;">'
                            + '<div class="vpnl_ctlElemWidth" style="' + textStyler.createCss(ctl.textStyle) + textStyler.createCssMargins(ctl.margins) + '"><span style="white-space:pre;">' +
                            textFormatter.formatVariable(ctl, api.getDeviceStateVariable(parseInt(ctl.display.deviceId, 10),
                                ctl.display.serviceId, ctl.display.variable, { dynamic: true }))
                                + '</span></div></div>';
                        break;
                    case 'multi_state_button':
                        html += '<div id="vpnl_ctlElem' + i + '" data-control_type="multi_state_button" class="on-off-device-container deviceButtonContainer width_75  vpnl_samplecontrol" data-controlordinal="' + i + '" data-min_state="0" data-max_state="1" data-state="0" data-min_label="" data-max_label="" style="cursor:pointer;background-color:white;position:absolute;top:' + ctl.position.top + 'px;left:' + ctl.position.left + 'px;">'
                             + '<div data-container="button" class="on-off-device cursor_pointer"/>'
                             + '<div data-container="label" class="stateLabel">'
                                + '<span style="white-space:pre;' + textStyler.createCss(ctl.offState.textStyle) + textStyler.createCssMargins(ctl.offState.margins) + '">' + ctl.offState.text + '</span>'
                             + '</div></div>';
                        break;
                	case 'slider':
                		var sliderval = api.getDeviceStateVariable(parseInt(ctl.display.deviceId, 10), ctl.display.serviceId, ctl.display.variable, { dynamic: true });
                		var sliderpos = (sliderval / (parseInt(ctl.values.max.value, 10) - parseInt(ctl.values.min.value, 10)) * 100);
                		if (isNaN(sliderpos))
                			sliderpos = 0;
                		html += '<div id="vpnl_ctlElem' + i + '" class="cpanel_device_control_slider vpnl_samplecontrol vpnl_ctlElemWidth" title="0" data-controlordinal="' + i + '" data-control_type="slider" style="cursor:pointer;position: absolute;top:' + ctl.position.top + 'px;left:' + ctl.position.left + 'px;width:' + ctl.position.width + 'px;height:' + ctl.position.height + 'px;">' +
							'<div class="cpanel_device_control_slider_content cursor_pointer ui-slider ui-slider-horizontal ui-widget ui-widget-content ui-corner-all" data-slider_content="slider_content" aria-disabled="false">' +
							'<div class="ui-slider-range ui-widget-header ui-corner-all ui-slider-range-min" style="width:' + sliderpos + '%;"></div><a class="ui-slider-handle ui-state-default ui-corner-all" href="javascript:void(0);" style="left: ' + sliderpos + '%;"></a></div></div>';
                        break;
                    case 'slider_vertical':
                        html += '<div id="vpnl_ctlElem' + i + '" class="cpanel_device_control_slider_vertical  vpnl_samplecontrol" data-controlordinal="' + i + '" data-control_type="slider_vertical" style="cursor:pointer;background-color:white;position:absolute;text-align:center;top:' + ctl.position.top + 'px;left:' + ctl.position.left + 'px;">'
                            + '<div class="cpanel_device_control_slider_vertical button_up cursor_pointer" data-slider_button="up"></div>'
                            + '<div class="clearfix"><div class="device_control_slider_vertical_slider_value pull_left" data-slider_value="value">0</div>'
                            + '<div class="cpanel_device_control_slider_vertical label pull_right" data-slider_value="prefix"></div></div>'
                            + '<div class="cpanel_device_control_slider_vertical button_down cursor_pointer" data-slider_button="down"></div></div>';
                        break;
                	case 'spinner_horizontal':
                		var spinnerval = api.getDeviceStateVariable(parseInt(ctl.display.deviceId, 10), ctl.display.serviceId, ctl.display.variable, { dynamic: true });
                		html += '<div id="vpnl_ctlElem' + i + '" class="device_control_spinner_horizontal vpnl_samplecontrol vpnl_ctlElemWidth vpnl_ctlElemHeight" data-controlordinal="' + i + '" data-control_type="spinner_horizontal" data-scenegroup="0" style="position:absolute;top:' + ctl.position.top + 'px;left:' + ctl.position.left + 'px;width:' + ctl.position.width + 'px;height:' + ctl.position.height + 'px;">'
							+ '<div class="clearfix spinner_horizontal_container_values" data-container="values"><span class="" data-container="value">' + spinnerval + '</span><span class="" data-container="symbol">' + ctl.values.suffix + '</span></div><div class="clearfix" data-container="controls">'
							+ '<div class="pull-left flip spinner_horizontal_arrow_down" data-container="control_down" title="' + ctl.values.min.text + '"></div>'
							+ '<div class="pull-left flip spinner_horizontal_arrow_up" data-container="control_up" title="' + ctl.values.max.text + '"></div></div></div>';
                		break;
                    case 'color_indicator':
                        html += '<div id="vpnl_ctlElem' + i + '" class="cpanel_device_control_label vpnl_samplecontrol vpnl_ctlElemWidth vpnl_ctlElemHeight" data-control_type="label" data-controlordinal="' + i + '" style="cursor:pointer;background-color:white;position:absolute;top:' + ctl.position.top + 'px;left:' + ctl.position.left + 'px;width:' + ctl.position.width + 'px;height:' + ctl.position.height + 'px;" >'
                        html += '<div class="vpnl_ctlElemWidth vpnl_ctlElemHeight" style="vertical-align:top;width:' + ctl.position.width + 'px;height:' + ctl.position.height + 'px; ' + textStyler.createCssMargins(ctl.margins) + ';display:inline-block;border:solid 2px gray;box-shadow:0px 0px 2px 2px rgba(50, 50, 50, 0.25);border-radius:' + ctl.position.width + 'px;background-color:' + evaluateColorExpressions(api.getDeviceStateVariable(parseInt(ctl.display.deviceId, 10), ctl.display.serviceId, ctl.display.variable, { dynamic: true }), ctl.defaultColor, ctl.expressions) + ';padding-left:1px;"/></div>';
                        break;

                }
            }
        }
        html += '</div><div id="vpnl_rubberbox" style="cursor:move;border:dotted 1px black;position:absolute;width:50px;height:20px;top:0px;left:0px;z-index:100000;display:none;"/></div></div>'
        return html;
    }
    function buildDashboardSample(configuration) {
        var cgid = 1
        var sgid = 1
        var html = '<div id="vpnl_panelDashboardSample" class="vpnl_dialogShadow" style="display:inline-block;border:solid 1px silver;border-radius:7px;padding:10px;background-color:azure;'
            + dialogManager.createGradient('119d2d', 'EE9A00') + ';"><div class="col-md-6 device_view_device_wrapper genericDevice generalDevice visible dashboardTypeLeftBorderSeparator" style="display:inline-block;color:gray;background-color:whitesmoke;width:100%;margin-bottom:0px;margin-top:0px;">'
            + '<div style="margin-left:20px;padding:2px;font-size:9pt;font-weight:bold;">Dashboard View</div>'
            + '<div class="clearfix"><div class="custom_device_device_details_wrapper full_width" style="background-color:white;width:350px;text-align:center;display:inline-block;border:dashed 1px silver;">';
        html += '<div id="device_id_' + configuration.deviceId + '_details">'
            + '<div id="device_id_' + configuration.deviceId + '_sg_' + sgid + '" class="device_scene_group">'
            + '<div id="device_id_' + configuration.deviceId + '_cg_' + cgid + '" class="device_control_group">';
        if (configuration.elements && configuration.elements.length > 0) {
            for (var i = 0; i < configuration.elements.length; i++) {
                ctl = configuration.elements[i];
                if (ctl.position && ctl.position.hideOnDashboard) continue;
                switch (ctl.controlType) {
                    case 'group_start':
                        cgid++;
                        if (typeof (ctl.isSceneGroup) != 'undefined' && ctl.isSceneGroup) {
                            sgid++;
                            html += '</div></div>'
                                + '<div id="device_id_' + configuration.deviceId + '_sg_' + sgid + '" class="device_scene_group">';
                            if (typeof (ctl.lineBreak) != 'undefined' && ctl.lineBreak) {
                                sgid++;
                                html += '<div id="device_id_' + configuration.deviceId + '_cg_' + cgid + '" class="device_control_group">'
                                    + '<div class="device_control_spacer pull-left" data-control_type="spacer" style="width: 300px;"/>'
                                    + '<div class="device_control_spacer pull-left" data-control_type="spacer" style="width: 300px;"/></div>';
                                html += '</div><div id="device_id_' + configuration.deviceId + '_sg_' + sgid + '" class="device_scene_group pull-left">';
                                cgid++;
                            }
                        }
                        else
                            html += '</div>';
                        html += '<div id="device_id_' + configuration.deviceId + '_cg_' + cgid + '" class="device_control_group">'
                        break;
                    case 'line_break':
                        html += '<div class="clearfix" data-control_type="line_break"/>';
                        break;
                    case 'label':
                        html += '<div class="device_control control_label pull-left" data-control_type="label">'
                            + '<div style="' + textStyler.createCss(ctl.textStyle) + textStyler.createCssMargins(ctl.margins) + '"><span style="white-space:pre;">' + ctl.text + '</span></div></div>';
                        break;
                    case 'horizontal_rule':
                        html += '<div class="device_control control_label pull-left" data-control_type="label">'
                            + '<hr style="border:' + ctl.lineStyle.style
                            + ' ' + translateLineWeight(ctl.lineStyle.weight)
                            + ' ' + ctl.lineStyle.color + ';' + textStyler.createCssMargins(ctl.margins)
                            + 'width:' + ctl.position.width + 'px;"/></div>'
                        break;
                    case 'button':
                        html += '<div class="device_control button pull-left" data-control_type="button" style="width:' + ctl.position.width + 'px">' + ctl.text + '</div>';
                        break;
                    case 'variable':
                        html += '<div class="device_control control_label pull-left" data-control_type="label">'
                            + '<div style="' + textStyler.createCss(ctl.textStyle) + textStyler.createCssMargins(ctl.margins) + '"><span style="white-space:pre;">' +
                            textFormatter.formatVariable(ctl, api.getDeviceStateVariable(parseInt(ctl.display.deviceId, 10),
                                ctl.display.serviceId, ctl.display.variable, { dynamic: true }))
                                + '</span></div></div>';
                        break;
                    case 'multi_state_button':
                        html += '<div data-control_type="multi_state_button" class="on-off-device-container deviceButtonContainer width_75 pull-left" data-min_state="0" data-max_state="1" data-state="0" data-min_label="" data-max_label="">'
                             + '<div data-container="button" class="on-off-device cursor_pointer"/>'
                             + '<div data-container="label" class="stateLabel">'
                                + '<span style="white-space:pre;' + textStyler.createCss(ctl.offState.textStyle) + textStyler.createCssMargins(ctl.offState.margins) + '">' + ctl.offState.text + '</span>'
                             + '</div></div>';
                        break;
                    case 'spacer':
                        html += '<div class="device_control_spacer pull-left" data-control_type="spacer" style="width:' + (ctl.position.width * 16) + 'px;"></div>';
                        break;
                	case 'slider':
                		var sliderval = api.getDeviceStateVariable(parseInt(ctl.display.deviceId, 10), ctl.display.serviceId, ctl.display.variable, { dynamic: true });
                		var sliderpos = (sliderval / (parseInt(ctl.values.max.value, 10) - parseInt(ctl.values.min.value, 10)) * 100);
                		if (isNaN(sliderpos))
                			sliderpos = 0;
                        html += '<div class="device_control slider pull-left" title="0" data-control_type="slider" style="width:' + ctl.position.width + 'px;">'
                            + '<div class="device_control slider slider_content ui-slider ui-slider-horizontal ui-widget ui-widget-content ui-corner-all" data-slider_content="slider_content" aria-disabled="false">'
                            + '<div class="ui-slider-range ui-widget-header ui-corner-all ui-slider-range-min" style="width:' + sliderpos + '%;"></div><a class="ui-slider-handle ui-state-default ui-corner-all" href="javascript:void(0);" style="left:' + sliderpos + '%;"></a>'
                            + '</div><div class="slider_label float_left" data-slider_label="label">' + sliderval
							+ (typeof (ctl.values.suffix) == 'undefined' ? '%' : ctl.values.suffix) + '</div><div class="breaker"></div></div>';
                        break;
                    case 'slider_vertical':
                        html += '<div class="cpanel_device_control_slider_vertical pull-left" data-control_type="slider_vertical">'
                            + '<div class="cpanel_device_control_slider_vertical button_up cursor_pointer" data-slider_button="up"></div>'
                            + '<div class="clearfix"><div class="device_control_slider_vertical_slider_value pull_left" data-slider_value="value">' + api.getDeviceStateVariable(parseInt(ctl.display.deviceId, 10),
                                ctl.display.serviceId, ctl.display.variable, { dynamic: true }) + '</div>'
                            + '<div class="cpanel_device_control_slider_vertical label pull_right" data-slider_value="prefix"></div></div>'
                            + '<div class="cpanel_device_control_slider_vertical button_down cursor_pointer" data-slider_button="down"></div></div>';
                        break;
                	case 'spinner_horizontal':
                		var spinnerval = api.getDeviceStateVariable(parseInt(ctl.display.deviceId, 10), ctl.display.serviceId, ctl.display.variable, { dynamic: true });
                		html += '<div class="device_control_spinner_horizontal pull-left" data-control_type="spinner_horizontal" data-scenegroup="1">'
							+ '<div class="clearfix spinner_horizontal_container_values" data-container="values">'
							+ '<span class="" data-container="value">' + spinnerval + '</span><span class="" data-container="symbol">' + (typeof (ctl.values.suffix) == 'undefined' ? '' : ctl.values.suffix) + '</span></div>'
							+ '<div class="clearfix" data-container="controls"><div class="pull-left flip spinner_horizontal_arrow_down" data-container="control_down"' + 'title="' + ctl.values.min.text + '"></div>'
							+ '<div class="pull-left flip spinner_horizontal_arrow_up" data-container="control_up" title="' + ctl.values.max.text + '"></div></div></div>';
                		break;
                    case 'color_indicator':
                        html += '<div class="device_control control_label pull-left" data-control_type="label">';
                        html += '<div style="vertical-align:top;width:' + ctl.position.width + 'px;height:' + ctl.position.height + 'px;' + textStyler.createCssMargins(ctl.margins) + ';display:inline-block;border:solid 2px gray;box-shadow:0px 0px 2px 2px rgba(50, 50, 50, 0.25);border-radius:' + ctl.position.width + 'px;background-color:' + evaluateColorExpressions(api.getDeviceStateVariable(parseInt(ctl.display.deviceId, 10), ctl.display.serviceId, ctl.display.variable, { dynamic: true }), ctl.defaultColor, ctl.expressions) + ';padding-left:1px;"/></div>';
                        break;

                }
            }
        }
        // End Scene/Control group
        html += '</div></div>';
        html += '</div></div></div></div></div>'
        return html;
    }
    function evaluateColorExpressions(value, defaultColor, expressions) {
        var sColor = defaultColor;
        for (var i = 0; i < expressions.length; i++) {
            if (eval('"' + value + '" ' + expressions[i].operator + ' "' + expressions[i].value + '"')) {
                sColor = expressions[i].color;
                break;
            }
        }
        return sColor;
    }
    function buildControlText(ctl) {
        switch (ctl.controlType) {
            case 'line_break':
                return "Line Break";
            case 'group_start':
                return "Group Start";
            case 'spacer':
                return "Spacer";
            case 'multi_state_button':
                return ctl.onState.text + '/' + ctl.offState.text;
            case 'label':
            case 'button':
                return ctl.text;
            case 'variable':
                return ctl.display.variable;
            default:
                return ctl.controlType;
        }

    }
    function buildControlLabel(ctl) {
        switch(ctl.controlType) {
            case 'group_start':
                var ctlText = 'GROUP START';
                if (typeof (ctl.lineBreak) != 'undefined' && ctl.lineBreak)
                    ctlText = 'GROUP START/LINE BREAK';
                if (typeof (ctl.isSceneGroup) != 'undefined' && ctl.isSceneGroup)
                    return '<div style="display:inline-block;background-color:#ff6000;width:100%;text-align:center;font-size:6pt;color:white;padding-top:1px;border-radius:5px;font-weight:bold;">SCENE ' + ctlText + '</span>';
                return '<div style="display:inline-block;background-color:orange;width:100%;text-align:center;font-size:6pt;color:white;padding-top:1px;border-radius:5px;font-weight:bold;">' + ctlText + '</span>';
            case 'line_break':
                return '<div style="display:inline-block;background-color:green;width:100%;text-align:center;font-size:6pt;color:white;padding-top:1px;border-radius:5px;font-weight:bold;">LINE BREAK</span>'
            case 'multi_state_button':
                if (typeof (ctl.onState.textStyle) == 'undefined')
                    ctl.onState.textStyle = { italic: false, bold: false, underline: false, fontFamily: '', fontSize: '1em', color: '' }
                if (typeof (ctl.offState.textStyle) == 'undefined')
                    ctl.offState.textStyle = { italic: false, bold: false, underline: false, fontFamily: '', fontSize: '1em', color: '' }
                return '<div class="on-off-device cursor_pointer" style="background-size:34px;height:20px;width:32px;display:inline-block;vertical-align:middle;"></div><div style="display:inline-block;"><span style="' + textStyler.createCss(ctl.onState.textStyle) + '">' + ctl.onState.text + '</span>/<span style="'
                    + textStyler.createCss(ctl.offState.textStyle) + '">' + ctl.offState.text + '</span></div>';
            case 'label':
                return '<div style="display:inline-block;font-size:14px;font-family:OpenSansLight;"><span style="display:inline-block;border:solid 1px silver;' + textStyler.createCss(ctl.textStyle) + '">' + ctl.text + '</span></div>';
            case 'horizontal_rule':
                if (typeof (ctl.lineStyle) == 'undefined')
                    ctl.lineStyle = { color: '', weight: 'fine', style: 'solid' }
                return '<div class="" style="height:14px;border:solid 1px gainsboro;border-radius:5px;overflow-y:hidden;"><div style="margin:0px auto;display:block;position:relative;top:3px;overflow-x:hidden;width:'
                    + Math.min(ctl.position.width, 190) + 'px;">'
                    + '<hr style="border:'
                    + ctl.lineStyle.style
                    + ' ' + translateLineWeight(ctl.lineStyle.weight)
                    + ' ' + ctl.lineStyle.color
                    + ';width:' + ctl.position.width
                    + 'px;margin:0px;"'
                    + '"/></div></div>';
            case 'variable':
                return '<div style="display:inline-block;font-size:14px;font-family:OpenSansLight;"><span style="' + textStyler.createCss(ctl.textStyle) + '">' + textStyler.transformVariableFormat(ctl.textStyle, ctl.display.variable) + '</span></div>';
            case 'spacer':
            	return '<span style="font-size:8pt;">Spacer [' + ctl.position.width + ']</span>';
        	case 'spinner_horizontal':
        		return '<div class="pull-left spinner_horizontal_arrow_down" style="width:15px;height:15px;background-size:15px 15px;display:inline-block;"/>'
					+ '<div class="pull-left spinner_horizontal_arrow_up" style="width:15px;height:15px;background-size:15px 15px;display:inline-block;"/><span>' + ctl.text + '</span>';
            default:
                return ctl.text;
        }
        return '';
    }
    function buildTriggerConfig(configuration) {
        var html = '<table id="vpnl_triggers" style="border:solid 1px black;margin-top:4px;width:100%;"><thead>';
        html += '<tr style="background-color:steelblue;color:white;font-weight:bold;" class="vpnl_titleGradient"><td colspan="5" style="padding-left:2px;padding-right:2px;border-bottom:solid 3px orange;">Defined Triggers</td></tr>';
        html += '<tr style="background-color:gainsboro;"><td style="width:24px;"/><td style="width:24px;"/><td style="width:200px;padding-left:4px;padding-right:4px;">Trigger</td><td style="width:37px;padding-left:4px;padding-right:4px;">Conditions</td><td style=padding-left:4px;padding-right:4px;"></td></tr></thead><tbody style="position:relative;">';
        if (configuration.triggers && configuration.triggers.length > 0) {
            for (var i = 0; i < configuration.triggers.length; i++) {
                t = configuration.triggers[i];
                html += '<tr data-ordinal="' + i + '" ' + 'title=""><td><img style="width:24px;" src="skins/default/img/scenes/create_scene_trigger_button_remove.png" onmouseover="this.src = \'skins/default/img/scenes/create_scene_trigger_button_remove_hover.png\'" onmouseout="this.src = \'skins/default/img/scenes/create_scene_trigger_button_remove.png\'" onclick="virtualPanelBuilder.onDeleteTrigger($(this).parents(\'tr:first\')[0].rowIndex - 2)" /></td>';
                html += '<td><img style="width:22px;" src="skins/default/img/scenes/scene_edit_button_static.png" onmouseover="this.src = \'skins/default/img/scenes/scene_edit_button_hover.png\'" onmouseout="this.src = \'skins/default/img/scenes/scene_edit_button_static.png\'" onclick="virtualPanelBuilder.onEditTrigger($(this).parents(\'tr:first\')[0].rowIndex - 2)" /></td>';
                html += '<td style="cursor:n-resize;padding-left:2px;padding-right:2px;">' + buildTriggerLabel(t) + '</td>';
                html += '<td/></tr>';
            }
        }
        html += '</tbody><tfoot><tr><td><img style="width:22px;cursor:pointer;" onclick="virtualPanelBuilder.onAddTrigger();" src="skins/default/img/other/plusOpenBtn.png"/></td><td colspan="4"><span style="white-space:nowrap;font-size:8pt;font-style:italic;">Click + to add another trigger. to your panel.</span></td></tr></tfoot></table>';
        return html;
    }
    function buildStateConfig(configuration) {
        var html = '<table id="vpnl_states" style="border:solid 1px black;margin-top:4px;width:100%;"><thead>';
        html += '<tr style="background-color:steelblue;color:white;font-weight:bold;" class="vpnl_titleGradient"><td colspan="5" style="padding-left:2px;padding-right:2px;border-bottom:solid 3px orange;">Defined States</td></tr>';
        html += '<tr style="background-color:gainsboro;"><td style="width:24px;"/><td style="width:24px;"/><td style="width:200px;padding-left:4px;padding-right:4px;">State Variable</td><td style="width:37px;padding-left:4px;padding-right:4px;">Conditions</td><td style=padding-left:4px;padding-right:4px;"></td></tr></thead><tbody style="position:relative;">';
        if (configuration.states && configuration.states.length > 0) {
            for (var i = 0; i < configuration.triggers.length; i++) {
                t = configuration.triggers[i];
                html += '<tr data-ordinal="' + i + '" ' + 'title=""><td><img style="width:24px;" src="skins/default/img/scenes/create_scene_trigger_button_remove.png" onmouseover="this.src = \'skins/default/img/scenes/create_scene_trigger_button_remove_hover.png\'" onmouseout="this.src = \'skins/default/img/scenes/create_scene_trigger_button_remove.png\'" onclick="virtualPanelBuilder.onDeleteStae($(this).parents(\'tr:first\')[0].rowIndex - 2)" /></td>';
                html += '<td><img style="width:22px;" src="skins/default/img/scenes/scene_edit_button_static.png" onmouseover="this.src = \'skins/default/img/scenes/scene_edit_button_hover.png\'" onmouseout="this.src = \'skins/default/img/scenes/scene_edit_button_static.png\'" onclick="virtualPanelBuilder.onEditState($(this).parents(\'tr:first\')[0].rowIndex - 2)" /></td>';
                html += '<td style="cursor:n-resize;padding-left:2px;padding-right:2px;">' + buildStateVariableLabel(t) + '</td>';
                html += '<td/></tr>';
            }
        }
        html += '</tbody><tfoot><tr><td><img style="width:22px;cursor:pointer;" onclick="virtualPanelBuilder.onAddState();" src="skins/default/img/other/plusOpenBtn.png"/></td><td colspan="4"><span style="white-space:nowrap;font-size:8pt;font-style:italic;">Click + to add another state variable. to your panel.</span></td></tr></tfoot></table>';
        return html;
    }

    /********************************************************
     * Element Templates
     ********************************************************/
    function buildEditButtonElementTemplate(elementState) {
        var html = '<div style="">';
        if (typeof (elementState.onState) == 'undefined')
            elementState.onState = {
                command: {
                    cmdType: "SetVariable",
                    action: { serviceId: "", deviceId: elementState.display.deviceId, action: "", parameters: [] },
                    scene: { sceneId: "" },
                    variable: { serviceId: "", variable: "", value: "", deviceId: elementState.display.deviceId },
                    http: { username: "", url: "", password: "" }
                }
            }
        if (typeof (elementState.offState) == 'undefined')
            elementState.offState = {
                command: {
                    cmdType: "SetVariable",
                    action: { serviceId: "", deviceId: elementState.display.deviceId, action: "", parameters: [] },
                    scene: { sceneId: "" },
                    variable: { serviceId: "", variable: "", value: "", deviceId: elementState.display.deviceId },
                    http: { username: "", url: "", password: "" }
                }
            }
        if (typeof (elementState.display.operator) == 'undefined') elementState.display.operator = '==';
        if (typeof (elementState.display.maxValue) == 'undefined') elementState.display.maxValue = '';
        html += '<div>';
        html += '<table style="width:100%;"><tbody>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Style</td>';
        html += '<td><select data-bind=".buttonStyle" class="vpnl_editField" style="" onchange="virtualPanelBuilder.onButtonStyleChanged(this, this.value);">' +
            '<option value="momentary" ' + (!elementState || !elementState.buttonStyle || elementState.buttonStyle == 'momentary' ? 'selected="selected"' : '') + '>Momentary</option>' +
            '<option value="persistent" ' + (elementState && elementState.buttonStyle == 'persistent' ? 'selected="selected"' : '') + '>Persistent</option>' +
            '<option value="toggle" ' + (elementState && elementState.buttonStyle == 'toggle' ? 'selected="selected"' : '') + '>Toggle</option>' +
            '<option value="persistenttoggle" ' + (elementState && elementState.buttonStyle == 'persistenttoggle' ? 'selected="selected"' : '') + '>Persistent Toggle</option>' +
            '</select></td><td/>';
        html += '</tr>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Text</td>';
        html += '<td style="padding-left:2px;padding-right:2px;padding-bottom:4px;"><input data-bind=".text" class="vpnl_editField" type="text" value="' + elementState.text + '" onchange="textStyler.updateSampleText(\'vpnl_buttonText\', this.value);"/>'
            // Vera tries to copy the styles into the title for the button which screws all this up.
            + '<div style="display:none;">' + textStyler.createHtml('vpnl_buttonText', '.textStyle', elementState.textStyle, elementState.text) + '</div>'
            + '</td><td/></tr>';

        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Size &amp; Position <span style="font-size:7pt;font-style:italic;">(Left/Top values for Detail Tab only)</span></td></tr>';
        html += '<tr><td colspan="3">' +
            buildElementPositionTemplate('.position', 
            elementState.position) + '</td></tr>';

        html += '<tr class="sizepos displayvar vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;' + (elementState.buttonStyle.substring(0, 10)  != 'persistent' ? 'display:none;' : '') + '"><td colspan="3" style="padding-left:4px;">Device Variable to Monitor <span style="font-size:8pt;font-style:italic">(used to set the clicked state of the button)</span></td></tr>';
        html += '<tr class="sizepos displayvar" style="' + (elementState.buttonStyle.substring(0, 10) != 'persistent' ? 'display:none;' : '') + '">';

        html += '<td colspan="3" style="padding-bottom:4px;">' + buildDeviceVariable('.display',
            elementState.display.deviceId,
            elementState.display.serviceId,
            elementState.display.variable, elementState.display.operator, elementState.display.value, elementState.display.maxValue) + '</td></tr>';

        html += '<tr class="btncmd_single" style="display:' + (elementState.buttonStyle == 'toggle' || elementState.buttonStyle == 'persistenttoggle' ? 'none' : '') + '"><td colspan="3"><table style="width:100%;"><tbody>';
        html += '<tr class="btncmd vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td style="padding-left:4px;">Command <span style="font-size:7pt;font-style:italic">(action to be performed when the button is clicked)</span></td></tr>';
        html += '<td>' + buildActionCommand(elementState.ordinal, '.command', elementState.command) + '</td></tr></tbody></table></tr>';

        html += '<tr class="btncmd_multi" style="display:' + (elementState.buttonStyle == 'toggle' || elementState.buttonStyle == 'persistenttoggle' ? '' : 'none') + '"><td colspan="3">';
        html += '<table style="width:100%"><tbody><tr class="btncmd vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td style="width:50%;padding-left:4px;">Command <span style="font-size:7pt;font-style:italic">(action to be performed for ON state)</span></td>';
        html += '<td style="padding-left:4px;">Command <span style="font-size:8pt;font-style:italic">(action to be performed for OFF state)</span></td></tr>'
        html += '<tr><td style="vertical-align:top;">' + buildActionCommand(elementState.ordinal, '.onState.command', elementState.onState.command) + '</td>';
        html += '<td style="vertical-align:top;">' + buildActionCommand(elementState.ordinal, '.offState.command', elementState.offState.command) + '</td></tr></tbody></table></td>';

        html += '</tr></tbody></table></div>'
        html += '</div>';
        return html;
    }
    function buildEditLabelElementTemplate(elementState) {
        var html = '<div style="">';
        if (typeof (elementState.margins) == 'undefined')
            elementState.margins = { left: '0', top: '0', right: '0', bottom: '0' }
        if (typeof (elementState.textStyle) == 'undefined')
            elementState.textStyle = { fontFamily: '', fontSize: '1em', color: '', bold: false, italic: false, underline: false, textAlign:'left' }
        html += '<table style="width:100%;"><tbody>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Text</td>';
        html += '<td style="padding-left:2px;padding-right:2px;padding-bottom:4px;"><input data-bind=".text" class="vpnl_editField" onchange="textStyler.updateSampleText(\'vpnl_labelText\', this.value);" type="text" value="' + elementState.text + '"/>' 
            + textStyler.createHtml('vpnl_labelText', '.textStyle', elementState.textStyle, elementState.text, { showAlignment: true }) + '</td><td/></tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Size &amp; Position <span style="font-size:7pt;font-style:italic;">(Left/Top values for Detail Tab only)</span></td></tr>';
        html += '<tr><td colspan="3">' +
            buildElementPositionTemplate('.position', 
            elementState.position) + '</td></tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Margins</td></tr>';
        html += '<tr><td colspan="3">' +
            buildElementMarginTemplate('.margins',
            elementState.margins.left, elementState.margins.top,
            elementState.margins.right, elementState.margins.bottom, null, null) + '</td></tr>';
        if (typeof (elementState.margins) == 'undefined')
            elementState.margins = { left: '0', top: '0', right: '0', bottom: '0' }

        html += '</tbody></table></div>';

        return html;
    }
    function buildEditSpacerElementTemplate(elementState) {
        var html = '<div style="">';
        html += '<table style="width:100%;"><tbody>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Width</td>';
        html += '<td style="padding-left:2px;padding-right:2px;padding-bottom:4px;"><input style="width:34px;" data-bind=".position.width" class="vpnl_editField" value="' + elementState.position.width + '"/>'
              + '</td><td/></tr>';
        html += '<tr><td colspan="2" style="padding:7px;">'
            + '<div class="vpnl_textStylerShadow" style="padding:4px;background-color:lightyellow;font-size:7pt;margin:0px auto;border-radius:5px;border:solid 1px gray;">Enter a value between <b>0</b> and <b>10</b>.  This will translate to roughly the percentage of width for the dashboard panel.  Larger numbers create a greater horizontal space while smaller numbers create a smaller horizontal space.</div>'

        html += '</td></tr></tbody></table></div>';
        return html;
    }
    function buildEditGroupStartElementTemplate(elementState) {
        var html = '<div style="">';
        html += '<table style="width:100%;"><tbody>';
        html += '<tr><td style="padding-left:2px;padding-right:2px;padding-bottom:4px;"><input type="checkbox" style="" data-bind=".isSceneGroup" class="vpnl_editField"'
            + ((typeof (elementState.isSceneGroup) == 'undefined' || !elementState.isSceneGroup) ? '' : 'checked="checked"') + '/><span style="vertical-align:top;">Is Scene Group</span>'
            + '</td><td/></tr>';
        html += '<tr><td style="padding-left:2px;padding-right:2px;padding-bottom:4px;"><input type="checkbox" style="" data-bind=".lineBreak" class="vpnl_editField"'
            + ((typeof (elementState.lineBreak) == 'undefined' || !elementState.lineBreak) ? '' : 'checked="checked"') + '/><span style="vertical-align:top;">Start a new Line</span>'
            + '</td><td/></tr>';

        html += '<tr><td colspan="2" style="padding:7px;">'
            + '<div class="vpnl_textStylerShadow" style="padding:4px;background-color:lightyellow;font-size:7pt;margin:0px auto;border-radius:5px;border:solid 1px gray;">Checking <i>Is Scene Group</i> will add margins to the top and bottom of the grouping.  Checking <i>Start a new line</i> will force the group to start below the previous element.</div>'
        html += '</td></tr></tbody></table></div>';
        return html;
    }
    function buildEditColorIndicatorElementTemplate(elementState) {
        var html = '<div style="">';
        if (typeof (elementState.margins) == 'undefined')
            elementState.margins = { left: '0', top: '0', right: '0', bottom: '0' }
        html += '<table style="width:100%;"><tbody>';
        html += '<tr style="display:none"><td style="width:77px;padding-left:2px;padding-right:2px;">Text</td>';
        html += '<td style="padding-left:2px;padding-right:2px;"><input data-bind=".text" style="border-radius:4px;" type="text" value="' + elementState.display.variable + '"/></td><td/></tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Variable to Monitor</td></tr>';
        html += '<td colspan="3" style="padding-bottom:4px;">' + buildDeviceVariable('.display',
            elementState.display.deviceId,
            elementState.display.serviceId,
            elementState.display.variable, null, null) + '</td></tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Size &amp; Position</td></tr>';
        html += '<tr><td colspan="3">' +
            buildElementPositionTemplate('.position',
            elementState.position) + '</td></tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Margins</td></tr>';
        html += '<tr><td colspan="3">' +
            buildElementMarginTemplate('.margins',
            elementState.margins.left, elementState.margins.top,
            elementState.margins.right, elementState.margins.bottom, null, null) + '</td></tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Conditions &amp; Colors</td></tr>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Default Color</td>'
            + '<td style="padding-left:2px;padding-right:2px;">'
                + '<input type="text" data-bind=".defaultColor" class="vpnl_editField" list="vpnl_colordatalist" value="' + elementState.defaultColor + '">'
                + '<datalist id="vpnl_colordatalist">';
                for (var clr in textStyler.colorList) {
                    html += '<option value="' + clr + '" style="color:' + clr + ';">' + textStyler.colorList[clr] + '</option>';
                }
        html += '</datalist></input>';
        html += '<div style="float:right;display:inline-block;padding-top:2px;">' + dialogManager.createTextPopupButton('vpnl_expressionInfo', 'Expression Information') + "</div>"
             + '</td><td/></tr>';
        html += '<tr><td colspan="3" style="padding-top:4px;">' 
        html += '<table id="vpnl_colorExpressions" style="margin:0px auto;border:solid 1px gray;"><tbody>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="4" style="padding-left:2px;padding-right:2px;border-bottom:solid 3px orange;">Expression <span style="font-style:italic;font-size:9pt;"></span></td></tr>';
        html += '<tr style="background-color:gainsboro;"><td style="width:24px;" /><td style="width:30px;padding-left:4px;padding-right:4px;">Operator</td><td style=padding-left:4px;padding-right:4px;">Value</td><td style=padding-left:4px;padding-right:4px;">Color</td>';
        if (elementState && elementState.expressions && elementState.expressions.length > 0) {
            for (var i = 0; i < elementState.expressions.length; i++)
                html += buildColorExpression(elementState.expressions[i].operator, elementState.expressions[i].value, elementState.expressions[i].color, i);
        }
        html += '<tr><td><img style="width:22px;cursor:pointer;" onclick="virtualPanelBuilder.addColorExpressionRow($(this).parents(\'#vpnl_colorExpressions\'));" src="skins/default/img/other/plusOpenBtn.png"/></td><td colspan="3"><span style="white-space:nowrap;font-size:8pt;font-style:italic;">Click + to add another condition to the expression.</span></td>';
        html += '</tbody></table></div>';
            + '<td></tr>';
        html += '</tbody></table>'
        html += '</div>'
        return html;
    }

    function buildEditVariableElementTemplate(elementState) {
        var html = '<div style="">';
        if (typeof (elementState.margins) == 'undefined')
            elementState.margins = { left: '0', top: '0', right: '0', bottom: '0' }
        if (typeof (elementState.textStyle) == 'undefined')
            elementState.textStyle = { fontFamily: '', fontSize: '1em', color: '', bold: false, italic: false, underline: false, format:'{0}', textAlign:'left' }
        html += '<table style="width:100%;"><tbody>';
        html += '<tr style="display:none"><td style="width:77px;padding-left:2px;padding-right:2px;">Text</td>';
        html += '<td style="padding-left:2px;padding-right:2px;"><input data-bind=".text" style="border-radius:4px;" type="text" value="' + elementState.display.variable + '"/></td><td/></tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Variable to Display</td></tr>';
        html += '<td colspan="3" style="padding-bottom:4px;">' + buildDeviceVariable('.display',
            elementState.display.deviceId,
            elementState.display.serviceId,
            elementState.display.variable, null, null) + '</td></tr>';
        html += '<tr><td colspan="3"><hr style="margin-top:1px;margin-bottom:1px;border:solid 2px darkgreen;" /></td></tr>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;vertical-align:top;">Format</td>';
        html += '<td style="padding-left:2px;padding-right:2px;padding-bottom:4px;">'
            + '<input data-bind=".textStyle.format" list="vpnl_formatDataList" style="width:277px;display:inline-block;" class="vpnl_editField vpnl_variableFormat" type="text" value="'
            + elementState.textStyle.format + '">'
            + '<datalist id="vpnl_formatDataList">'
            + '<option value="{0}">General</option>'
            + '<option value="{0:DateTime,%m/%d/%Y %I:%M%p}">Date Time (US)</option>'
            + '<option value="{0:DateTime,%x %X}">Date Time (locale)</option>'
            + '<option value="{0:Number,%.0f}">Integer</option>'
            + '<option value="{0:Number,%.1f}">1 Decimal Point</option>'
            + '<option value="{0:Number,%.2f}">2 Decimal Points</option>'
            + '<option value="{0:Percent,%.1f}%">Decimal Percentage</option>'
            + '<option value="{0:Percent,%.0f}%">Integer Percentage</option>'
            + '<option value="{0}&deg;">Degrees</option>'
            + '<option value="{0}&deg;F">Degrees F</option>'
            + '<option value="{0}&deg;C">Degrees C</option>'
            + '<option value="&quot;{0}&quot;">Quoted</option>'
            + '<option value="&bull;{0}">Circle Bullet</option>'
            + '<option value="&raquo;{0}">Right Angle</option>'
            + '<option value="&laquo;{0}">Left Angle</option>'
            + '<option value="${0}">Dollars</option>'
            + '<option value="&pound;{0}">Pounds</option>'
            + '<option value="&yen;{0}">Yen</option>'
            + '<option value="&euro;{0}">Euro</option>'
            + '<option value="{0}&copy;">Copyright</option>'
            + '</datalist>'
            + '</input>' + dialogManager.createTextPopupButton('vpnl_variableFormatInfo', 'Formatting Help')
            + textStyler.createHtml('vpnl_labelText', '.textStyle', elementState.textStyle, 
                textStyler.transformVariableFormat(elementState.textStyle, elementState.display.variable), { showAlignment: true }) + '</td><td/></tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Size &amp; Position</td></tr>';
        html += '<tr><td colspan="3">' +
            buildElementPositionTemplate('.position', 
            elementState.position) + '</td></tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Margins</td></tr>';
        html += '<tr><td colspan="3">' +
            buildElementMarginTemplate('.margins',
            elementState.margins.left, elementState.margins.top,
            elementState.margins.right, elementState.margins.bottom, null, null) + '</td></tr>';

        html += '</tbody></table></div>'
        return html;
    }
    function buildVariableFormatInfo() {
        var html = '<div style="width:375px;font-size:9pt;"><div style="font-weight:bold;font-size:12pt;">Format Syntax:<span> {0</span><span style="font-weight:normal;">[:DataType,FormatSpecifiers]</span>}</span><hr style="margin:0px;"/></div>'
            + '<div>'
            + '<div><span style="font-weight:bold;">DataType: </span>DateTime | Number | Percent</span></div>'
            + '<p style="font-size:8pt;">The data type can be either DateTime, Number, or Percent.  This will tell the formatter how it should handle the format specifiers.</p>'
            + '<div style="padding-left:7px;"><div><span style="font-weight:bold;">DateTime Format Specifiers (partial reference):</span>'
                + '<div style="font-size:7pt">* for a complete list google strftime</div>'
                + '<ul style="font-size:8pt;">'
                + '<li><span style="font-weight:bold;margin-right:4px;">%a/%A</span>Abbreviated/Full weekday name</li>'
                + '<li><span style="font-weight:bold;margin-right:4px;">%b/%B</span>Abbreviated/Full month name</li>'
                + '<li><span style="font-weight:bold;margin-right:4px;">%m</span>Month as number</li>'
                + '<li><span style="font-weight:bold;margin-right:4px;">%d</span>Day as number</li>'
                + '<li><span style="font-weight:bold;margin-right:4px;">%Y/%y</span>Year with/without century</li>'
                + '<li><span style="font-weight:bold;margin-right:4px;">%H/%h</span>Hour 24hr clock/12hr clock</li>'
                + '<li><span style="font-weight:bold;margin-right:4px;">%M</span>Minute as number</li>'
                + '<li><span style="font-weight:bold;margin-right:4px;">%S</span>Seconds as number</li>'
                + '<li><span style="font-weight:bold;margin-right:4px;">%p</span>AM or PM</li>'
                + '</ul></div>'
            + '</div>'
            + '<div style="padding-left:7px;"><div><span style="font-weight:bold;">Number Format Specifiers (partial reference):</span>'
                + '<div style="font-size:7pt">* for a complete list google printf format specifiers</div>'
                + '<ul style="font-size:8pt;">'
                + '<li><span style="font-weight:bold;margin-right:4px;">%i</span>Show as an integer</li>'
                + '<li><span style="font-weight:bold;margin-right:4px;">%.2f</span>Show as a decimal to 2 decimal places where the 2 represents the number of decimals to display</li>'
                + '</ul></div>'
            + '<div style="padding-left:7px;"><div><span style="font-weight:bold;">Percent Format Specifiers:</span>'
                + '<div style="font-size:7pt">Percentage formats are used to normalize percentage values.  Typically .01 represents 1% this Datatype will normalize this value and use format specifiers described for Number values.</div>'
                + '</div>'
            + '</div>';

        return html;
    }
    function buildLuaContextInfo() {
        var html = '<div style="width:600px;font-size:9pt;"><div style="font-weight:bold;font-size:12pt;">Lua context variable:<span></div>'
            + '<div>'
            + '<div><span style="font-weight:bold;">A table named context is passed into your lua code.  This table contains information related to the panel and action you are performing</span></div>'
            + '<p style="font-size:8pt;">context will be initialized and available as a local table during the execution of the code you enter.</p>'
            + '<div style="padding-left:7px;"><div><span style="font-weight:bold;">context variable members:</span>'
                + '<ul style="font-size:8pt;">'
                    + '<li><span style="font-weight:bold;margin-right:4px;">panelDeviceId</span>The device id of the virtual panel.</li>'
                    + '<li><span style="font-weight:bold;margin-right:4px;">controlValue</span>The current control value for the control.  For instance if the control is a slider this will be the current percentage of the slider.</li>'
                + '</ul>'
            + '</div>'
            + '<div style="padding-left:7px;"><div><span style="font-weight:bold;">Usage Examples:</span></div>'
            + '<div style="padding-left:14px;"><div style="font-weight:bold">This will set another slider to the value of the slider that is being moved.</div>'
            + '<div style="font-family:courier new, monospace;font-size:8pt;">'
                + '<div style="color:gray;">-- Set dimmerDeviceId to the device id of the dimmer</div>'
                + '<div style="color:gray;">-- that is to follow the dimmer on the panel</div>'
                + '<div><span style="color:blue;">local</span> dimmerDeviceId = 123</div>'
                + '<div>luup.call_action(<span style="color:crimson">"urn:upnp-org:serviceId:Dimming1"</span>, <span style="color:crimson">"SetLoadLevelTarget"</span>,</div>'
                    + '<div style="padding-left:40px;">{["newLoadlevelTarget"] = <span style="font-weight:bold;color:green;background-color:yellow;">context.controlValue</span>}, dimmerDeviceId)</div>'
            + '</div>'
        + '</div>';
        return html;
    }
    function buildColorExpressionInfo() {
        var html = '<div style="font-size:9pt;width:270px;"><div style="font-weight:bold;font-size:12pt;display:inline-block;">Expressions:<span></div>'
            + '<div>'
            + '<p style="font-size:8pt;">Expressions are evaluated as a series of conditions from the topmost to the bottom in the list.  The first condition that evaluates to true will set the indicator to the selected color.</p>'
            + '<p style="font-size:8pt;">If none of the conditions in the list evaluate to true the default color will be selected.</p>'
            + '</div></div>';
        return html;
    }

    function buildEditToggleElementTemplate(elementState) {
    	var html = '<div style="">';
    	if (typeof (elementState.onState.display.operator) == 'undefined') elementState.onState.display.operator = '==';
    	if (typeof (elementState.onState.display.maxValue) == 'undefined') elementState.onState.display.maxValue = '';
    	if (typeof (elementState.offState.display.operator) == 'undefined') elementState.offState.display.operator = '==';
    	if (typeof (elementState.offState.display.maxValue) == 'undefined') elementState.offState.display.maxValue = '';

        html += '<table style="width:100%;"><tbody>';
        html += '</tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Size &amp; Position <span style="font-size:7pt;font-style:italic">(Left/Top values for Detail Tab only)</span></td></tr>';
        html += '<tr><td colspan="3">' +
            buildElementPositionTemplate('.position', 
            elementState.position) + '</td></tr>';

        html += '<tr><td colspan="3">'
            + '<table style="width:100%"><tbody>'
            + '<tr><td style="vertical-align:top;width:50%;">'
            + '<table style="width:100%;"><tbody>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Toggled On State <span style="font-size:7pt;font-style:italic">(used to set the ON state of the toggle)</span></td></tr>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Text</td>';
        html += '<td style="padding-left:2px;padding-right:2px;"><input data-bind="onState.text" class="vpnl_editField" onchange="textStyler.updateSampleText(\'vpnl_toggleOnStateText\', this.value);" type="text" value="' + elementState.onState.text + '"/>'
            + textStyler.createHtml('vpnl_toggleOnStateText', '.onState.textStyle', elementState.onState.textStyle, elementState.onState.text) + '</td><td/></tr>';
        html += '<td colspan="3">' + buildDeviceVariable('.onState.display',
            elementState.onState.display.deviceId,
            elementState.onState.display.serviceId,
            elementState.onState.display.variable, elementState.onState.display.operator, elementState.onState.display.value, elementState.onState.display.maxValue) + '</td>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:lightsteelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Command <span style="font-size:7pt;font-style:italic">(action to be performed when ON state is selected)</span></td></tr>';
        html += '<tr><td colspan="3">' + buildActionCommand(elementState.ordinal, '.onState.command', elementState.onState.command) + '</td></tr>';
        html += '</tbody></table>'
            + '</td><td style="vertical-align:top;width:50%;">'
            + '<table style="width:100%;"><tbody>';

        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Toggled Off State <span style="font-size:7pt;font-style:italic">(used to set the OFF state of the toggle)</span></td></tr>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Text</td>';
        html += '<td style="padding-left:2px;padding-right:2px;"><input data-bind="offState.text" class="vpnl_editField" onchange="textStyler.updateSampleText(\'vpnl_toggleOffStateText\', this.value);" type="text" value="' + elementState.offState.text + '"/>'
            + textStyler.createHtml('vpnl_toggleOffStateText', '.offState.textStyle', elementState.offState.textStyle, elementState.offState.text) + '</td><td/></tr>';
        html += '<tr><td colspan="3">' + buildDeviceVariable('offState.display',
            elementState.offState.display.deviceId,
            elementState.offState.display.serviceId,
            elementState.offState.display.variable, elementState.offState.display.operator, elementState.offState.display.value, elementState.offState.display.value) + '</td>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:lightsteelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Command <span style="font-size:8pt;font-style:italic">(action to be performed when the OFF state is selected)</span></td></tr>';
        html += '<tr><td colspan="3">' + buildActionCommand(elementState.ordinal, '.offState.command', elementState.offState.command) + '</td></tr>';
        html += '</tbody></table></td></tr></tbody></table>'
        html += '</tbody></table></div>'
        return html;
    }
    function buildEditSliderElementTemplate(elementState) {
        var html = '<div style="">';
        html += '<div>';
        html += '<table style="width:100%;"><tbody>';

        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Size &amp; Position <span style="font-size:7pt;font-style:italic;">(Left/Top values for Detail Tab only)</span></td></tr>';
        html += '<tr><td colspan="3">' +
            buildElementPositionTemplate('.position', elementState.position) + '</td></tr>';
        html += '<tr class="sizepos vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Device Variable to Monitor <span style="font-size:7pt;font-style:italic">(used to set the slider position)</span></td></tr>';
        html += '<tr class="sizepos">';
        html += '<td colspan="3">' + buildDeviceVariable('.display',
            elementState.display.deviceId,
            elementState.display.serviceId,
            elementState.display.variable, null, null) + '</td>';
        html += '<tr class="btncmd vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Command <span style="font-size:7pt;font-style:italic">(action to be performed when the slider is moved)</span></td></tr>';
        html += '<td colspan="3">' + buildActionCommand(elementState.ordinal, '.command', elementState.command) + '</td>';
        html += '<tr class="btncmd vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Validation</td></tr>';
        if (typeof (elementState.values) == 'undefined')
            elementState.values = {
                min: { text: 'OFF', value: 0 },
                max: { text: 'ON', value: 100 },
				suffix: '%'
            }
        if(typeof(elementState.values.suffix) == 'undefined')
        	elementState.values.suffix = '%';
        html += '<tr><td colspan="3" style="padding:2px;">'
            + '<table><tbody>'
            + '<tr style=""><td style="width:77px;padding-left:2px;padding-right:2px;"/><td style="width:77px;padding-left:2px;padding-right:2px;border:solid 1px silver;background-color:gainsboro;">Min</td><td style="width:77px;padding-left:2px;padding-right:2px;border:solid 1px silver;background-color:gainsboro;">Max</td></tr>'
            + '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Text</td><td style="width:77px;padding-left:2px;padding-right:2px"><input class="vpnl_editField" data-bind=".values.min.text" value="' + elementState.values.min.text + '"/></td>'
            + '<td style="width:77px;padding-left:2px;padding-right:2px"><input class="vpnl_editField" data-bind=".values.max.text" value="' + elementState.values.max.text + '"/></td>'
            + '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Value</td><td style="width:77px;padding-left:2px;padding-right:2px"><input class="vpnl_editField" data-bind=".values.min.value" value="' + elementState.values.min.value + '" /></td>'
            + '<td style="width:77px;padding-left:2px;padding-right:2px"><input class="vpnl_editField" data-bind=".values.max.value" value="' + elementState.values.max.value + '"/></td>'
            + '<tr><tr><td>Suffix</td><td colspan="2"><input class="vpnl_editField" data-bind=".values.suffix" style="width:77px;" value="' + elementState.values.suffix + '"/><td></tr></tbody></table></td>';
        html += '</tr></tbody></table></div>';
        html += '</div>';
        return html;
    }
    function buildEditVerticalSliderElementTemplate(elementState) {
        var html = '<div style="font-size:9pt">';
        html += '<div>';
        html += '<table style="width:100%;"><tbody>';

        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Size &amp; Position <span style="font-size:7pt;font-style:italic">(Left/Top values for Detail Tab only)</span></td></tr>';
        html += '<tr><td colspan="3">' +
            buildElementPositionTemplate('.position',
            elementState.position) + '</td></tr>';
        html += '<tr class="vpnl_dialogSubHeader" class="sizepos" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Device Variable to Monitor <span style="font-size:7pt;font-style:italic">(used to set the slider position)</span></td></tr>';
        html += '<tr class="sizepos">';
        html += '<td colspan="3">' + buildDeviceVariable('.display',
            elementState.display.deviceId,
            elementState.display.serviceId,
            elementState.display.variable, null, null) + '</td>';
        html += '<tr class="btncmd vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Command <span style="font-size:7pt;font-style:italic">(action to be performed when the button is clicked)</span></td></tr>';
        html += '<td colspan="3">' + buildActionCommand(elementState.ordinal, '.command', elementState.command) + '</td>';
        html += '</tr>';
        html += '<tr class="btncmd vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Validation</td></tr>';

        if (typeof (elementState.values) == 'undefined')
        	elementState.values = {
        		min: { text: 'LOW', value: 0 },
        		max: { text: 'HIGH', value: 100 },
        		suffix: '&deg;'
        	}
        if (typeof (elementState.values.suffix) == 'undefined')
        	elementState.values.suffix = '&deg;';
        html += '<tr><td colspan="3" style="padding:2px;">'
            + '<table><tbody>'
            + '<tr style=""><td style="width:77px;padding-left:2px;padding-right:2px;"/><td style="width:77px;padding-left:2px;padding-right:2px;border:solid 1px silver;background-color:gainsboro;">Min</td><td style="width:77px;padding-left:2px;padding-right:2px;border:solid 1px silver;background-color:gainsboro;">Max</td></tr>'
            + '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Text</td><td style="width:77px;padding-left:2px;padding-right:2px"><input class="vpnl_editField" data-bind=".values.min.text" value="' + elementState.values.min.text + '"/></td>'
            + '<td style="width:77px;padding-left:2px;padding-right:2px"><input class="vpnl_editField" data-bind=".values.max.text" value="' + elementState.values.max.text + '"/></td>'
            + '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Value</td><td style="width:77px;padding-left:2px;padding-right:2px"><input class="vpnl_editField" data-bind=".values.min.value" value="' + elementState.values.min.value + '" /></td>'
            + '<td style="width:77px;padding-left:2px;padding-right:2px"><input class="vpnl_editField" data-bind=".values.max.value" value="' + elementState.values.max.value + '"/></td>'
            + '<tr><tr style="display:none;"><td>Suffix</td><td colspan="2"><input class="vpnl_editField" data-bind=".values.suffix" style="width:77px;" value="' + elementState.values.suffix + '"/><td></tr></tbody></table></td>';
        html += '</tr></tbody></table></div>';
        return html;
    }
    function buildEditHorizontalSpinnerElementTemplate(elementState) {
    	var html = '<div style="font-size:9pt">';
    	html += '<div>';
    	html += '<table style="width:100%;"><tbody>';

    	html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Size &amp; Position <span style="font-size:7pt;font-style:italic">(Left/Top values for Detail Tab only)</span></td></tr>';
    	html += '<tr><td colspan="3">' +
            buildElementPositionTemplate('.position',
            elementState.position) + '</td></tr>';
    	html += '<tr class="vpnl_dialogSubHeader" class="sizepos" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Device Variable to Monitor <span style="font-size:7pt;font-style:italic">(used to set the spinner value)</span></td></tr>';
    	html += '<tr class="sizepos">';
    	html += '<td colspan="3">' + buildDeviceVariable('.display',
            elementState.display.deviceId,
            elementState.display.serviceId,
            elementState.display.variable, null, null) + '</td>';
    	html += '<tr class="btncmd vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Command <span style="font-size:7pt;font-style:italic">(action to be performed when the button is clicked)</span></td></tr>';
    	html += '<td colspan="3">' + buildActionCommand(elementState.ordinal, '.command', elementState.command) + '</td>';
    	html += '</tr>';
    	html += '<tr class="btncmd vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Validation</td></tr>';

    	if (typeof (elementState.values) == 'undefined')
    		elementState.values = { allowEdit: true,
    			min: { text: 'Decrease', value: 0 },
    			max: { text: 'Increase', value: 100 },
    			suffix: '&deg;'
    		}
    	if (typeof (elementState.values.suffix) == 'undefined') elementState.values.suffix = '&deg;';
    	if (typeof (elementState.values.step) == 'undefined') elementState.values.step = "1";
    	html += '<tr><td colspan="3" style="padding:2px;">'
			+ '<div style="padding-left:87px;"><input type="checkbox" data-bind=".values.allowEdit" id="vpnl_allowEdit"' + (elementState.values.allowEdit ? ' checked="checked"' : '') + '/><label style="white-space:nowrap;font-weight:normal;padding-left:4px;vertical-align:top;" for="vpnl_allowEdit">Allow Edit Value</label></div>'
            + '<table><tbody>'
            + '<tr style=""><td style="width:87px;padding-left:2px;padding-right:2px;"/><td style="width:77px;padding-left:2px;padding-right:2px;border:solid 1px silver;background-color:gainsboro;text-align:center;">Decrement</td><td style="width:77px;padding-left:2px;padding-right:2px;border:solid 1px silver;background-color:gainsboro;text-align:center;">Increment</td></tr>'
            + '<tr><td style="width:87px;padding-left:2px;padding-right:2px;">Hover Text</td><td style="width:77px;padding-left:2px;padding-right:2px"><input class="vpnl_editField" data-bind=".values.min.text" value="' + elementState.values.min.text + '"/></td>'
				+ '<td style="width:87px;padding-left:2px;padding-right:2px"><input class="vpnl_editField" data-bind=".values.max.text" value="' + elementState.values.max.text + '"/></td>'
            + '<tr style=""><td style="width:87px;padding-left:2px;padding-right:2px;"/><td style="width:77px;padding-left:2px;padding-right:2px;border:solid 1px silver;background-color:gainsboro;text-align:center;">Min</td><td style="width:77px;padding-left:2px;padding-right:2px;border:solid 1px silver;background-color:gainsboro;text-align:center;">Max</td></tr>'
            + '<tr><td style="width:87px;padding-left:2px;padding-right:2px;">Min/Max Value</td><td style="width:77px;padding-left:2px;padding-right:2px"><input class="vpnl_editField" data-bind=".values.min.value" value="' + elementState.values.min.value + '" /></td>'
				+ '<td style="width:87px;padding-left:2px;padding-right:2px"><input class="vpnl_editField" data-bind=".values.max.value" value="' + elementState.values.max.value + '"/></td>'
			+ '<tr><tr><td style="width:87px;padding-left:2px;padding-right:2px;white-space:nowrap;">Increment Step</td><td colspan="2" style="padding-left:2px;"><input class="vpnl_editField" data-bind=".values.step" style="width:37px;" value="' + elementState.values.step + '"/><td></tr>'
            + '<tr><tr><td style="width:87px;padding-left:2px;padding-right:2px">Suffix</td><td colspan="2" style="padding-left:2px;"><input class="vpnl_editField" data-bind=".values.suffix" style="width:37px;" value="' + elementState.values.suffix + '"/><td></tr></tbody></table></td>'
    	html += '</tr></tbody></table></div>';
    	return html;
    }

    function buildEditHorizontalRuleElementTemplate(elementState) {
        function createUpdateSampleEvent() {
            return '$(\'#vpnl_modalContent\').find(\'td.vpnl_horizontalRuleSample\').html(virtualPanelBuilder.buildHorizontalRuleSample(dialogManager.fromDialog()));';
        }
        if (typeof (elementState) == 'undefined')
            elementState = {
                controlType: 'horizontal_rule',
                position: { left: '0', top: '0', width: '47', height: '0' },
                lineStyle: { weight: 'fine', style: 'solid', color: '' },
                margins: { left: '0', top: '0', right: '0', bottom: '0' }
            }
        var html = '<div style="">';
        html += '<div>';
        html += '<table style="width:100%;"><tbody>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Size &amp; Position <span style="font-size:7pt;font-style:italic;">(Left/Top values for Detail Tab only)</span></td></tr>';
        html += '<tr><td colspan="3">' +
            buildElementPositionTemplate('.position',
            elementState.position,
            null, null, createUpdateSampleEvent()) + '</td></tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Appearance</td></tr>';
        html += '<tr><td style="padding-left:2px;width:77px;">Style</td>';
        html += '<td><select class="vpnl_editField" data-bind=".lineStyle.style" onchange="' + createUpdateSampleEvent() + '">'
                + '<option value="solid"' + (elementState.lineStyle.style == 'solid' || typeof(elementState.lineStyle.style) == 'undefined' || elementState.lineStyle.style == '' ? ' selected = "selected"' : '') + '>Solid</option>'
                + '<option value="dashed"' + (elementState.lineStyle.style == 'dashed' ? ' selected = "selected"' : '') + '>Dashed</option>'
                + '<option value="dotted"' + (elementState.lineStyle.style == 'dotted' ? ' selected = "selected"' : '') + '>Dotted</option>'
                + '<option value="double"' + (elementState.lineStyle.style == 'double' ? ' selected = "selected"' : '') + '>Double</option>'
                + '<option value="groove"' + (elementState.lineStyle.style == 'groove' ? ' selected = "selected"' : '') + '>Groove</option>'
                + '<option value="ridge"' + (elementState.lineStyle.style == 'ridge' ? ' selected = "selected"' : '') + '>Ridge</option>'
                + '<option value="inset"' + (elementState.lineStyle.style == 'inset' ? ' selected = "selected"' : '') + '>Inset</option>'
                + '<option value="outset"' + (elementState.lineStyle.style == 'outset' ? ' selected = "selected"' : '') + '>Outset</option>'
            + '</td><td/></tr>';

        html += '<tr><td style="padding-left:2px;width:77px;">Weight</td>';
        html += '<td><select class="vpnl_editField" data-bind=".lineStyle.weight" onchange="' + createUpdateSampleEvent() + '">'
                + '<option value="extrafine"' + (elementState.lineStyle.weight == 'extrafine' ? ' selected = "selected"' : '') + '>Extra Fine</option>'
                + '<option value="fine"' + (elementState.lineStyle.weight == 'fine' || typeof(elementState.lineStyle.weight) == 'undefined' || elementState.lineStyle.weight == '' ? ' selected = "selected"' : '') + '>Fine</option>'
                + '<option value="medium"' + (elementState.lineStyle.weight == 'medium' ? ' selected = "selected"' : '') + '>Medium</option>'
                + '<option value="thick"' + (elementState.lineStyle.weight == 'thick' ? ' selected = "selected"' : '') + '>Thick</option>'
                + '<option value="extrathick"' + (elementState.lineStyle.weight == 'extrathick' ? ' selected = "selected"' : '') + '>Extra Thick</option>'
            + '</td><td/></tr>';

        html += '<tr><td style="padding-left:2px;width:77px;">Color</td><td style="padding-bottom:2px;">';
        html += '<input type="text" data-bind=".lineStyle.color" class="vpnl_editField" list="vpnl_linecolordatalist" value="' + elementState.lineStyle.color + '" onchange="' + createUpdateSampleEvent() + '" >'
        html += '<datalist id="vpnl_linecolordatalist">';
        for (var clr in textStyler.colorList) {
            html += '<option value="' + clr + '" style="color:' + clr + ';">' + textStyler.colorList[clr] + '</option>';
        }
        html += '</datalist></input></td><td/></tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Margins</td></tr>';
        html += '<tr><td colspan="3">' +
            buildElementMarginTemplate('.margins',
            null, elementState.margins.top,
            null,
            elementState.margins.bottom, null, createUpdateSampleEvent()) + '</td></tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Sample</td></tr>';
        html += '<tr><td colspan="3" class="vpnl_horizontalRuleSample" style="padding:4px;">' + buildHorizontalRuleSample(elementState) + '</td></tr>';
        html += '</tbody></table></div>';
        return html;
    }
    function buildElementMarginTemplate(binding, left, top, right, bottom, sMsg, sOnChange) {
        var html = '<table><tbody><tr><td><div style="display:inline-block;"><table style=""><tbody>';
        html += '<tr>';
        html += '<td style="width:77px;padding-left:2px;padding-right:2px;' + ((left == null) ? 'display:none;' : '') + '">Left</td>';
        html += '<td style="width:34px;' + ((left == null) ? 'display:none;' : '') + '">'
            + '<input data-bind="' + binding + '.left" type="text" class="vpnl_editField" data-datatype="float" style="width:34px;text-align:right;" ' 
                + (left == null ? 'disabled="disabled"' : '')
                + ' value="' + (left == null ? '0' : left) + '"'
                + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '/>'
            + '</td><td style="padding-left:2px;' + ((left == null) ? 'display:none;' : '') + '">px</td>';

        html += '<td style="padding-left:27px;padding-right:2px;' + ((right == null) ? 'display:none;' : '') + '">Right</td>';
        html += '<td style="width:34px;' + ((right == null) ? 'display:none;' : '') + '">'
            + '<input data-bind="' + binding + '.right" type="text" class="vpnl_editField" data-datatype="float" style="width:34px;text-align:right;" '
                + (right == null ? 'disabled="disabled"' : '')
                + ' value="' + (right == null ? '0' : right) + '"'
                + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '/>'
            + '</td><td style="padding-left:2px;' + ((right == null) ? 'display:none;' : '')
            + '"' + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '">px</td>';
        html += '</tr>';
        html += '<tr>';
        html += '<td style="width:77px;padding-left:2px;padding-right:2px;' + ((top == null) ? 'display:none;' : '') + '">Top</td>';
        html += '<td style="width:34px;' + ((top == null) ? 'display:none;' : '') + '">'
            + '<input data-bind="' + binding + '.top" type="text" class="vpnl_editField" data-datatype="float" style="width:34px;text-align:right;" '
                + (top == null ? 'disabled="disabled"' : '')
                + ' value="' + (top == null ? '0' : top) + '"'
                + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '/>'
            + '</td><td style="padding-left:2px;' + ((top == null) ? 'display:none;' : '')
            + '"' + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '">px</td>';
        html += '<td style="padding-left:27px;padding-right:2px;' + ((bottom == null) ? 'display:none' : '') + '">Bottom</td>';
        html += '<td style="width:34px;' + ((bottom == null) ? 'display:none;' : '') + '">'
            + '<input data-bind="' + binding + '.bottom" type="text" class="vpnl_editField" data-datatype="float" style="width:34px;text-align:right;" '
                + (bottom == null ? 'disabled="disabled"' : '')
                + ' value="' + (bottom == null ? '0' : bottom) + '"'
                + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '/>'
            + '</td><td style="padding-left:2px;' + ((bottom == null) ? 'display:none;' : '')
            + '"' + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '">px</td>';
        html += '</tr></tbody></table></div>';
        if (typeof (sMsg) != 'undefined' && sMsg != null) {
            html += '<td><div style="margin:4px;position:inline-block;background-color:lightyellow;border:solid 1px gray;padding:4px;border-radius:7px;font-size:7pt;">' + sMsg + '</div></td>';

        }
        html += '</tr></tbody></table>';
        return html;
    }
    function buildElementPositionTemplate(binding, pos, sMsg, sOnChange) {
        var html = '<table><tbody><tr><td><div style="display:inline-block;"><table style=""><tbody>';
        html += '<tr>';
        html += '<td style="width:77px;padding-left:2px;padding-right:2px;' + ((pos.left == null) ? 'display:none;' : '') + '">Left</td>';
        html += '<td style="width:34px;' + ((pos.left == null) ? 'display:none' : '') + '">'
            + '<input data-bind="' + binding + '.left" type="text" class="vpnl_editField" data-datatype="float" style="width:34px;text-align:right;" '
                + (pos.left == null ? 'disabled="disabled"' : '')
                + ' value="' + (pos.left == null ? '0' : pos.left) + '"'
                + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '/>'
            + '</td><td style="padding-left:2px;' + ((pos.left == null) ? 'display:none;' : '') + '">px</td>';

        html += '<td style="padding-left:27px;padding-right:2px;' + ((pos.width == null) ? 'display:none;' : '') + '">Width</td>';
        html += '<td style="width:34px;' + ((pos.width == null) ? 'display:none' : '') + '">'
            + '<input data-bind="' + binding + '.width" type="text" class="vpnl_editField" data-datatype="float" style="width:34px;text-align:right;" '
                + (pos.width == null ? 'disabled="disabled"' : '')
                + ' value="' + (pos.width == null ? '0' : pos.width) + '"'
                + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '/>'
            + '</td><td style="padding-left:2px;' + ((pos.width == null) ? 'display:none;' : '') + '">px</td><td/>';
        html += '</tr>';
        html += '<tr>';
        html += '<td style="width:77px;padding-left:2px;padding-right:2px;' + ((pos.top == null) ? 'display:none;' : '') + '">Top</td>';
        html += '<td stlye="width:34px;' + ((pos.top == null) ? 'display:none' : '') + '">'
            + '<input data-bind="' + binding + '.top" type="text" class="vpnl_editField" data-datatype="float" style="width:34px;text-align:right;" '
                + (pos.top == null ? 'disabled="disabled"' : '')
                + ' value="' + (pos.top == null ? '0' : pos.top) + '"'
                + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '/>'
            + '</td><td style="padding-left:2px;' + ((pos.top == null) ? 'display:none;' : '') + '">px</td>';

        html += '<td style="padding-left:27px;padding-right:2px;' + ((pos.height == null) ? 'display:none;' : '') + '">Height</td>';
        html += '<td style="width:34px;' + ((pos.height == null) ? 'display:none;' : '') + '">'
            + '<input data-bind="' + binding + '.height" type="text" class="vpnl_editField" data-datatype="float" style="width:34px;text-align:right;" '
                + (pos.height == null ? 'disabled="disabled"' : '')
                + ' value="' + (pos.height == null ? '0' : pos.height) + '"'
                + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '/>'
            + '</td><td style="padding-left:2px;' + ((pos.height == null) ? 'display:none;' : '') + '">px</td><td/>';
        html += '</tr></tbody></table></div>';
        html += '<td><div style="padding-left:14px;"><input id="vpnl_hideOnDashboard" type="checkbox" data-bind="' + binding + '.hideOnDashboard" ' + ((pos.hideOnDashboard) ? ' checked="checked"' : '') + '/><label for="vpnl_hideOnDashboard" style="vertical-align:top;font-weight:normal;white-space:nowrap;">Hide on Dashboard</label></div>'
        if (typeof (sMsg) != 'undefined' && sMsg != null) {
            html += '<div style="margin:4px;position:inline-block;background-color:lightyellow;border:solid 1px gray;padding:4px;border-radius:7px;font-size:7pt;">' + sMsg + '</div>';

        }
        html += '</td></tr></tbody></table>';
        return html;
    }

    function buildElementPositionTemplate2(binding, x, y, width, height, sMsg, sOnChange) {
        var html = '<table><tbody><tr><td><div style="display:inline-block;"><table style=""><tbody>';
        html += '<tr>';
        html += '<td style="width:77px;padding-left:2px;padding-right:2px;' + ((x == null) ? 'display:none;' : '') + '">Left</td>';
        html += '<td style="width:34px;' + ((x == null) ? 'display:none' : '') + '">'
            + '<input data-bind="' + binding + '.left" type="text" class="vpnl_editField" data-datatype="float" style="width:34px;text-align:right;" '
                + (x == null ? 'disabled="disabled"' : '')
                + ' value="' + (x == null ? '0' : x) + '"'
                + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '/>'
            + '</td><td style="' + ((x == null) ? 'display:none;' : '') + '">px</td>';

        html += '<td style="padding-left:27px;padding-right:2px;' + ((width == null) ? 'display:none;' : '') + '">Width</td>';
        html += '<td style="width:34px;' + ((width == null) ? 'display:none' : '') + '">'
            + '<input data-bind="' + binding + '.width" type="text" class="vpnl_editField" data-datatype="float" style="width:34px;text-align:right;" '
                + (width == null ? 'disabled="disabled"' : '')
                + ' value="' + (width == null ? '0' : width) + '"'
                + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '/>'
            + '</td><td style="' + ((width == null) ? 'display:none;' : '') + '">px</td><td/>';
        html += '</tr>';
        html += '<tr>';
        html += '<td style="width:77px;padding-left:2px;padding-right:2px;' + ((y == null) ? 'display:none;' : '') + '">Top</td>';
        html += '<td stlye="width:34px;' + ((y == null) ? 'display:none' : '') + '">'
            + '<input data-bind="' + binding + '.top" type="text" class="vpnl_editField" data-datatype="float" style="width:34px;text-align:right;" '
                + (y == null ? 'disabled="disabled"' : '')
                + ' value="' + (y == null ? '0' : y) + '"'
                + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '/>'
            + '</td><td style="' + ((y == null) ? 'display:none;' : '') + '">px</td>';

        html += '<td style="padding-left:27px;padding-right:2px;' + ((height == null) ? 'display:none;' : '') + '">Height</td>';
        html += '<td style="width:34px;' + ((height == null) ? 'display:none;' : '') + '">'
            + '<input data-bind="' + binding + '.height" type="text" class="vpnl_editField" data-datatype="float" style="width:34px;text-align:right;" '
                + (height == null ? 'disabled="disabled"' : '')
                + ' value="' + (height == null ? '0' : height) + '"'
                + ((sOnChange == null) ? '' : (' onchange="' + sOnChange + '"')) + '/>'
            + '</td><td style="' + ((height == null) ? 'display:none;' : '') + '">px</td><td/>';
        html += '</tr></tbody></table></div>';
        if (typeof (sMsg) != 'undefined' && sMsg != null) {
            html += '<td><div style="margin:4px;position:inline-block;background-color:lightyellow;border:solid 1px gray;padding:4px;border-radius:7px;font-size:7pt;">' + sMsg + '</div></td>';

        }
        html += '</tr></tbody></table>';
        return html;
    }
    function buildDeviceVariable(binding, deviceId, serviceId, variable, operator, value, maxValue) {
        var html = '<table class="vpnl_deviceVariable" style="width:100%;"><tbody>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Device</td>';
        html += '<td>' + buildAvailableDevices(deviceId, binding) + '</td><td/></tr>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;white-space:nowrap;">Service Id</td>';
        html += '<td><input type="text" class="vpnl_editField vpnl_serviceId" list="vpnl_datalist' + binding.replace(/\./g, '_') + '_serviceId" data-bind="' + binding + '.serviceId" style="width:297px;" value="' + serviceId + '" onchange="virtualPanelBuilder.onReplaceVariableDataList(this);">'
            + '<div id="vpnl_listContainer' + binding.replace(/\./g, '_') + '_serviceId">' + buildServiceIdDataList(deviceId, binding) + '</div></input></td><td/></tr>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Variable</td>';
        html += '<td><table><tbody><tr>'
        html += '<td><input type="text" class="vpnl_editField vpnl_variableName" list="vpnl_datalist' + binding.replace(/\./g, '_') + '_variable" data-bind="' + binding + '.variable" style="width:187px;" value="' + variable + '"/>'
            + '<div id="vpnl_listContainer' + binding.replace(/\./g, '_') + '_variable">' + buildVariableDataList(deviceId, serviceId, binding) + '</div></td>';
        if (operator != undefined && operator != '=') {
        	html += '<td><select class="vpnl_editField" data-bind="' + binding + '.operator" style="width:100%;min-width:77px;" onchange="virtualPanelBuilder.onVariableOperatorChanged(this, this.value);">' +
                    '<option value="==" ' + (!operator || operator == '==' ? 'selected="selected"' : '') + '>==</option>' +
                    '<option value="&gt;=" ' + (operator == '>=' ? 'selected="selected"' : '') + '>&gt;=</option>' +
                    '<option value="&lt;=" ' + (operator == '<=' ? 'selected="selected"' : '') + '>&lt;=</option>' +
                    '<option value="&gt;" ' + (operator == '>' ? 'selected="selected"' : '') + '>&gt;</option>' +
                    '<option value="&lt;" ' + (operator == '<' ? 'selected="selected"' : '') + '>&lt;</option>' +
					'<option value="between" ' + (operator == 'between' ? 'selected="selected"' : '') + '>between</option>' +
                '</select></td></tr><tr>';
        }
        else if (value != undefined && operator != '=')
            html += '<td style="padding-left:2px;padding-right:2px;text-align:center;">==</td>'
        else if (value != undefined && operator == '=')
            html += '<td style="padding-left:2px;padding-right:2px;text-align:center;">Value</td>'
        else
            html += '<td/>';
        if (value != undefined) {
        	html += '<td><table><tbody><tr><td style=""><input type="text" class="vpnl_editField" data-bind="' + binding + '.value" style="width:77px;" value="' + value + '"/></td>';
        	if (maxValue != undefined) {
        		html += '<td class="vpnl_betweenOperator" style="padding-left:2px;padding-right:2px;' + ((operator == 'between') ? '' : 'display:none;') + '">and</td><td class="vpnl_betweenOperator" style="' + ((operator == 'between') ? '' : 'display:none;') + '"><input type="text" class="vpnl_editField" data-bind="' + binding + '.maxValue" style="width:77px;" value="' + maxValue + '"/></td>';
        	}
        	html += '</td></tr></tbody></table></td></tr>'
        }
        else
        	html += '<td/></tr>';
        html += '</tbody></table></td></tr></tbody></table>';
        return html;
    }
    function buildSetDeviceVariable(binding, deviceId, serviceId, variable, value, useControlValue) {
    	console.log(useControlValue);
    	var html = '<table class="vpnl_deviceVariable" style="width:100%;"><tbody>';
    	html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Device</td>';
    	html += '<td>' + buildAvailableDevices(deviceId, binding) + '</td><td/></tr>';
    	html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;white-space:nowrap;">Service Id</td>';
    	html += '<td><input type="text" class="vpnl_editField vpnl_serviceId" list="vpnl_datalist' + binding.replace(/\./g, '_') + '_serviceId" data-bind="' + binding + '.serviceId" style="width:297px;" value="' + serviceId + '" onchange="virtualPanelBuilder.onReplaceVariableDataList(this);">'
            + '<div id="vpnl_listContainer' + binding.replace(/\./g, '_') + '_serviceId">' + buildServiceIdDataList(deviceId, binding) + '</div></input></td><td/></tr>';
    	html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Variable</td>';
    	html += '<td><input type="text" class="vpnl_editField vpnl_variableName" list="vpnl_datalist' + binding.replace(/\./g, '_') + '_variable" data-bind="' + binding + '.variable" style="width:187px;" value="' + variable + '"/>'
            + '<div id="vpnl_listContainer' + binding.replace(/\./g, '_') + '_variable">' + buildVariableDataList(deviceId, serviceId, binding) + '</div></td>';
    	html += '<tr><td style="padding-left:2px;" colspan="2"><input id="vpnl_useControlValue" style="display:inline-block;vertical-align:middle;margin-top:5px;margin-bottom:4px;" onchange="virtualPanelBuilder.onUseControlValueChecked(this, this.value);" type="checkbox" data-bind="' + binding + '.useControlValue"' + ((useControlValue === true) ? ' checked="checked"' : '') + '/><label for="vpnl_useControlValue" style="white-space:nowrap;padding-left:2px;font-weight:normal;margin:0px;vertical-align:middle;">Use Control Value</label>';
    	html += '<input type="text" class="vpnl_editField vpnl_setVariableValue" data-bind="' + binding + '.value" style="width:77px;margin-left:4px;' + ((typeof(useControlValue) == 'undefined' || useControlValue == false) ? '' : 'display:none;') + '" value="' + value + '"/></td></tr>'
    	html += '</tbody></table>';
    	return html;
    }

    function buildServiceIdDataList(deviceId, binding) {
        var html = '<datalist id="vpnl_datalist' + binding.replace(/\./g, '_') + '_serviceId">';
        if (parseInt(deviceId, 10) > 0 && !isNaN(parseInt(deviceId, 10))) {
            for (var i = 0; i < deviceList.devices.length; i++) {
                var dev = deviceList.devices[i];
                if (parseInt(deviceId, 10) == parseInt(dev.id, 10)) {
                    var arrServices = new Array();
                    for (var j = 0; j < dev.states.length; j++) {
                        var state = dev.states[j];
                        if (arrServices.indexOf(state.service) == -1) {
                            arrServices.push(state.service);
                        }
                    }
                    for (var k = 0; k < arrServices.length; k++) {
                        html += '<option value="' + arrServices[k] + '"/>';
                    }
                    break;
                }
            }
        }
        html += '</datalist>';
        return html;
    }
    function buildVariableDataList(deviceId, serviceId, binding) {
        var html = '<datalist id="vpnl_datalist' + binding.replace(/\./g, '_') + '_variable">';
        if (deviceId && deviceId.length > 0 && serviceId && serviceId.length > 0) {
            for (var i = 0; i < deviceList.devices.length; i++) {
                var dev = deviceList.devices[i];
                if (parseInt(deviceId, 10) == parseInt(dev.id, 10)) {
                    var arrVariables = new Array();
                    for (var j = 0; j < dev.states.length; j++) {
                        var state = dev.states[j];
                        if (state.service == serviceId && arrVariables.indexOf(state.variable) == -1) {
                            arrVariables.push(state.variable);
                        }
                    }
                    for (var k = 0; k < arrVariables.length; k++) {
                        html += '<option value="' + arrVariables[k] + '"/>';
                    }
                    break;
                }
            }
        }
        html += '</datalist>';
        return html;
    }
    function buildAvailableDevices(deviceId, binding) {
        var html = '<select class="vpnl_editField vpnl_deviceId" data-bind="' + binding + '.deviceId" onchange="virtualPanelBuilder.onReplaceServiceIdDataList(this);">';
        for (var key in availableDevices) {
            var room = availableDevices[key];
            if (room.devices && room.devices.length > 0) {
                html += '<optgroup style="font-weight:bold;" label="' + room.roomName + '">';
                for (k = 0; k < room.devices.length; k++) {
                    var d = room.devices[k];
                    html += '<option value="' + d.deviceId + '" ' + ((deviceId == d.deviceId) ? 'selected="selected"' : '') + '>' + d.deviceName + '</option>';
                }
                html += '</optgroup>';
            }
        }
        html += '</select>';
        return html;
    }
    function buildActionCommand(ordinal, binding, cmdState) {
        var html = '<table class="cmdActions" style="width:100%;"><tbody>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Action</td>';
        html += '<td>' + buildSelectCommandTypes(binding, cmdState.cmdType) + '</td><td/></tr>';

        html += '<tr class="cmdTypeSetVariable cmdType" style="' + (cmdState.cmdType != 'SetVariable' ? 'display:none;' : '') + '"><td colspan="2">' + buildSetDeviceVariable(binding + '.variable', cmdState.variable.deviceId, cmdState.variable.serviceId, cmdState.variable.variable, cmdState.variable.value, cmdState.variable.useControlValue) + '</td></tr>';
        html += '<tr class="cmdTypeCallAction cmdType" style="' + (cmdState.cmdType != 'CallAction' ? 'display:none;' : '') + '"><td colspan="2">' + buildCallActionParams(binding + '.action', cmdState.action) + '</td></tr>';
        html += '<tr class="cmdTypeExecScene cmdType" style="' + (cmdState.cmdType != 'ExecScene' ? 'display:none;' : '') + '"><td colspan="2">' + buildExecuteScene(binding + '.scene', cmdState.scene) + '</td></tr>';
        html += '<tr class="cmdTypeHttpGET cmdType" style="' + (cmdState.cmdType != 'HttpGET' ? 'display:none;' : '') + '"><td colspan="2">' + buildHttpGETParams(binding + '.http', cmdState.http) + '</td></tr>';
        html += '<tr class="cmdTypeHttpPOST cmdType" style="' + (cmdState.cmdType != 'HttpPOST' ? 'display:none;' : '') + '"><td colspan="2">' + buildHttpPOSTParams(binding + '.http', cmdState.http) + '</td></tr>';
        html += '<tr class="cmdTypeExecLua cmdType" style="' + (cmdState.cmdType != 'ExecLua' ? 'display:none;' : '') + '"><td colspan="2">' + buildExecLua(ordinal, binding + '.execLua', cmdState.execLua) + '</td></tr>';
        html += '</tbody></table>';
        return html;
    }
    function buildSelectCommandTypes(binding, sType) {
        return '<select class="vpnl_editField" data-bind="' + binding + '.cmdType" onchange="virtualPanelBuilder.onSetCommandParams($(this).parents(\'.cmdActions:first\')[0], this.value);" style="">' +
        '<option value="NoAction" ' + (!sType || sType == 'NoAction' ? 'selected="selected"' : '') + '>Do Nothing (Display Only)</option>' +
        '<option value="SetVariable" ' + (!sType || sType == 'SetVariable' ? 'selected="selected"' : '') + '>Set a Variable on a Device</option>' +
        '<option value="CallAction" ' + (sType == 'CallAction' ? 'selected="selected"' : '') + '>Call an Action on a Device</option>' +
        '<option value="ExecScene" ' + (sType == 'ExecScene' ? 'selected="selected"' : '') + '>Execute a Scene</option>' +
        '<option value="HttpGET" ' + (sType == 'HttpGET' ? 'selected="selected"' : '') + '>Send an HTTP GET Command</option>' +
        '<option value="HttpPOST" ' + (sType == 'HttpPOST' ? 'selected="selected"' : '') + '>Send an HTTP POST Command</option>' +
        '<option value="ExecLua" ' + (sType == 'ExecLua' ? 'selected="selected"' : '') + '>Execute Lua</option>' +
        '</select>';
    }
    function buildCallActionParams(binding, params) {
        var html = '<table><tbody>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Device</td><td>' + buildAvailableDevices(params.deviceId, binding) + '</td><td/></tr>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Service Id</td>';
        //    html += '<td><input type="text" data-bind="' + binding + '.serviceId" style="border-radius:4px;width:297px;" value="' + params.serviceId + '"/></td><td/></tr>';
        html += '<td><input type="text" class="vpnl_editField" list="vpnl_datalist' + binding.replace(/\./g, '_') + '_serviceId" data-bind="' + binding + '.serviceId" style="width:297px;" value="' + params.serviceId + '"/>'
            + '<div id="vpnl_listContainer' + binding.replace(/\./g, '_') + '_serviceId">' + buildServiceIdDataList(params.deviceId, binding) + '</div></td><td/></tr>';

        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Action</td>';
        html += '<td><input class="vpnl_editField" type="text" data-bind="' + binding + '.action" style="width:297px;" value="' + params.action + '"/></td><td/></tr>';
        html += '<tr><td style="padding-left:37px;" colspan="3">';

        html += '<table class="cmdActionParams" style="width:100%;border:solid 1px black;margin-top:4px;"><tbody>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="4" style="padding-left:2px;padding-right:2px;border-bottom:solid 3px orange;">Action Parameters <span style="font-style:italic;font-size:9pt;"></span></td></tr>';
        html += '<tr style="background-color:gainsboro;"><td style="width:24px;" /><td style="width:147px;">Name</td><td style=padding-left:4px;padding-right:4px;">Value</td><td>&nbsp;</td>';
        if (params && params.parameters && params.parameters.length > 0) {
            for (var i = 0; i < params.parameters.length; i++) {
                html += buildActionParameter(binding + '.parameters[' + i + ']', params.parameters[i].name, params.parameters[i].value, params.parameters[i].isControlValue);
            }
        }
        html += '<tr><td><img style="width:22px;cursor:pointer;" onclick="virtualPanelBuilder.onAddActionParameterRow($(this).parents(\'table.cmdActionParams:first\'), \''
            + binding + '.parameters[\' + ($(this).parents(\'table.cmdActionParams:first\')[0].rows.length - 3) + \']\');" src="skins/default/img/other/plusOpenBtn.png"/></td><td colspan="3"><span style="white-space:nowrap;font-size:8pt;font-style:italic;">Click + to add another parameter to the action.</span></td>';
        html += '</tbody></table></td></tr></tbody></table>';
        return html;
    }
    function buildActionParameter(binding, sName, sValue, isControlValue) {
        var html = '<tr><td style="width:24px;"><img style="width:22px;cursor:pointer;" src="skins/default/img/other/minusLightGreen.png" onclick="$(this).parents(\'tr:first\').remove();"/></td>';
        html += '<td><input type="text" class="vpnl_editField" data-bind="' + binding + '.name" style="width:100%;" type="text" value="' + sName + '"/></td>';
        html += '<td><input type="text" class="vpnl_editField" data-bind="' + binding + '.value" style="width:100%;" type="text" value="' + sValue + '" ' + (isControlValue ? 'disabled="disabled"' : '') + '/></td>';
        html += '<td title="Check this box to use the value of the control" style="padding-left:1px;"><input type="checkbox" onclick="$(this).parents(\'tr:first\')[0].cells[2].children[0].disabled = this.checked; $(this).parents(\'tr:first\')[0].cells[2].children[0].value = \'\';" data-bind="' + binding + '.isControlValue" ' + ((isControlValue) ? 'checked="checked"' : '') + '/></tr>';
        return html;
    }
    function buildExecuteScene(binding, params) {
        var html = '<table><tbody>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Scene</td><td>' + buildAvailableScenes(params.sceneId, binding) + '</td><td/></tr>';
        html += '</tbody></table>';
        return html;
    }
    function buildExecLua(ordinal, binding, params) {
        var sCode = (typeof (params) == 'undefined' || typeof (params.code) == 'undefined') ? '' : params.code;
        var html = '<div class="vpnl_execLuaContainer" style="padding:2px;"><textarea class="vpnl_currentLua" rows="10" style="width:100%;font-family:Courier New, monospace;" onkeyup="virtualPanelBuilder.onExecLuaCodeChanged(event, this);" class="vpnl_editField" data-bind="' + binding + '.code">' + sCode + '</textarea>';
        html += '<input type="text" data-bind="' + binding + '.callName' + '" style="display:none;" value="' + (binding).replace(/\./g, '_') + '"/>';
        html += '<span class="vpnl_testedLua" style="display:none;">' + sCode + '</span>';
        
        html += '<div class="vpnl_execLuaMessage"></div>';
        html += '<div style="float:left;display:inline-block;">' + dialogManager.createTextPopupButton('vpnl_luaContextInfo', 'Execute Lua Help') + "</div>";
        html += '<div style="float:right;font-size:8pt;display:inline-block;">';
        html += '<input class="vBtn pull-left vpnl_testLuaBtn" type="button" style="margin:2px;" data-iscodetested="true" onclick="virtualPanelBuilder.onCheckLuaSyntax(event, this);" onmouseover="if(!$(this).data(\'iscodetested\')) this.style.backgroundColor=\'red\';" onmouseout="if(!$(this).data(\'iscodetested\')) this.style.backgroundColor=\'orangered\';" value="Check Syntax"/>';
        html += '<input class="vBtn pull-left" type="button" style="margin:2px;" onclick="virtualPanelBuilder.onTestLuaCode(event, this);" value="Run Lua"/>';
        html += '</div>';
        html += '</div>';
        return html;
    }
    function buildAvailableScenes(sceneId, binding) {
        var html = '<select class="vpnl_editField" data-bind="' + binding + '.sceneId">';
        for (var key in availableDevices) {
            var room = availableDevices[key];
            if (room.scenes && room.scenes.length > 0) {
                html += '<optgroup style="font-weight:bold;" label="' + room.roomName + '">';
                for (k = 0; k < room.scenes.length; k++) {
                    var s = room.scenes[k];
                    html += '<option value="' + s.sceneId + '" ' + ((sceneId == s.sceneId) ? 'selected="selected"' : '') + '>' + s.sceneName + '</option>';
                }
                html += '</optgroup>';
            }
        }
        html += '</select>';
        return html;
    }
    function buildHttpGETParams(binding, params) {
        var html = '<table style="width:100%;"><tbody>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">URL</td><td style="padding-right:2px;"><input class="vpnl_editField" type="text" data-bindifhidden="true" data-bind="' + binding + '.url" style="width:100%;" value="' + params.url + '"/></td><td/></tr>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Username</td><td><input class="vpnl_editField" type="text" data-bindifhidden="true" data-bind="' + binding + '.username" value="' + params.username + '"/></td><td/></tr>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Password</td><td><input class="vpnl_editField" type="text" data-bindifhidden="true" data-bind="' + binding + '.password" value="' + params.password + '"/></td><td/></tr>';
        html += '</tbody></table>';
        return html;

    }
    function buildHttpPOSTParams(binding, params) {
        var html = '<table style="width:100%;"><tbody>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">URL</td><td style="padding-right:2px;"><input class="vpnl_editField" type="text" data-bindifhidden="false" data-bind="' + binding + '.url" style="width:100%;" value="' + params.url + '"/></td><td/></tr>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Username</td><td><input class="vpnl_editField" type="text" data-bindifhidden="false" data-bind="' + binding + '.username" value="' + params.username + '"/></td><td/></tr>';
        html += '<tr><td style="width:77px;padding-left:2px;padding-right:2px;">Password</td><td><input class="vpnl_editField" type="text" data-bindifhidden="false" data-bind="' + binding + '.password" value="' + params.password + '"/></td><td/></tr>';
        html += '</tbody></table>';
        return html;
    }
    function translateLineWeight(sVal) {
        switch (sVal) {
            case 'extrafine':
                return '1px';
            case 'fine':
                return '1.5px';
            case 'medium':
                return '2px';
            case 'thick':
                return '3px';
            case 'extrathick':
                return '4px';
            default:
                return '1px';
        }
    }
    function buildHorizontalRuleSample(elementState) {
        var html = '<div class="vpnl_textStylerShadow vpnl_textStylerSample" style="height:40px;border:solid 1px silver; border-radius:4px;"><div style="margin:0px auto;display:block;position:relative;top:20px;overflow-x:hidden;width:'
            + elementState.position.width + 'px;">'
            + '<hr style="border:'
            + elementState.lineStyle.style
            + ' ' + translateLineWeight(elementState.lineStyle.weight)
            + ' ' + elementState.lineStyle.color
            + ';width:' + elementState.position.width
            + 'px;margin-left:0px;margin-right:0px;margin-top:' + elementState.margins.top + 'px;margin-bottom:' + elementState.margins.bottom + 'px;'
            + '"/></div></div>';
        return html;
    }

    /********************************************************
     * Dialog Templates
     ********************************************************/
    function buildSelectIconDlgTemplate() {
        var html = '<div id="vpnl_SelectIconTemplate" style="display:block;width:100%;background-color:white;">'
        html += '<table style="position:relative;width:100%;height:100%"><tbody>';
        html += '<tr><td style="padding-left:4px;padding-right:4px;">Available Images</td></tr>'
        html += '<tr style="height:100%"><td style="width:100%;height:100%;">';
        html += '<div style="position:relative;width:100%;height:220px;border:solid 1px silver;">';
        html += '<div id="vpnl_iconList" style="position:relative;width:100%;height:100%;padding-right:4px;overflow-y:auto;"/>'
        html += '</div></div></td></tr></tbody></table></div>'
        return html;
    }
    function buildStateIconDlgTemplate(iconState) {
        var html = buildSelectIconDlgTemplate();
        html += '<div><table id="vpnl_iconStateConditions" style="width:100%;border:solid 1px black;margin-top:4px;"><tbody>';
        html += '<tr style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="6" style="padding-left:2px;padding-right:2px;border-bottom:solid 3px orange;">State Conditions <span style="font-style:italic;font-size:9pt;">(When all conditions are true the selected icon will be displayed in the dashboard)</span></td></tr>';
        html += '<tr style="background-color:gainsboro;"><td style="width:24px;" /><td style="width:77px;">Device</td><td style="width:230px;">Service</td><td style="width:130px;padding-left:4px;padding-right:4px;">Variable</td><td style="width:30px;padding-left:4px;padding-right:4px;">Operator</td><td style=padding-left:4px;padding-right:4px;">Value</td></tr>';
        if (iconState && iconState.conditions && iconState.conditions.length > 0) {
            for (var i = 0; i < iconState.conditions.length; i++) {
                if (typeof (iconState.conditions[i].deviceId) == 'undefined')
                    iconState.conditions[i].deviceId = device;
                html += buildStateIconCondition(iconState.conditions[i].deviceId, iconState.conditions[i].serviceId, iconState.conditions[i].variable, iconState.conditions[i].operator, iconState.conditions[i].value, i);
            }
        }
        html += '<tr><td><img style="width:22px;cursor:pointer;" onclick="virtualPanelBuilder.addIconStateConditionRow($(this).parents(\'#vpnl_iconStateConditions\'));" src="skins/default/img/other/plusOpenBtn.png"/></td><td colspan="5"><span style="white-space:nowrap;font-size:8pt;font-style:italic;">Click + to add another condition to display the selected icon.</span></td>';
        html += '</tbody></table></div>';
        return html;
    }
    function buildStateIconCondition(deviceId, sService, sVariable, sOperator, sValue, ndx) {
        var binding = '_iconState' + ndx;
        var html = '<tr>';
        html += '<td style="width:24px;"><img style="width:22px;cursor:pointer;" src="skins/default/img/other/minusLightGreen.png" onclick="$(this).parents(\'tr:first\').remove();"/></td>';
        html += '<td>' + buildAvailableDevices(deviceId, binding) + '</td>';
        html += '<td><input class="vpnl_editField" style="width:100%;" type="text" data-bind="' + binding + '.serviceId" value="' + sService + '" onchange="virtualPanelBuilder.onReplaceVariableDataList(this);" list="vpnl_datalist' + binding + '_serviceId">'
            + '<div id="vpnl_listContainer' + binding + '_serviceId">' + buildServiceIdDataList(deviceId, binding) + '</div></input></td>';
        html += '<td><input class="vpnl_editField" style="width:100%;" type="text" value="' + sVariable + '" list="vpnl_datalist' + binding + '_variable">'
            + '<div id="vpnl_listContainer' + binding + '_variable">' + buildVariableDataList(deviceId, sService, binding) + '</div></input></td>';
        html += '<td><select class="vpnl_editField" style="width:100%;">' +
            '<option value="==" ' + (!sOperator || sOperator == '==' ? 'selected="selected"' : '') + '>==</option>' +
            '<option value="&gt;=" ' + (sOperator == '>=' ? 'selected="selected"' : '') + '>&gt;=</option>' +
            '<option value="&lt;=" ' + (sOperator == '<=' ? 'selected="selected"' : '') + '>&lt;=</option>' +
            '<option value="&gt;" ' + (sOperator == '>' ? 'selected="selected"' : '') + '>&gt;</option>' +
            '<option value="&lt;" ' + (sOperator == '<' ? 'selected="selected"' : '') + '>&lt;</option>' +
        '</select></td>';
        html += '<td><input class="vpnl_editField" style="width:100%;" type="text" value="' + sValue + '"/></td>';
        return html;
    }
    function buildColorExpression(sOperator, sValue, sColor, ndx) {
        var binding = '_colorState' + ndx;
        var html = '<tr>';
        html += '<td style="width:24px;"><img style="width:22px;cursor:pointer;" src="skins/default/img/other/minusLightGreen.png" onclick="$(this).parents(\'tr:first\').remove();"/></td>';
        html += '<td><select class="vpnl_editField" style="width:100%;">' +
            '<option value="==" ' + (!sOperator || sOperator == '==' ? 'selected="selected"' : '') + '>==</option>' +
            '<option value="&gt;=" ' + (sOperator == '>=' ? 'selected="selected"' : '') + '>&gt;=</option>' +
            '<option value="&lt;=" ' + (sOperator == '<=' ? 'selected="selected"' : '') + '>&lt;=</option>' +
            '<option value="&gt;" ' + (sOperator == '>' ? 'selected="selected"' : '') + '>&gt;</option>' +
            '<option value="&lt;" ' + (sOperator == '<' ? 'selected="selected"' : '') + '>&lt;</option>' +
        '</select></td>';
        html += '<td><input class="vpnl_editField" style="width:100%;" type="text" value="' + sValue + '"/></td>';
        html += '<td><input class="vpnl_editField" style="width:100%;" list="vpnl_colordatalist" type="text" value="' + sColor + '"/></td>';
        return html;
    }

    function buildSelectControlTypes(sType) {
        var html = '';
        html += '<select id="vpnl_elemControlType" class="vpnl_editField" style="width:100%;" onchange="$(&quot;#vpnl_controlTypeDetail&quot;).html(virtualPanelBuilder.getControlTypeDetail(this.value));">' +
        '<option value="button" ' + (!sType || sType == 'button' ? 'selected="selected"' : '') + '>Button</option>' +
        '<option value="label" ' + (sType == 'label' ? 'selected="selected"' : '') + '>Label</option>' +
        '<option value="color_indicator" ' + (sType == 'color_indicator' ? 'selected="selected"' : '') + '>Color Indicator</option>' +
        '<option value="variable" ' + (sType == 'variable' ? 'selected="selected"' : '') + '>Variable</option>' +
        '<option value="multi_state_button" ' + (sType == 'multi_state_button' ? 'selected="selected"' : '') + '>Multi-state Button</option>' +
        '<option value="horizontal_rule" ' + (sType == 'horizontal_rule' ? 'selected="selected"' : '') + '>Horizontal Rule</option>' +
		'<option value="spinner_horizontal" ' + (sType == 'spinner_horizontal' ? 'selected="selected"' : '') + '>Horizontal Spinner</option>' +
        '<option value="spacer" ' + (sType == 'spacer' ? 'selected="selected"' : '') + '>Horizontal Spacer</option>' +
        '<option value="line_break" ' + (sType == 'line_break' ? 'selected="selected"' : '') + '>Line Break</option>' +
        '<option value="group_start" ' + (sType == 'line_break' ? 'selected="selected"' : '') + '>Group Start</option>' +
        //'<option value="input" ' + (sType == 'input' ? 'selected="selected"' : '') + '>Input Field</option>' +
        //'<option value="checkbox" ' + (sType == 'checkbox' ? 'selected="selected"' : '') + '>Checkbox</option>' +
        '<option value="slider" ' + (sType == 'checkbox' ? 'selected="selected"' : '') + '>Horizontal Slider</option>' +
        '<option value="slider_vertical" ' + (sType == 'checkbox' ? 'selected="selected"' : '') + '>Vertical Slider</option>' +
        '</select>';
        html += '<div id="vpnl_controlTypeDetail">' + getControlTypeDetail((!sType) ? 'button' : sType); + '</div>';
        return html;
    }
    function getControlTypeDetail(sType) {
        var html = '<div style="padding:4px;">';
        switch (sType) {
            case 'button':
                html += '<span>Displays a button that you can use to turn items on or off.  Buttons contain several styles that will allow you to toggle items or send a single command.</span>';
                break;
            case 'label':
                html += '<span>Displays static text on the panel.  You can set the display style including font, size, and color for the text.<span>';
                break;
            case 'variable':
                html += '<span>Displays variable data from a device.  You can set the display style including font, size, and color for the text.<span>';
                break;
            case 'multi_state_button':
                html += '<span>Displays an on/off style toggle button.  Mult-state buttons allow you to change the text and the commands issued for the on or off state.  You can set the display style including font, size, and color for the text.<span>';
                break;
            case 'horizontal_rule':
                html += '<span>Displays a horizontal line on the panel.  You can change the line style as well as the weight of the line<span>';
                break;
            case 'spacer':
                html += 'Adds horizontal space.  When added between elements the element after the spacer will be proportionally spaced to the right. This only has an effect in the Dashboard View.';
                break;
            case 'line_break':
                html += 'Forces the element immediately following the line break to start below the previous element.  This only has an effect in the dasboard view.';
                break;
            case 'group_start':
                html += 'Starts a new group of elements and ends any previously defined group of the same type.  This is either a control group or scene group.';
                break;
            case 'slider':
                html += 'Creates a horizontal slider that can be used for dimming controls';
                break;
            case 'slider_vertical':
                html += 'Creates a vertical slider such as those used by thermostats';
                break;
            case 'color_indicator':
                html += 'Displays a read-only indicator that reflects the state of a variable based upon a series of expressions.  Different colors can be used to display the state.';
                break;
        }
        return html + "</div>";
    }
    function addIconStateConditionRow(tbl) { tbl.find('tr:last').before(buildStateIconCondition(device, '', '', '==', '', tbl[0].rows.length - 2)); }
    function addColorExpressionRow(tbl) { tbl.find('tr:last').before(buildColorExpression('==', '', 'green', tbl[0].rows.length - 2)); }
    function buildIconList(configuration, sSelected) {
        var sHtml = '<table style="width:100%;"><tbody>';
        for (var i = 0; i < configuration.length; i++) {
            sHtml += '<tr class="' + ((sSelected == configuration[i]) ? 'vpnl_iconListSelected' : '') + '" style="cursor:pointer;'
            sHtml += (sSelected == configuration[i]) ? 'background-color:steelblue;color:white;" data-selected="true"' : '" data-selected="false" ';
            sHtml += 'onclick="virtualPanelBuilder.setListSelected($(this).parents(\'table\')[0], this.rowIndex);" onmouseover="if($(this).data(\'selected\') != true) this.style.backgroundColor = \'lightsteelblue\';" onmouseout="if($(this).data(\'selected\') != true) this.style.backgroundColor=\'\'"><td><div style="display:inline-block;border:outset 2px;border-radius:5px;padding:4px;background-color:white;">';
            sHtml += '<img style="width:32px;" src="skins/default/img/devices/device_states/' + configuration[i] + '" /></div></td>'
            sHtml += '<td style="width:100%;"><span style="font-size:14pt;">' + configuration[i] + '</span></td>';
            sHtml += '</tr>';
        }
        sHtml += '</tbody></table>';
        return sHtml;
    }
    function buildEditTriggerTemplate(state) {
        state = {
            ordinal: -1, description: '', variable: { deviceId: device, serviceId: '', variable: '' },
            arguments: []
        }

        var html = '<div style="">';
        html += '<table style="width:100%;"><tbody>';
        html += '<tr style=""><td style="width:77px;padding-left:2px;padding-right:2px;">Name</td>';
        html += '<td style="padding-left:2px;padding-right:2px;padding-bottom:2px;"><input data-bind=".description" class="vpnl_editField" style="width:270px;" type="text" value="' + state.description + '"/></td><td/></tr>';
        html += '<tr class="vpnl_dialogSubHeader" style="background-color:steelblue;color:white;font-weight:bold;"><td colspan="3" style="padding-left:4px;">Trigger Variable</td></tr>';
        html += '<td colspan="3" style="padding-bottom:4px;">' + buildDeviceVariable('.variable',
            state.variable.deviceId,
            state.variable.serviceId,
            state.variable.variable, null, null) + '</td></tr>';
        html += '</tbody></table></div>'
        return html;
    }

    /********************************************************
     * Event Handlers
     ********************************************************/
    function onClickSelectDefaultIcon() {
        modalDialog.showModalDialog({
            title: "Select a Default Icon for the Control Panel",
            width: "450px",
            content: buildSelectIconDlgTemplate(),
            init: function () {
                loadIcons('vpnl_iconList', panelConfig.icon);
            },
            buttons: [
                { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                {
                    text: "Select Icon", event: function () {
                        var row = modalDialog.getSelectedRow($('#vpnl_iconList')[0]);
                        if (row) {
                            panelConfig.icon = $(row.cells[1]).text();
                            $('#vpnl_panelIcon')[0].src = 'skins/default/img/devices/device_states/' + panelConfig.icon;
                        }
                        saveIcons();
                    }
                }
            ]
        });
    }
    function onClickCreateStateIcon() { onClickEditStateIcon(-1); }
    function onClickEditStateIcon(index) {
        var iconState = null;
        if(index != -1)
            iconState = panelConfig.stateIcons[index];
        modalDialog.showModalDialog({
            title: "Define an Icon to display when the conditions are true",
            width: "777px",
            content: buildStateIconDlgTemplate(iconState),
            init: function () {
                loadIcons('vpnl_iconList', iconState ? iconState.icon : '');
                if (index != -1)
                    this.buttons.push({
                        text: "Delete Icon", event: function () {
                            panelConfig.stateIcons.splice(index, 1);
                            $('#vpnl_stateIconDisplay').html(buildStateIconConfig(panelConfig));
                            saveIcons();
                            modalDialog.closeDialog();
                        }
                    });

            },
            buttons: [
                { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                {
                    text: "Save Icon", event: function () {
                        $('#vpnl_modalStatusBar').html('');
                        var row = modalDialog.getSelectedRow($('#vpnl_iconList')[0]);
                        if (row) {
                            var o = new Object();
                            o.icon = $(row.cells[1]).text();
                            var tbl = $('#vpnl_iconStateConditions')[0];

                            o.conditions = new Array();
                            for (var i = 2; i < tbl.rows.length - 1; i++) {
                                var r = tbl.rows[i];
                                var c = new Object();
                                c.deviceId = $(r.cells[1]).children('select').val();
                                c.serviceId = $(r.cells[2]).children('input').val();
                                c.variable = $(r.cells[3]).children('input').val();
                                c.operator = $(r.cells[4]).children('select').val();
                                c.value = $(r.cells[5]).children('input').val();
                                if (c.serviceId.length > 0 && c.variable.length > 0 && c.value.length > 0)
                                    o.conditions.push(c);
                            }
                            if (o.conditions.length > 0) {
                                if (!panelConfig.stateIcons) panelConfig.stateIcons = new Array();
                                if (index == -1)
                                    panelConfig.stateIcons.push(o);
                                else
                                    panelConfig.stateIcons[index] = o;
                                saveIcons();
                            }
                            else {
                                $('#vpnl_modalStatusBar').html('You must define at least 1 valid condition before saving.');
                            }
                        }
                        else {
                            $('#vpnl_modalStatusBar').html('You must select an icon to display before saving.');
                        }
                    }
                }

            ]
        });
    }
    function onDeleteControlElement(ndx) {
        var el = panelConfig.elements[ndx];
        modalDialog.showMessageDialog({
            title: 'Confirm Delete (' + buildControlText(el) + ')',
            message: 'Are you sure you want to delete (' + buildControlText(el) + ')?  If so press the Delete button otherwise press Cancel. <p style="font-size:8pt">',
            width: '500px',
            buttons: [{ text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                {
                    text: 'Delete ' + buildControlText(el),
                    event: function () {
                        panelConfig.elements.splice(ndx, 1);
                        $('#vpnl_panelElements').html(buildPanelElements(panelConfig));
                        deleteElement(el.ordinal);
                        modalDialog.closeDialog();
                    }
                }

            ]
        });
    }
    function onEditControlElement(ndx) {
        var elementState = panelConfig.elements[ndx];
        switch (elementState.controlType) {
            case 'color_indicator':
                modalDialog.showModalDialog({
                    title: "Edit Color Indicator",
                    width: "470px",
                    content: buildEditColorIndicatorElementTemplate(elementState),
                    init: function () {
                        $('.vpnl_variableName').on('change', function (e) {
                            var v = dialogManager.fromDialog();
                            textStyler.updateSampleText('vpnl_labelText',
                                textStyler.transformVariableFormat(v.textStyle, v.display.variable));
                        });
                        $('.vpnl_variableFormat').on('change', function (e) {
                            var v = dialogManager.fromDialog();
                            textStyler.updateSampleText('vpnl_labelText',
                                textStyler.transformVariableFormat(v.textStyle, v.display.variable));
                        });
                        var exprInfo = dialogManager.createTextPopup($('#vpnl_expressionInfo')[0], 'vpnl_expressionInfo',
                            buildColorExpressionInfo());
                    },
                    params: { elementIndex: ndx },
                    buttons: [
                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                        {
                            text: "Save Indicator", event: function () {
                                var state = dialogManager.fromDialog();
                                state.text = state.display.variable;
                                state.controlType = elementState.controlType;
                                state.expressions = new Array();
                                var tbl = $('#vpnl_colorExpressions')[0];
                                for (var i = 2; i < tbl.rows.length - 1; i++) {
                                    state.expressions.push({
                                        operator: $(tbl.rows[i].cells[1].children[0]).val(),
                                        value: $(tbl.rows[i].cells[2].children[0]).val(),
                                        color: $(tbl.rows[i].cells[3].children[0]).val()
                                    });
                                }

                                saveElement(ndx, state);
                            }
                        },
                        {
                            text: "Add Indicator", align: 'left', event: function () {
                                var state = dialogManager.fromDialog();
                                state.controlType = elementState.controlType;
                                state.ordinal = -1;
                                state.expressions = new Array();
                                var tbl = $('#vpnl_colorExpressions')[0];
                                for (var i = 2; i < tbl.rows.length - 1; i++) {
                                    state.expressions.push({
                                        operator: $(tbl.rows[i].cells[1].children[0]).val(),
                                        value: $(tbl.rows[i].cells[2].children[0]).val(),
                                        color: $(tbl.rows[i].cells[3].children[0]).val()
                                    });
                                }
                                addElement(state);
                            }
                        }

                    ]
                });
                break;
            case 'button':
                modalDialog.showModalDialog({
                    title: "Edit Button",
                    width: elementState.buttonStyle == 'toggle' || elementState.buttonStyle == 'persistenttoggle' ? "777px" : "527px",
                    content: buildEditButtonElementTemplate(elementState),
                    init: function () {
                        var ctxInfo = dialogManager.createTextPopup($('#vpnl_luaContextInfo')[0], 'vpnl_luaContextInfo',
                            buildLuaContextInfo());
                    },
                    params: { elementIndex: ndx },
                    buttons: [
                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                        {
                            text: "Save Button", event: function () {
                                var state = dialogManager.fromDialog();
                                state.controlType = elementState.controlType;
                                saveElement(ndx, state);
                            }
                        },
                        {
                            text: "Add Button", align: 'left', event: function () {
                                var state = dialogManager.fromDialog();
                                state.text = state.display.variable;
                                state.controlType = elementState.controlType;
                                state.ordinal = -1;
                                addElement(state);
                            }
                        }
                    ]
                });
                break;
            case 'label':
                modalDialog.showModalDialog({
                    title: "Edit Label",
                    width: "470px",
                    content: buildEditLabelElementTemplate(elementState),
                    init: function () { },
                    params: { elementIndex: ndx },
                    buttons: [
                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                        {
                            text: "Save Label", event: function () {
                                var state = dialogManager.fromDialog();
                                state.textStyle.width = state.position.width;
                                state.controlType = elementState.controlType;
                                saveElement(ndx, state);
                            }
                        },
                        { 
                            text: "Add Label", align: 'left', event: function () {
                                var state = dialogManager.fromDialog();
                                state.textStyle.width = state.position.width;
                                state.controlType = elementState.controlType;
                                state.ordinal = -1;
                                addElement(state);
                            }
                        }
                    
                    ]
                });
                break;
            case 'spacer':
                modalDialog.showModalDialog({
                    title: "Edit Spacer",
                    width: "370px",
                    content: buildEditSpacerElementTemplate(elementState),
                    init: function () { },
                    params: { elementIndex: ndx },
                    buttons: [
                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                        {
                            text: "Save Spacer", event: function () {
                                var state = dialogManager.fromDialog();
                                state.controlType = elementState.controlType;
                                saveElement(ndx, state);
                            }
                        },
                        {
                            text: "Add Spacer", align: 'left', event: function () {
                                var state = dialogManager.fromDialog();
                                state.controlType = elementState.controlType;
                                state.ordinal = -1;
                                addElement(state);
                            }
                        }

                    ]
                });
                break;
            case 'group_start':
                modalDialog.showModalDialog({
                    title: "Edit Grouping",
                    width: "370px",
                    content: buildEditGroupStartElementTemplate(elementState),
                    init: function () { },
                    params: { elementIndex: ndx },
                    buttons: [
                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                        {
                            text: "Save Grouping", event: function () {
                                var state = dialogManager.fromDialog();
                                state.controlType = elementState.controlType;
                                saveElement(ndx, state);
                            }
                        }
                    ]
                });
                break;
            case 'horizontal_rule':
                modalDialog.showModalDialog({
                    title: "Edit Horizontal Rule",
                    width: "470px",
                    content: buildEditHorizontalRuleElementTemplate(elementState),
                    init: function () { },
                    params: { elementIndex: ndx },
                    buttons: [
                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                        {
                            text: "Save Horizontal Rule", event: function () {
                                var state = dialogManager.fromDialog();
                                state.controlType = elementState.controlType;
                                saveElement(ndx, state);
                            }
                        },
                        {
                            text: "Add Horizontal Rule", align: 'left', event: function () {
                                var state = dialogManager.fromDialog();
                                state.controlType = elementState.controlType;
                                state.ordinal = -1;
                                addElement(state);
                            }

                        }
                    ]
                });
                break;
            case 'variable':
                modalDialog.showModalDialog({
                    title: "Edit Variable",
                    width: "470px",
                    content: buildEditVariableElementTemplate(elementState),
                    init: function () {
                        $('.vpnl_variableName').on('change', function (e) {
                            var v = dialogManager.fromDialog();
                            textStyler.updateSampleText('vpnl_labelText', 
                                textStyler.transformVariableFormat(v.textStyle, v.display.variable));
                        });
                        $('.vpnl_variableFormat').on('change', function (e) {
                            var v = dialogManager.fromDialog();
                            textStyler.updateSampleText('vpnl_labelText',
                                textStyler.transformVariableFormat(v.textStyle, v.display.variable));
                        });
                        var fmtInfo = dialogManager.createTextPopup($('#vpnl_variableFormatInfo')[0], 'vpnl_variableFormatInfo',
                            buildVariableFormatInfo());
                    },
                    params: { elementIndex: ndx },
                    buttons: [
                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                        {
                            text: "Save Variable", event: function () {
                                var state = dialogManager.fromDialog();
                                state.textStyle.width = state.position.width;
                                state.text = state.display.variable;
                                state.controlType = elementState.controlType;
                                saveElement(ndx, state);
                            }
                        },
                        {
                            text: "Add Variable", align: 'left', event: function () {
                                var state = dialogManager.fromDialog();
                                state.textStyle.width = state.position.width;
                                state.controlType = elementState.controlType;
                                state.ordinal = -1;
                                addElement(state);
                            }
                        }

                    ]
                });
                break;
            case 'multi_state_button':
                modalDialog.showModalDialog({
                    title: "Edit Toggle Button",
                    width: "747px",
                    content: buildEditToggleElementTemplate(elementState),
                    init: function () {
                        var ctxInfo = dialogManager.createTextPopup($('#vpnl_luaContextInfo')[0], 'vpnl_luaContextInfo',
                            buildLuaContextInfo());
                    },
                    params: { elementIndex: ndx },
                    buttons: [
                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                        {
                            text: "Save Toggle Button", event: function () {
                                var state = dialogManager.fromDialog();
                                state.controlType = elementState.controlType;
                                state.text = state.onState.text + ' / ' + state.offState.text;
                                saveElement(ndx, state);
                            }
                        },
                        {
                            text: "Add Toggle Button", align: 'left', event: function () {
                                var state = dialogManager.fromDialog();
                                state.text = state.onState.text + ' / ' + state.offState.text;
                                state.controlType = elementState.controlType;
                                state.ordinal = -1;
                                addElement(state);
                            }
                        }

                    ]
                });
                break;
            case 'slider':
                modalDialog.showModalDialog({
                    title: "Edit Slider",
                    width: "470px",
                    content: buildEditSliderElementTemplate(elementState),
                    init: function () {
                        var ctxInfo = dialogManager.createTextPopup($('#vpnl_luaContextInfo')[0], 'vpnl_luaContextInfo',
                            buildLuaContextInfo());
                    },
                    params: { elementIndex: ndx },
                    buttons: [
                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                        {
                            text: "Save Slider", event: function () {
                                var state = dialogManager.fromDialog();
                                state.controlType = elementState.controlType;
                                state.text = state.display.variable;
                                console.log(state);
                                saveElement(ndx, state);
                            }
                        },
                        {
                            text: "Add Slider", align: 'left', event: function () {
                                var state = dialogManager.fromDialog();
                                state.text = state.display.variable;
                                state.controlType = elementState.controlType;
                                state.ordinal = -1;
                                addElement(state);
                            }
                        }

                    ]
                });
                break;
            case 'slider_vertical':
                modalDialog.showModalDialog({
                    title: "Edit Vertical Slider",
                    width: "470px",
                    content: buildEditVerticalSliderElementTemplate(elementState),
                    init: function () {
                        var ctxInfo = dialogManager.createTextPopup($('#vpnl_luaContextInfo')[0], 'vpnl_luaContextInfo',
                            buildLuaContextInfo());
                    },
                    params: { elementIndex: ndx },
                    buttons: [
                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                        {
                            text: "Save Vertical Slider", event: function () {
                                var state = dialogManager.fromDialog();
                                state.controlType = elementState.controlType;
                                state.text = state.display.variable;
                                saveElement(ndx, state);
                            }
                        },
                        {
                            text: "Add Vertical Slider", align: 'left', event: function () {
                                var state = dialogManager.fromDialog();
                                state.text = state.display.variable;
                                state.controlType = elementState.controlType;
                                state.ordinal = -1;
                                addElement(state);
                            }
                        }

                    ]
                });
                break;
        	case 'spinner_horizontal':
        		modalDialog.showModalDialog({
        			title: "Edit Horizontal Spinner",
        			width: "470px",
        			content: buildEditHorizontalSpinnerElementTemplate(elementState),
        			init: function () {
        				var ctxInfo = dialogManager.createTextPopup($('#vpnl_luaContextInfo')[0], 'vpnl_luaContextInfo',
                            buildLuaContextInfo());
        			},
        			params: { elementIndex: ndx },
        			buttons: [
                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                        {
                        	text: "Save Horizontal Spinner", event: function () {
                        		var state = dialogManager.fromDialog();
                        		state.controlType = elementState.controlType;
                        		state.text = state.display.variable;
                        		saveElement(ndx, state);
                        	}
                        },
                        {
                        	text: "Add Horizontal Spinner", align: 'left', event: function () {
                        		var state = dialogManager.fromDialog();
                        		state.text = state.display.variable;
                        		state.controlType = elementState.controlType;
                        		state.ordinal = -1;
                        		addElement(state);
                        	}
                        }

        			]
        		});
        		break;

        }
    }
    function onAddControlElement() {
        modalDialog.showModalDialog({
            title: "Select the type of element to add",
            width: "370px",
            content: buildSelectControlTypes(),
            init: function () { },
            buttons: [
                { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                {
                    text: "Select Control Type", event: function () {
                        var elementState = {};
                        elementState.controlType = $('#vpnl_elemControlType').val();
                        modalDialog.closeDialog();
                        switch (elementState.controlType) {
                            case 'button':
                                elementState = {
                                    controlType: elementState.controlType,
                                    text: (panelConfig.elements && panelConfig.elements.length) ? 'Button ' + (panelConfig.elements.length + 1).toString() : 'Button 1',
                                    ordinal: -1, position: { left: 0, top: 0, width: 77, height: 20 }, buttonStyle: 'momentary',
                                    display: { deviceId: device, serviceId: '', variable: '', value: '' },
                                    command: {
                                        cmdType: 'NoAction',
                                        variable: { deviceId: device, serviceId: '', variable: '', value: '' },
                                        action: { deviceId: device, serviceId: '', action: '', parameters: [] },
                                        scene: { sceneId: 0 },
                                        http: { url: '', username: '', password: '' }
                                    }
                                };
                                modalDialog.showModalDialog({
                                    title: "Add New Button",
                                    width: "527px",
                                    content: buildEditButtonElementTemplate(elementState),
                                    init: function () {
                                        var ctxInfo = dialogManager.createTextPopup($('#vpnl_luaContextInfo')[0], 'vpnl_luaContextInfo',
                                            buildLuaContextInfo());
                                    },
                                    buttons: [
                                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                                        {
                                            text: "Add Button", event: function () {
                                                var state = dialogManager.fromDialog();
                                                state.controlType = elementState.controlType;
                                                addElement(state);
                                            }
                                        }
                                    ]
                                });
                                break;
                            case 'line_break':
                                elementState = {
                                    controlType: elementState.controlType,
                                    ordinal: -1
                                };
                                addElement(elementState);
                                break;
                            case 'group_start':
                                elementState = {
                                    controlType: elementState.controlType,
                                    ordinal: -1,
                                    isSceneGroup: false,
                                    lineBreak: false
                                };
                                modalDialog.showModalDialog({
                                    title: "Add New Grouping",
                                    width: "370px",
                                    content: buildEditGroupStartElementTemplate(elementState),
                                    init: function () { },
                                    buttons: [
                                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                                        {
                                            text: "Add Grouping", event: function () {
                                                var state = dialogManager.fromDialog();
                                                state.controlType = elementState.controlType;
                                                addElement(state);
                                            }
                                        }
                                    ]
                                });
                                break;
                            case 'spacer':
                                elementState = {
                                    controlType: elementState.controlType,
                                    text: "---- SPACER ----",
                                    ordinal: -1, position: { left: 0, top: 0, width: 5, height: 0 }
                                };
                                modalDialog.showModalDialog({
                                    title: "Add New Spacer",
                                    width: "370px",
                                    content: buildEditSpacerElementTemplate(elementState),
                                    init: function () { },
                                    buttons: [
                                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                                        {
                                            text: "Add Spacer", event: function () {
                                                var state = dialogManager.fromDialog();
                                                state.controlType = elementState.controlType;
                                                addElement(state);
                                            }
                                        }
                                    ]
                                });
                                break;
                            case 'horizontal_rule':
                                elementState = {
                                    controlType: elementState.controlType,
                                    text: "Horizontal Rule",
                                    ordinal: -1, position: { left: 0, top: 0, width: 47, height: 4 },
                                    lineStyle: { width: 'fine', style: 'solid', color: '' },
                                    margins: { left: 0, top: 0, right: 0, bottom: 0 }
                                };
                                modalDialog.showModalDialog({
                                    title: "Add New Horizontal Rule",
                                    width: "470px",
                                    content: buildEditHorizontalRuleElementTemplate(elementState),
                                    init: function () { },
                                    buttons: [
                                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                                        {
                                            text: "Add Horizontal Rule", event: function () {
                                                var state = dialogManager.fromDialog();
                                                state.controlType = elementState.controlType;
                                                addElement(state);
                                            }
                                        }
                                    ]
                                });
                                break;

                            case 'label':
                                elementState = {
                                    controlType: elementState.controlType,
                                    text: (panelConfig.elements && panelConfig.elements.length) ? 'Label ' + (panelConfig.elements.length + 1).toString() : 'Label 1',
                                    ordinal: -1, position: { left: 0, top: 0, width: 77, height: 20 }
                                };
                                modalDialog.showModalDialog({
                                    title: "Add New Label",
                                    width: "470px",
                                    content: buildEditLabelElementTemplate(elementState),
                                    init: function () { },
                                    buttons: [
                                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                                        {
                                            text: "Add Label", event: function () {
                                                var state = dialogManager.fromDialog();
                                                state.textStyle.width = state.position.width;
                                                state.controlType = elementState.controlType;
                                                addElement(state);
                                            }
                                        }
                                    ]
                                });
                                break;
                            case 'color_indicator':
                                elementState = {
                                    controlType: elementState.controlType,
                                    defaultColor: 'gray',
                                    expressions: [],
                                    ordinal: -1, position: { left: 0, top: 0, width: 10, height: 10 },
                                    display: { deviceId: device, serviceId: '', variable: '', value: '' }
                                };
                                modalDialog.showModalDialog({
                                    title: "Add New Color Indicator",
                                    width: "470px",
                                    content: buildEditColorIndicatorElementTemplate(elementState),
                                    init: function () {
                                        $('.vpnl_variableName').on('change', function (e) {
                                            var v = dialogManager.fromDialog();
                                            textStyler.updateSampleText('vpnl_labelText',
                                                textStyler.transformVariableFormat(v.textStyle, v.display.variable));
                                        });
                                        var fmtInfo = dialogManager.createTextPopup($('#vpnl_variableFormatInfo')[0], 'vpnl_variableFormatInfo',
                                            buildVariableFormatInfo());
                                    },
                                    buttons: [
                                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                                        {
                                            text: "Add Indicator", event: function () {
                                                var state = dialogManager.fromDialog();
                                                state.text = state.display.variable;
                                                state.controlType = elementState.controlType;
                                                state.expressions = new Array();
                                                var tbl = $('#vpnl_colorExpressions')[0];
                                                for (var i = 2; i < tbl.rows.length - 1; i++) {
                                                    state.expressions.push({
                                                        operator: $(tbl.rows[i].cells[1].children[0]).val(),
                                                        value: $(tbl.rows[i].cells[2].children[0]).val(),
                                                        color: $(tbl.rows[i].cells[3].children[0]).val()
                                                    });
                                                }
                                                addElement(state);
                                            }
                                        }
                                    ]
                                });
                                break;

                            case 'variable':
                                elementState = {
                                    controlType: elementState.controlType,
                                    text: (panelConfig.elements && panelConfig.elements.length) ? 'Variable ' + (panelConfig.elements.length + 1).toString() : 'Variable 1',
                                    ordinal: -1, position: { left: 0, top: 0, width: 77, height: 20 },
                                    display: { deviceId: device, serviceId: '', variable: '', value: '' }
                                };
                                modalDialog.showModalDialog({
                                    title: "Add New Variable",
                                    width: "470px",
                                    content: buildEditVariableElementTemplate(elementState),
                                    init: function () {
                                        $('.vpnl_variableName').on('change', function (e) {
                                            var v = dialogManager.fromDialog();
                                            textStyler.updateSampleText('vpnl_labelText',
                                                textStyler.transformVariableFormat(v.textStyle, v.display.variable));
                                        });
                                        $('.vpnl_variableFormat').on('change', function (e) {
                                            var v = dialogManager.fromDialog();
                                            textStyler.updateSampleText('vpnl_labelText',
                                                textStyler.transformVariableFormat(v.textStyle, v.display.variable));
                                        });
                                        var fmtInfo = dialogManager.createTextPopup($('#vpnl_variableFormatInfo')[0], 'vpnl_variableFormatInfo',
                                            buildVariableFormatInfo());
                                    },
                                    buttons: [
                                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                                        {
                                            text: "Add Variable", event: function () {
                                                var state = dialogManager.fromDialog();
                                                state.textStyle.width = state.position.width;
                                                state.text = state.display.variable;
                                                state.controlType = elementState.controlType;
                                                addElement(state);
                                            }
                                        }
                                    ]
                                });
                                break;
                            case 'multi_state_button':
                                elementState = {
                                    controlType: elementState.controlType,
                                    text: 'On/Off',
                                    ordinal: -1, position: { left: 0, top: 0, width: 47, height: 20 },
                                    display: { deviceId: device, serviceId: '', variable: '', value: '' },
                                    offState: {
                                        text: 'OFF',
                                        display: { deviceId: device, serviceId: '', variable: '', value: '' },
                                        command: {
                                            cmdType: 'NoAction',
                                            variable: { deviceId: device, serviceId: '', variable: '', value: '' },
                                            action: { deviceId: device, serviceId: '', action: '', parameters: [] },
                                            scene: { sceneId: 0 },
                                            http: { url: '', username: '', password: '' }
                                        }
                                    },
                                    onState: {
                                        text: 'ON',
                                        display: { deviceId: device, serviceId: '', variable: '', value: '' },
                                        command: {
                                            cmdType: 'NoAction',
                                            variable: { deviceId: device, serviceId: '', variable: '', value: '' },
                                            action: { deviceId: device, serviceId: '', action: '', parameters: [] },
                                            scene: { sceneId: 0 },
                                            http: { url: '', username: '', password: '' }
                                        }
                                    }

                                };
                                modalDialog.showModalDialog({
                                    title: "Add New Toggle Button",
                                    width: "747px",
                                    content: buildEditToggleElementTemplate(elementState),
                                    init: function () {
                                        var ctxInfo = dialogManager.createTextPopup($('#vpnl_luaContextInfo')[0], 'vpnl_luaContextInfo',
                                            buildLuaContextInfo());
                                    },
                                    buttons: [
                                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                                        {
                                            text: "Add Toggle Button", event: function () {
                                                var state = dialogManager.fromDialog();
                                                state.text = state.onState.text + ' / ' + state.offState.text;
                                                state.controlType = elementState.controlType;
                                                addElement(state);
                                            }
                                        }
                                    ]
                                });
                                break;
                            case 'slider':
                                elementState = {
                                    controlType: elementState.controlType,
                                    text: 'Slider',
                                    ordinal: -1, position: { left: 0, top: 0, width: 47, height: 20 },
                                    display: { deviceId: device, serviceId: '', variable: '' },
                                    style: 'numeric',
                                    values: {
                                        min: { text: 'OFF', value: 0 },
                                        max: { text: 'ON', value: 100 }
                                    },
                                    command: {
                                        cmdType: 'NoAction',
                                        variable: { deviceId: device, serviceId: '', variable: '', value: '' },
                                        action: { deviceId: device, serviceId: '', action: '', parameters: [] },
                                        scene: { sceneId: 0 },
                                        http: { url: '', username: '', password: '' }
                                    }
                                };
                                modalDialog.showModalDialog({
                                    title: "Add New Slider",
                                    width: "470px",
                                    content: buildEditSliderElementTemplate(elementState),
                                    init: function () {
                                        var ctxInfo = dialogManager.createTextPopup($('#vpnl_luaContextInfo')[0], 'vpnl_luaContextInfo',
                                            buildLuaContextInfo());
                                    },
                                    buttons: [
                                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                                        {
                                            text: "Add Slider", event: function () {
                                                var state = dialogManager.fromDialog();
                                                state.text = state.display.variable;
                                                state.controlType = elementState.controlType;
                                                addElement(state);
                                            }
                                        }
                                    ]
                                });
                                break;
                            case 'slider_vertical':
                                elementState = {
                                    controlType: elementState.controlType,
                                    text: 'Vertical Slider',
                                    ordinal: -1, position: { left: 0, top: 0, width: 20, height: 47 },
                                    display: { deviceId: device, serviceId: '', variable: '' },
                                    style: 'numeric',
                                    command: {
                                        cmdType: 'NoAction',
                                        variable: { deviceId: device, serviceId: '', variable: '', value: '' },
                                        action: { deviceId: device, serviceId: '', action: '', parameters: [] },
                                        scene: { sceneId: 0 },
                                        http: { url: '', username: '', password: '' }
                                    }
                                };
                                modalDialog.showModalDialog({
                                    title: "Add New Vertical Slider",
                                    width: "470px",
                                    content: buildEditVerticalSliderElementTemplate(elementState),
                                    init: function () {
                                        var ctxInfo = dialogManager.createTextPopup($('#vpnl_luaContextInfo')[0], 'vpnl_luaContextInfo',
                                            buildLuaContextInfo());
                                    },
                                    buttons: [
                                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                                        {
                                            text: "Add Vertical Slider", event: function () {
                                                var state = dialogManager.fromDialog();
                                                state.text = state.display.variable;
                                                state.controlType = elementState.controlType;
                                                addElement(state);
                                            }
                                        }
                                    ]
                                });
                                break;
                        	case 'spinner_horizontal':
                        		elementState = {
                        			controlType: elementState.controlType,
                        			text: 'Vertical Slider',
                        			ordinal: -1, position: { left: 0, top: 0, width: 64, height: 52 },
                        			display: { deviceId: device, serviceId: '', variable: '' },
                        			style: 'numeric',
                        			command: {
                        				cmdType: 'NoAction',
                        				variable: { deviceId: device, serviceId: '', variable: '', value: '' },
                        				action: { deviceId: device, serviceId: '', action: '', parameters: [] },
                        				scene: { sceneId: 0 },
                        				http: { url: '', username: '', password: '' }
                        			}
                        		};
                        		modalDialog.showModalDialog({
                        			title: "Add New Horizontal Spinner",
                        			width: "470px",
                        			content: buildEditHorizontalSpinnerElementTemplate(elementState),
                        			init: function () {
                        				var ctxInfo = dialogManager.createTextPopup($('#vpnl_luaContextInfo')[0], 'vpnl_luaContextInfo',
                                            buildLuaContextInfo());
                        			},
                        			buttons: [
                                        { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                                        {
                                        	text: "Add Horizontal Spinner", event: function () {
                                        		var state = dialogManager.fromDialog();
                                        		state.text = state.display.variable;
                                        		state.controlType = elementState.controlType;
                                        		addElement(state);
                                        	}
                                        }
                        			]
                        		});
                        		break;

                        }
                    }
                }
            ]
        });
    }
    function onAddTrigger() {
        modalDialog.showModalDialog({
            title: "Add New Trigger",
            width: "470px",
            content: buildEditTriggerTemplate({
                ordinal: -1, description: '', variable: { deviceId: device, serviceId: '', variable: '' },
                arguments: []
                }),
            init: function () { },
            buttons: [
                { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                {
                    text: "Add Trigger", event: function () {
                        var state = dialogManager.fromDialog();
                        addTrigger(state);
                    }
                }
            ]
        });
    }
    function onEditTrigger(ndx) {
        modalDialog.showModalDialog({
            title: "Edit Trigger",
            width: "470px",
            content: buildEditTriggerTemplate(panelConfig.triggers[ndx]),
            init: function () { },
            buttons: [
                { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                {
                    text: "Add Trigger", event: function () {
                        var state = dialogManager.fromDialog();
                        addTrigger(state);
                    }
                }
            ]
        });
    }
    function onAddState() {
        modalDialog.showModalDialog({
            title: "Add New State Variable",
            width: "470px",
            content: buildEditTriggerTemplate({
                ordinal: -1, description: '', variable: { deviceId: device, serviceId: '', variable: '' },
                arguments: []
            }),
            init: function () { },
            buttons: [
                { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                {
                    text: "Add Trigger", event: function () {
                        var state = dialogManager.fromDialog();
                        addTrigger(state);
                    }
                }
            ]
        });
    }
    function onEditState(ndx) {
        modalDialog.showModalDialog({
            title: "Edit State Variable",
            width: "470px",
            content: buildEditTriggerTemplate(panelConfig.triggers[ndx]),
            init: function () { },
            buttons: [
                { text: "Cancel", event: function () { modalDialog.closeDialog(); } },
                {
                    text: "Add Trigger", event: function () {
                        var state = dialogManager.fromDialog();
                        addTrigger(state);
                    }
                }
            ]
        });
    }

    function onReplaceServiceIdDataList(elDeviceId) {
        // Build out a datalist for the the selected device id.
        var binding = $(elDeviceId).data('bind');
        if (!binding || binding.length == 0 || binding.lastIndexOf('.') == -1) return;
        binding = binding.substring(0, binding.lastIndexOf('.'));
        var deviceId = elDeviceId.value;
        // Find the input field with the associated binding.
        $('#vpnl_modalDialog').find('input')
            .each(function () {
                if ($(this).data('bind') == binding + '.serviceId') {
                    // Ok we have the service id.
                    $('#vpnl_listContainer' + binding.replace(/\./g, '_') + '_serviceId').html(buildServiceIdDataList(deviceId, binding));

                    // While we are here we shoudl also replace any dependent variables.
                    if ($('#vpnl_listContainer' + binding.replace(/\./g, '_') + '_variable').length > 0) {
                        $('#vpnl_listContainer' + binding.replace(/\./g, '_') + '_variable').html(buildVariableDataList(deviceId, this.value, binding));
                    }
                    return false;
                }
                return true;
            });
    }
    function onReplaceVariableDataList(elServiceId) {
        var binding = $(elServiceId).data('bind');
        if (!binding || binding.length == 0 || binding.lastIndexOf('.') == -1) return;
        binding = binding.substring(0, binding.lastIndexOf('.'));

        var serviceId = elServiceId.value;
        $('#vpnl_modalDialog').find('select')
            .each(function () {
                if ($(this).data('bind') == binding + '.deviceId') {
                    // Ok we have the service id.
                    var deviceId = this.value;
                    // While we are here we shoudl also replace any dependent variables.
                    if ($('#vpnl_listContainer' + binding.replace(/\./g, '_') + '_variable').length > 0) {
                        $('#vpnl_listContainer' + binding.replace(/\./g, '_') + '_variable').html(buildVariableDataList(deviceId, serviceId, binding));
                    }
                    return false;
                }
                return true;
            });
    }
    function onButtonStyleChanged(el, style) {
        if(style.substring(0, 10) == 'persistent') 
            $('tr.displayvar').show(); 
        else
            $('tr.displayvar').hide();
        if (style == 'toggle') {
            $('tr.btncmd_single').hide();
            $('tr.btncmd_multi').show();
            modalDialog.resizeTo("777px", true);
        }
        else if (style == 'persistenttoggle') {
            $('tr.btncmd_single').hide();
            $('tr.btncmd_multi').show();
            modalDialog.resizeTo("777px", true);
        }
        else {
            $('tr.btncmd_single').show();
            $('tr.btncmd_multi').hide();
            modalDialog.resizeTo("527px");
        }
    }
    function onVariableOperatorChanged(el, val) {
    	var tbl = $(el).parents('table.vpnl_deviceVariable');
    	switch (val) {
    		case 'between':
    			tbl.find('.vpnl_betweenOperator').show();
    			break;
    		default:
    			tbl.find('.vpnl_betweenOperator').hide();
    			break;
    	}
    }
    function onUseControlValueChecked(el, val) {
    	if ($(el).is(':checked'))
    		$(el).parents('tr').find('input.vpnl_setVariableValue').hide();
    	else
    		$(el).parents('tr').find('input.vpnl_setVariableValue').show();
    }
    function onSetCommandParams(el, sCmdType) { $(el).find('.cmdType').hide(); $(el).find('.cmdType' + sCmdType).show(); }
    function onAddActionParameterRow(tbl, binding) { tbl.find('tr:last').before(buildActionParameter(binding, '', '', false)); }
    function onReloadLuup() {
        modalDialog.showStatusDialog({
            title: "Performing Action",
            width: '350px',
            message: 'Reloading Luup... Please wait!'
        });
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_ReloadLuup",
                    deviceId: device,
                    rand: Math.random(),
                    output_format: "json"
                },
                onSuccess: function (response) {
                    modalDialog.closeDialog();
                },
                onFailure: function () {
                    modalDialog.showMessageDialog({
                        title: 'Error',
                        width: '350px',
                        message: 'Error reloading Luup'
                    });
                }
            });
        }
        catch (err) {
            alert('Error saving configuration');
        }
    }
    function onSaveDirtyPositions() {
        var arrPos = [];
        for (var i in dirtyPositions) {
            var el = dirtyPositions[i];
            arrPos.push({ ordinal: el.ordinal, left: el.position.left, top: el.position.top, width:el.position.width, height: el.position.height });
        }
        if (arrPos.length > 0)
            saveDirtyPositions(arrPos);
        else
            modalDialog.closeDialog();

    }
    function onRevertDirtyPositions() {
        modalDialog.showStatusDialog({
            title: "Reverting Positions...",
            width: '350px',
            message: 'Reverting elements to last saved positions... Please wait!'
        });
        loadPanelConfig('vpnl_panelconfig');
        modalDialog.closeDialog();
    }
    function onExecLuaCodeChanged(event, txtArea) {
        var cont = $(txtArea).parents('div.vpnl_execLuaContainer:first');
        var sNewCode = $(txtArea).val();
        var sOldCode = cont.find('span.vpnl_testedLua').text();
        var btn = cont.find('input.vpnl_testLuaBtn');

        if (sNewCode != sOldCode) {
            btn.data('iscodetested', false);
            btn.css({ backgroundColor: 'orangered', color: 'yellow' });
        }
        else {
            btn.data('iscodetested', true);
            btn.css({ backgroundColor: '', color: '' });
        }
    }
    function onCheckLuaSyntax(evt, button) {
        var btn = $(button);
        var cont = $(button).parents('div.vpnl_execLuaContainer:first');
        var txtArea = cont.find('textarea.vpnl_currentLua:first');
        var sNewCode = txtArea.val();
        console.log('Testing code:' + sNewCode);
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_CheckLuaSyntax",
                    deviceId: device,
                    luaCode: sNewCode.replace(/ /g, '%20'),
                    rand: Math.random(),
                    output_format: "json"
                },
                onSuccess: function (response) {
                    console.log(response);
                    var rc = JSON.parse(response.responseText);
                    if (rc != undefined) {
                        var msg = cont.find('div.vpnl_execLuaMessage:first');
                        if (rc.HasError == false) {
                            btn.data('iscodetested', true);
                            btn.css({ backgroundColor: '', color: '' });
                            cont.find('span.vpnl_testedLua').text(sNewCode);
                            msg.text(rc.Message);
                        }
                        else {
                            btn.data('iscodetested', true);
                            btn.css({ backgroundColor: 'orangered', color: 'yellow' });
                            cont.find('span.vpnl_testedLua').text('');
                            msg.text(rc.Message);
                        }
                    }
                    else {
                        alert(response.responseText);
                    }

                },
                onFailure: function (response) {
                    console.log("Error checking lua syntax" + response.statusText);
                }
            });
        }
        catch (err) {
            console.log('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Error checking Lua code.' + err.message + '</p>');
        }


    }
    function onTestLuaCode(evt, button) {
        var cont = $(button).parents('div.vpnl_execLuaContainer:first');
        var btn = cont.find('input.vpnl_testLuaBtn');
        var txtArea = cont.find('textarea.vpnl_currentLua:first');
        var sNewCode = txtArea.val();
        console.log('Testing code:' + sNewCode);
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_RunLua",
                    deviceId: device,
                    luaCode: sNewCode.replace(/ /g, '%20'),
                    rand: Math.random(),
                    output_format: "json"
                },
                onSuccess: function (response) {
                    //console.log(response);
                    var rc = JSON.parse(response.responseText);
                    if (rc != undefined) {
                        var msg = cont.find('div.vpnl_execLuaMessage:first');
                        if (rc.HasError == false) {
                            btn.data('iscodetested', true);
                            btn.css({ backgroundColor: '', color: '' });
                            cont.find('span.vpnl_testedLua').text(sNewCode);
                            msg.text(rc.Message);
                        }
                        else {
                            btn.data('iscodetested', true);
                            btn.css({ backgroundColor: 'orangered', color: 'yellow' });
                            cont.find('span.vpnl_testedLua').text('');
                            msg.text(rc.Message);
                        }
                    }
                    else {
                        alert(response.responseText);
                    }

                },
                onFailure: function (response) {
                    console.log("Error checking lua syntax" + response.statusText);
                }
            });
        }
        catch (err) {
            console.log('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Error checking Lua code.' + err.message + '</p>');
        }
    }

    /********************************************************
     * Loading functions
     ********************************************************/
    function loadIcons(panelId, sSelected) {
        $('#' + panelId).html('<div style="font-size:10pt;font-style:italic;">Loading Icons... pleast wait!</div>');
        if (iconList && iconList.loaded) {
            var html = buildIconList(iconList, sSelected);
            $('#' + panelId).html(html);
            if (typeof (sSelected) != 'undefined' && sSelected && sSelected.length > 0) {
                setTimeout(function () { $('#' + panelId).scrollTop($('#' + panelId).find('tr.vpnl_iconListSelected:first').position().top); }, 100);
            }
        }
        else {
            try {
                new Ajax.Request("../port_3480/data_request", {
                    method: "get",
                    parameters: {
                        id: "lr_vpnl_GetIcons",
                        rand: Math.random(),
                        output_format: "json"
                    },
                    onSuccess: function (response) {
                        var configuration = JSON.parse(response.responseText);
                        if (configuration != undefined) {
                            var html = '';
                            // Success. Populate the output with all the information from the device.
                            html += buildIconList(configuration, sSelected);
                            iconList = configuration;
                            iconList.loaded = true;
                            $('#' + panelId).html(html);
                            if (typeof (sSelected) != 'undefined' && sSelected && sSelected.length > 0) {
                                setTimeout(function () { $('#' + panelId).scrollTop($('#' + panelId).find('tr.vpnl_iconListSelected:first').position().top); }, 100);
                            }

                        }
                        else {
                            jQuery('#' + panelId).html(response.responseText);
                        }

                    },
                    onFailure: function (response) {
                        jQuery('#' + panelId).html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Could not get configuration for the panel:' + response.statusText + '</p>');
                    }
                });
            }
            catch (err) {
                jQuery('#' + panelId).html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Error getting configuration from panel.' + err.message + '</p>');
            }
        }
    }
    function loadAvailableDevices() {
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_GetAllDevices",
                    rand: Math.random(),
                    output_format: "json"
                },
                onSuccess: function (response) {
                    availableDevices = JSON.parse(response.responseText);
                },
                onFailure: function (response) {
                    jQuery('#vpnl_deviceresponse').html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Could not get devices from Vera:' + response.statusText + '</p>');
                }
            });
        }
        catch (err) {
            jQuery('#vpnl_deviceresponse').html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Error getting devices from Vera.' + err.message + '</p>');
        }
    }
    function loadUserData() {
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "status",
                    rand: Math.random(),
                    output_format: "json"
                },
                onSuccess: function (response) {
                    //jQuery('#vpnl_configresponse').html(response.responseText);
                    var configuration = JSON.parse(response.responseText);
                    if (configuration != undefined)
                        deviceList = configuration;
                },
                onFailure: function (response) {
                    alert(response);
                }
            });
        }
        catch (err) {
            alert(err);
        }
    }
    function loadPanelConfig(panelId) {
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_GetPanelConfiguration",
                    panelDeviceId: device,
                    rand: Math.random(),
                    output_format: "json"
                },
                onSuccess: function (response) {
                    $('#vpnl_configresponse').html(dialogManager.beautifyJson(response.responseText));
                    var configuration = JSON.parse(response.responseText);
                    if (configuration != undefined) {
                        // Success. Populate the output with all the information from the device.
                        panelConfig = configuration;
                        $('#vpnl_configresponse').html(JSON.stringify(panelConfig, null, '\t'));
                        if (typeof (panelConfig.deviceId) == 'undefined') {
                            $('#' + panelId).html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Could not get configuration for the panel (try reloading your browser). [' + response.responseText + ']</p>');
                        }
                        else {
                            $('#' + panelId).html(buildPanelConfig(configuration));
                            registerElementsSortable();
                        }
                    }
                    else
                        $('#' + panelId).html(response.responseText);
                },
                onFailure: function (response) {
                    jQuery('#' + panelId).html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Could not get configuration for the panel:' + response.statusText + '</p>');
                }
            });
        }
        catch (err) {
            $('#' + panelId).html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Error getting configuration from panel.' + err.message + '</p>');
        }
    }
    function loadTriggerConfig(panelId) {
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_GetPanelConfiguration",
                    panelDeviceId: device,
                    rand: Math.random(),
                    output_format: "json"
                },
                onSuccess: function (response) {
                    $('#vpnl_configresponse').html(dialogManager.beautifyJson(response.responseText));
                    var configuration = JSON.parse(response.responseText);
                    if (configuration != undefined) {
                        // Success. Populate the output with all the information from the device.
                        panelConfig = configuration;
                        $('#vpnl_configresponse').html(JSON.stringify(panelConfig, null, '\t'));
                        if (typeof (panelConfig.deviceId) == 'undefined') {
                            $('#' + panelId).html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Could not get configuration for the panel (try reloading your browser). [' + response.responseText + ']</p>');
                        }
                        else {
                            $('#' + panelId).html(buildTriggerConfig(configuration));
                        }
                    }
                    else
                        $('#' + panelId).html(response.responseText);
                },
                onFailure: function (response) {
                    jQuery('#' + panelId).html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Could not get trigger configuration for the panel:' + response.statusText + '</p>');
                }
            });
        }
        catch (err) {
            $('#' + panelId).html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Error getting trigger configuration from panel.' + err.message + '</p>');
        }
    }
    function loadStateConfig(panelId) {
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_GetPanelConfiguration",
                    panelDeviceId: device,
                    rand: Math.random(),
                    output_format: "json"
                },
                onSuccess: function (response) {
                    $('#vpnl_configresponse').html(dialogManager.beautifyJson(response.responseText));
                    var configuration = JSON.parse(response.responseText);
                    if (configuration != undefined) {
                        // Success. Populate the output with all the information from the device.
                        panelConfig = configuration;
                        $('#vpnl_configresponse').html(JSON.stringify(panelConfig, null, '\t'));
                        if (typeof (panelConfig.deviceId) == 'undefined') {
                            $('#' + panelId).html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Could not get configuration for the panel (try reloading your browser). [' + response.responseText + ']</p>');
                        }
                        else {
                            $('#' + panelId).html(buildStateConfig(configuration));
                        }
                    }
                    else
                        $('#' + panelId).html(response.responseText);
                },
                onFailure: function (response) {
                    jQuery('#' + panelId).html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Could not get state configuration for the panel:' + response.statusText + '</p>');
                }
            });
        }
        catch (err) {
            $('#' + panelId).html('<p style="font-weight: bold; text-align: center;background-color:red;color:yellow;">Error getting state configuration from panel.' + err.message + '</p>');
        }
    }

    /********************************************************
     * Configuration Saving functions
     ********************************************************/
    function saveIcons() {
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_SaveIcons",
                    deviceId: device,
                    rand: Math.random(),
                    stateIcons: JSON.stringify(panelConfig.stateIcons).replace(/ /g, '%20'),
                    defaultIcon: panelConfig.icon,
                    output_format: "json"
                },
                onSuccess: function (response) {
                    var configuration = JSON.parse(response.responseText);
                    if (configuration != undefined) {
                        var html = '';
                        // Success. Populate the output with all the information from the device.
                        panelConfig = configuration;
                        $('#vpnl_configresponse').html(JSON.stringify(panelConfig, null, '\t'));
                        $('#vpnl_stateIconDisplay').html(buildStateIconConfig(panelConfig));
                        $('#vpnl_panelElements').html(buildPanelElements(panelConfig));
                        registerElementsSortable();
                        modalDialog.closeDialog();
                    }
                },
                onFailure: function () {
                    modalDialog.showMessageDialog({
                        title: 'Error',
                        width: '350px',
                        message: 'Error deleting element'
                    });
                }
            });
        }
        catch (err) {
            alert('Error saving configuration');
        }

    }
    function addElement(elem) {
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_AddElement",
                    deviceId: device,
                    rand: Math.random(),
                    elem: JSON.stringify(elem).replace(/ /g, '%20'),
                    output_format: "json"
                },
                onSuccess: function (response) {
                    var configuration = JSON.parse(response.responseText);
                    if (configuration != undefined) {
                        var html = '';
                        // Success. Populate the output with all the information from the device.
                        panelConfig = configuration;
                        $('#vpnl_configresponse').html(JSON.stringify(panelConfig, null, '\t'));
                        $('#vpnl_panelElements').html(buildPanelElements(panelConfig));
                        registerElementsSortable();
                        modalDialog.closeDialog();
                    }
                },
                onFailure: function () {
                    modalDialog.showMessageDialog({
                        title: 'Error',
                        width: '350px',
                        message: 'Error adding element'
                    });
                }
            });
        }
        catch (err) {
            alert('Error saving configuration');
        }
    }
    function saveElement(ndx, elem) {
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_SaveElement",
                    deviceId: device,
                    rand: Math.random(),
                    ordinal: ndx + 1,
                    elem: JSON.stringify(elem).replace(/ /g, '%20'),
                    output_format: "json"
                },
                onSuccess: function (response) {
                    var configuration = JSON.parse(response.responseText);
                    if (configuration != undefined) {
                        // Success. Populate the output with all the information from the device.
                        panelConfig = configuration;
                        jQuery('#vpnl_configresponse').html(JSON.stringify(panelConfig, null, '\t'));
                        $('#vpnl_panelElements').html(buildPanelElements(panelConfig));
                        registerElementsSortable();
                        modalDialog.closeDialog();
                    }
                },
                onFailure: function () {
                    modalDialog.showMessageDialog({
                        title: 'Error',
                        width: '350px',
                        message: 'Error saving element'
                    });
                }
            });
        }
        catch (err) {
            alert('Error saving configuration');
        }
    }
    function saveDirtyPositions(arrPos) {
        //alert(JSON.stringify(arrPos));
        //return;
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_SaveDirtyPositions",
                    deviceId: device,
                    rand: Math.random(),
                    positions: JSON.stringify(arrPos).replace(/ /g, '%20'),
                    output_format: "json"
                },
                onSuccess: function (response) {
                    var configuration = JSON.parse(response.responseText);
                    if (configuration != undefined) {
                        // Success. Populate the output with all the information from the device.
                        panelConfig = configuration;
                        jQuery('#vpnl_configresponse').html(JSON.stringify(panelConfig, null, '\t'));
                        $('#vpnl_panelElements').html(buildPanelElements(panelConfig));
                        registerElementsSortable();
                        modalDialog.closeDialog();
                    }
                },
                onFailure: function () {
                    modalDialog.showMessageDialog({
                        title: 'Error',
                        width: '350px',
                        message: 'Error saving positions'
                    });
                }
            });
        }
        catch (err) {
            alert('Error saving positions');
        }
    }

    function deleteElement(ordinal) {
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_DeleteElement",
                    deviceId: device,
                    rand: Math.random(),
                    ordinal: ordinal,
                    output_format: "json"
                },
                onSuccess: function (response) {
                    var configuration = JSON.parse(response.responseText);
                    if (configuration != undefined) {
                        var sHtml = '';
                        // Success. Populate the output with all the information from the device.
                        panelConfig = configuration;
                        $('#vpnl_configresponse').html(JSON.stringify(panelConfig, null, '\t'));
                        $('#vpnl_panelElements').html(buildPanelElements(panelConfig));
                        registerElementsSortable();
                        modalDialog.closeDialog();
                    }
                },
                onFailure: function () {
                    modalDialog.showMessageDialog({
                        title: 'Error',
                        width: '350px',
                        message: 'Error deleting element'
                    });
                }
            });
        }
        catch (err) {
            alert('Error saving configuration');
        }
    }
    function moveElement(oldOrdinal, newOrdinal) {
        try {
            new Ajax.Request("../port_3480/data_request", {
                method: "get",
                parameters: {
                    id: "lr_vpnl_MoveElement",
                    deviceId: device,
                    rand: Math.random(),
                    oldOrdinal: oldOrdinal,
                    newOrdinal: newOrdinal,
                    output_format: "json"
                },
                onSuccess: function (response) {
                    var configuration = JSON.parse(response.responseText);
                    if (configuration != undefined) {
                        var sHtml = '';
                        // Success. Populate the output with all the information from the device.
                        panelConfig = configuration;
                        $('#vpnl_configresponse').html(JSON.stringify(panelConfig, null, "\t"));
                        $('#vpnl_panelElements').html(buildPanelElements(panelConfig));
                        registerElementsSortable();
                        modalDialog.closeDialog();
                    }
                },
                onFailure: function () {
                    modalDialog.showMessageDialog({
                        title: 'Error',
                        width: '350px',
                        message: 'Error deleting element'
                    });
                }
            });
        }
        catch (err) {
            alert('Error saving configuration');
        }
    }
    function onUIInit() {
        console.log('Hit the UI Init code');
    }
    function init() {
        // Register to events...
        api.registerEventHandler('on_ui_cpanel_before_close', myModule, 'onBeforeCpanelClose');
        //api.registerEventHandler('on_ui_initFinished', myModule, 'onUIInit');
        // Create Styles
        dialogManager.createStyles();
        textStyler.createStyles();
        // Set up the scrolling for the sample.
        $(window).on('scroll',
            function () {
                var elcont = $('#vpnl_elementsContainer');
                var samp = $('#vpnl_panelSamples');
                if (elcont[0] && samp[0]) {
                    rc = elcont[0].getBoundingClientRect();
                    samp.css({ top: ((rc.top < 0) ? -rc.top : 0) + 'px' });
                }

            });
        $(document.body).on('click', function () { $('#vpnl_rubberbox').hide(); });
        $(document).on('keydown',
            function (e) {
                if (e.which < 37 || e.which > 40) return;
                var box = $('#vpnl_rubberbox');
                var ord = box.data('controlordinal');
                var pos = box.position();
                var ctl = $('#vpnl_ctlElem' + ord);
                var l = 0;
                if (!box.is(':visible'))
                    return;
                var el = panelConfig.elements[ord];
                $('#vpnl_btnSavePositions').show();
                $('#vpnl_btnRevertPositions').show();
                if (e.shiftKey) {
                    switch (e.which) {
                        case 37: // left
                            if (ctl[0].clientWidth > 2) {
                                l = (Math.floor(parseInt(ctl[0].style.width, 10)) - 1);
                                ctl.find('.vpnl_ctlElemWidth').css({ width: l + 'px' });
                                if (ctl.hasClass('vpnl_ctlElemWidth')) {
                                    box.css({ width: l + 'px' });
                                    ctl.css({ width: l + 'px' });
                                    //console.log("WIDTH:%s OFFSET:%s JWIDTH:%s HWIDTH:%s", l, ctl[0].clientWidth, ctl.width(), parseInt(ctl[0].style.width, 10));
                                }
                                
                                if (!isNaN(l)) {
                                    el.position.width = l;
                                    dirtyPositions[el.ordinal] = el;
                                }
                            }
                            break;
                        case 38: // up
                            if (ctl[0].offsetHeight > 2) {
                                l = (Math.ceil(parseInt(ctl[0].style.height) - 1));
                                if (ctl.hasClass('vpnl_ctlElemHeight'))
                                    box.css({ height: l + 'px' });
                                ctl.find('.vpnl_ctlElemHeight').css({ height: l + 'px' });
                                if (ctl.hasClass('vpnl_ctlElemHeight')) {
                                    ctl.css({ height: l + 'px' });
                                }
                                if (!isNaN(l)) {
                                    el.position.height = l;
                                    dirtyPositions[el.ordinal] = el;
                                }
                            }
                            break;
                        case 39: // right
                            if (pos.left + ctl[0].clientWidth < ctl[0].offsetParent.clientWidth) {
                                l = (Math.ceil(parseInt(ctl[0].style.width, 10)) + 1);
                                if (ctl.hasClass('vpnl_ctlElemWidth')) {
                                    box.css({ width: l + 'px' });
                                    ctl.css({ width: l + 'px' });
                                }
                                ctl.find('.vpnl_ctlElemWidth').css({ width: l + 'px' });
                                //console.log("WIDTH:%s OFFSET:%s JWIDTH:%s HWIDTH:%s", l, ctl[0].clientWidth, ctl.width(), parseInt(ctl[0].style.width, 10));
                                if (!isNaN(l)) {
                                    dirtyPositions[el.ordinal] = el;
                                    el.position.width = l;
                                }
                            }
                            break;
                        case 40: //down
                            if (pos.top + ctl[0].offsetHeight < ctl[0].offsetParent.offsetHeight) {
                                l = (Math.ceil(parseInt(ctl[0].style.height) + 1));
                                if (ctl.hasClass('vpnl_ctlElemHeight')) {
                                    box.css({ height: l + 'px' });
                                    ctl.css({ height: l + 'px' });
                                }
                                ctl.find('.vpnl_ctlElemHeight').css({ height: l + 'px' });
                                if (!isNaN(l)) {
                                    el.position.height = l;
                                    dirtyPositions[el.ordinal] = el;
                                }
                            }
                            break;
                        default:
                            return;
                    }
                }
                else {
                    switch (e.which) {
                        case 37: // left
                            if (pos.left > 0) {
                                l = (Math.floor(pos.left) - 1);
                                box.css({ left: l + 'px' });
                                ctl.css({ left: l + 'px' });
                                el.position.left = l;
                                dirtyPositions[el.ordinal] = el;
                            }
                            break;
                        case 38: // up
                            if (pos.top > 0) {
                                l = (Math.floor(pos.top) - 1);
                                box.css({ top: l + 'px' });
                                ctl.css({ top: l + 'px' });
                                el.position.top = l;
                                dirtyPositions[el.ordinal] = el;
                            }
                            break;
                        case 39: // right
                            if (pos.left + box[0].offsetWidth < box[0].offsetParent.offsetWidth) {
                                l = (Math.ceil(pos.left) + 1);
                                box.css({ left: l + 'px' });
                                ctl.css({ left: l + 'px' });
                                el.position.left = l;
                                dirtyPositions[el.ordinal] = el;
                            }
                            break;
                        case 40: //down
                            if (pos.top + box[0].offsetHeight < box[0].offsetParent.offsetHeight) {
                                l = (Math.ceil(pos.top) + 1);
                                box.css({ top: l + 'px' });
                                ctl.css({ top: l + 'px' });
                                el.position.top = l;
                                dirtyPositions[el.ordinal] = el;
                            }
                            break;
                        default:
                            return;
                    }
                }
                e.preventDefault();
                e.stopPropagation();
            });
    }
    function onBeforeCpanelClose() {
        //console.log('handler for before cpanel close');
    }
    function elementSortHelper(e, tr) {
        var $originals = tr.children();
        var $helper = tr.clone();
        $helper.children().each(function(index) {
            $(this).width($originals.eq(index).width())
        });
        return $helper;
    }
    function moveArrayItem(arr, old_index, new_index) {
        while (old_index < 0) {
            old_index += arr.length;
        }
        while (new_index < 0) {
            new_index += arr.length;
        }
        if (new_index >= arr.length) {
            var k = new_index - arr.length;
            while ((k--) + 1) {
                arr.push(undefined);
            }
        }
        arr.splice(new_index, 0, arr.splice(old_index, 1)[0]);
        return arr; 
    };
    function updateElementIndexes(e, ui) {
        $('td.index', ui.item.parent()).each(function (i) {
            $(this).html(i + 1);
        });
        var oldIndex = parseInt($(ui.item).data('ordinal'), 10);
        var newIndex = parseInt(ui.item[0].rowIndex, 10) - 2;
        //alert('Updating index on ' + oldIndex + ' to ' + newIndex);
        moveArrayItem(panelConfig.elements, oldIndex, newIndex);
        $('#vpnl_panelElements').html(buildPanelElements(panelConfig));
        registerElementsSortable();
        moveElement(oldIndex + 1, newIndex + 1);
    }
    function updateDragElement(e, ui) {
        var ord = $('#vpnl_rubberbox').data('controlordinal');
        var pos = $('#vpnl_rubberbox').position();
        $('#vpnl_ctlElem' + ord).css({ top: Math.round(pos.top) + 'px', left: Math.round(pos.left) + 'px' });
        $('#vpnl_rubberbox').css({ top: Math.round(pos.top) + 'px', left: Math.round(pos.left) + 'px' });
        var el = panelConfig.elements[ord];
        if (el) {
            el.position.left = Math.round(pos.left);
            el.position.top = Math.round(pos.top);
            dirtyPositions[ord] = el;
            $('#vpnl_btnSavePositions').show();
            $('#vpnl_btnRevertPositions').show();
        }
    }
    function registerElementsSortable() {
        // Register Draggable on the elements grid.
        if (!$.isFunction($.fn['sortable'])) {
            alert('JQuery sortable function is no longer loaded on Vera');
        }
        $("#vpnl_elements tbody").sortable({
            containment: 'parent',
            helper: virtualPanelBuilder.elementSortHelper,
            stop: virtualPanelBuilder.updateElementIndexes,
            axis: 'y',
            cursorAt: { left: 5 }
        }).disableSelection();
        $('#vpnl_panelSamples').find('.vpnl_samplecontrol').on('click',
            function (event) {
                var pos = $(this).position();
                var offset = $(this).offset();
                var box = $('#vpnl_rubberbox');
                box.css({
                    display: '',
                    top: Math.round(pos.top) + 'px',
                    left: Math.round(pos.left) + 'px',
                    width: Math.round(this.offsetWidth) + 'px',
                    height: Math.round(this.offsetHeight) + 'px'
                }).data('controlordinal', $(this).data('controlordinal'));
                event.stopPropagation();
                event.preventDefault();
            });
        $('#vpnl_rubberbox').draggable({
            containment: 'parent',
            stop: virtualPanelBuilder.updateDragElement
        }).disableSelection();
        var elcont = $('#vpnl_elementsContainer');
        var samp = $('#vpnl_panelSamples');
        if (elcont[0] && samp[0]) {
            rc = elcont[0].getBoundingClientRect();
            samp.css({ top: ((rc.top < 0) ? -rc.top : 0) + 'px' });
        }
        else
            console.log('Not initialized %s %s', elcont[0], samp[0]);
    }
    function cleanUp() { }
    myModule = {
        uuid: uuid,
        init:init,
        showLayoutTab: showLayoutTab,
        showTriggersTab: showTriggersTab,
        showStatesTab: showStatesTab,
        cleanUp: cleanUp,
        onReloadLuup: onReloadLuup,
        onClickSelectDefaultIcon: onClickSelectDefaultIcon,
        onClickEditStateIcon: onClickEditStateIcon,
        onClickCreateStateIcon: onClickCreateStateIcon,
        onDeleteControlElement: onDeleteControlElement,
        onEditControlElement: onEditControlElement,
        onAddControlElement: onAddControlElement,
        onAddTrigger: onAddTrigger,
        onEditTrigger: onEditTrigger,
        onAddSate: onAddState,
        onEditState: onEditState,
        onReplaceServiceIdDataList: onReplaceServiceIdDataList,
        onReplaceVariableDataList: onReplaceVariableDataList,
        onSetCommandParams: onSetCommandParams,
        onAddActionParameterRow: onAddActionParameterRow,
        onButtonStyleChanged: onButtonStyleChanged,
		onUseControlValueChecked: onUseControlValueChecked,
		onVariableOperatorChanged: onVariableOperatorChanged,
        setListSelected: modalDialog.setListSelected,
        addIconStateConditionRow: addIconStateConditionRow,
        addColorExpressionRow: addColorExpressionRow,
        elementSortHelper: elementSortHelper,
        updateElementIndexes: updateElementIndexes,
        buildHorizontalRuleSample: buildHorizontalRuleSample,
        updateDragElement: updateDragElement,
        onSaveDirtyPositions: onSaveDirtyPositions,
        onRevertDirtyPositions: onRevertDirtyPositions,
        onBeforeCpanelClose: onBeforeCpanelClose,
        getControlTypeDetail: getControlTypeDetail,
        onUIInit: onUIInit,
        onExecLuaCodeChanged: onExecLuaCodeChanged,
        onCheckLuaSyntax: onCheckLuaSyntax,
        onTestLuaCode: onTestLuaCode
    };
    init();
    return myModule;
})(api);



